/*
 *	Copyright (c) 2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>

#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif
#include "gd.h"
#include "gdfonts.h"

#ifndef	IFDATA_FORMAT
#define	IFDATA_FORMAT	"ToOpPdDeE"
#endif

/* range of best dimensions in pixels */
#define	MIN_WIDTH	480
#define	MAX_WIDTH	1536
#define	MIN_HEIGHT	160
#define	MAX_HEIGHT	1024

/* own types */
typedef enum { Hour, Day, Week, Month, Year } TimeScale;
typedef	struct ifdata_entry {
	u_int32_t absTime;	/* time in seconds from Epoch (UTC) */
	u_int64_t inOctets;	/* input bytes */
	u_int64_t outOctets;	/* output bytes */
	u_int64_t inPackets;	/* input packets */
	u_int64_t outPackets;	/* output packets */
	u_int64_t inDrops;	/* discarded input packets */
	u_int64_t outDrops;	/* discarded output packets */
	u_int64_t inErrors;	/* input erorrs */
	u_int64_t outErrors;	/* output errors */

	struct ifdata_entry *next;
} IfDataEntry;

/* function prototypes */
void read_ifdata(FILE *fp, char *fmt, IfDataEntry **list);
int make_argv(char *buf, char ***avp, int max);
int stat_ifdata(IfDataEntry **list, IfDataEntry *stat);
unsigned int find_bandwidth(unsigned int value);
char *time_text(unsigned int abs_time);
char *bps_text(unsigned int bps);
char *pps_text(unsigned int pps);
char *strippath(char *path);
void draw_ifdata(gdImagePtr im, IfDataEntry *list, IfDataEntry *stat);
int draw_timescale(gdImagePtr img, int legend, int color,
		   TimeScale time_scale, IfDataEntry *list, IfDataEntry *stat);
int time_sharp(time_t *clock, TimeScale time_scale);
void draw_point(gdImagePtr img, int x1, int y1, int color);
void draw_line(gdImagePtr img, int x1, int y1, int x2, int y2, int color);
void draw_htext(gdImagePtr img, int x, int y, char *text, int color);
void draw_vtext(gdImagePtr img, int x, int y, char *text, int color);
void usage();

/* misc definitions */
#ifdef	DEBUG_IFGRAPH
void dump_ifdata(IfDataEntry *list, IfDataEntry *stat);
#define	trace(x)	{ printf x; }
#else
#define	trace(x)
#endif
#ifndef	ABS
#define	ABS(a)		((a) >= 0  ? (a) : -(a))
#endif
#ifndef	MAX
#define	MAX(a, b)	((b) < (a) ? (a) : (b))
#endif
#ifndef	MIN
#define	MIN(a, b)	((a) < (b) ? (a) : (b))
#endif

/* global variables */
int Black, White, Red, Green, Blue, Brown;
int XGap, YGap;
gdFontPtr CurrFont;
struct tm tim;

int
main(argc, argv)
	int argc;
	char **argv;
{
	int opt, width, height, bg_color;
	char *fmt;
	IfDataEntry *ifdata_list = 0, ifdata_stat;
	gdImagePtr image;
	extern int optind, opterr;
	extern char *optarg;

	/* set values by default */
	width = -1;
	height = -1;
	bg_color = -1;
	fmt = IFDATA_FORMAT;
	CurrFont = gdFontSmall;

	/* deal with command line arguments */
	opterr = 0;
	while ((opt = getopt(argc, argv, "b:f:g:?")) != EOF) {
		switch (opt) {
		case 'b': /* background color */
			if (sscanf(optarg, "%6x", &bg_color) != 1) {
				fprintf(stderr, "Wrong color format \"%s\"\n",
					optarg);
				exit(1);
			}
			break;
		case 'f': /* input data format */
			fmt = optarg;
			break;
		case 'g': /* image geometry */
			if (sscanf(optarg, "%dx%d", &width, &height) != 2) {
				fprintf(stderr, "Wrong geometry format \"%s\"\n",
					optarg);
				exit(1);
			}
			if (width < MIN_WIDTH || height < MIN_HEIGHT) {
				fprintf(stderr, "Geometry %dx%d too small\n",
					width, height);
				exit(1);
			}
			if (width > MAX_WIDTH || height > MAX_HEIGHT) {
				fprintf(stderr, "Geometry %dx%d too large\n",
					width, height);
				exit(1);
			}
			break;
		case '?':
		default:
			usage();
		}
	}
	if (argv[optind] != 0) {
		FILE *fp;
		do {
			fp = fopen(argv[optind], "r");
			if (!fp) {
				perror(argv[optind]);
				exit(1);
			}
			if (*fmt == 't') {
				if (sscanf(strippath(argv[optind]), "%u.%u.%u",
					   &tim.tm_year, &tim.tm_mon, &tim.tm_mday) != 3) {
					fprintf(stderr, "Wrong format name input files \"%s\"\n",
						argv[optind]);
					exit(1);
				}
				tim.tm_sec = tim.tm_min = tim.tm_hour = 0;
				tim.tm_mon -= 1;
				tim.tm_year -= 1900;
			}
			read_ifdata(fp, fmt, &ifdata_list);
			fclose(fp);
		} while (argv[++optind] != 0);
	} else {
		if (*fmt == 't') {
			struct timeval now;
			gettimeofday(&now, 0);
			memcpy(&tim, localtime((time_t *)&now.tv_sec),
				sizeof(struct tm));
		}
		read_ifdata(stdin, fmt, &ifdata_list);
	}

	if ((opt = stat_ifdata(&ifdata_list, &ifdata_stat)) < 2) {
		fprintf(stderr, "No input data collected\n");
		exit(1);
	}
#ifdef	DEBUG_IFGRAPH
	dump_ifdata(ifdata_list, &ifdata_stat);
#endif
	XGap = CurrFont->w * 4;
	YGap = CurrFont->h * 4;

	/* find best image size for the data */
	if (width < MIN_WIDTH || height < MIN_HEIGHT) {
		width = opt;
		while (width > MAX_WIDTH)
			width = width * 2 / 3;
		width -= width % 5;
		if (width < MIN_WIDTH)
			width = MIN_WIDTH;
		if (width >= 600)
			height = (int)((double)width * 0.25);
		else if (width >= 400)
			height = (int)((double)width * 0.33);
		else if (width >= 300)
			height = (int)((double)width * 0.4);
		else	height = (int)((double)width * 0.5);
		height -= height % 5;
		if (height < MIN_HEIGHT)
			height = MIN_HEIGHT;
	}
	image = gdImageCreate(width, height);
	if (!image) {
		perror("gdImageCreate");
		exit(1);
	}
	/* the first color allocated is background color */
	if (bg_color == 0) { /* make it transparent */
		bg_color = gdImageColorAllocate(image, 1, 1, 1);
		(void)gdImageColorTransparent(image, bg_color);
	} else if (bg_color != -1) {
		(void)gdImageColorAllocate(image, (bg_color >> 16) & 0xff,
					   (bg_color >> 8) & 0xff,
					   bg_color & 0xff);
	}
	White = gdImageColorAllocate(image, 0xff, 0xff, 0xff);
	Black = gdImageColorAllocate(image, 0x00, 0x00, 0x00);
	Red = gdImageColorAllocate(image, 0xff, 0x00, 0x00);
	Green = gdImageColorAllocate(image, 0x00, 0xaa, 0x00);
	Blue = gdImageColorAllocate(image, 0x10, 0x00, 0xff);
	Brown = gdImageColorAllocate(image, 0xff, 0xaa, 0x00);

	/* covered rectangle */
/*	gdImageRectangle(image,
			 XGap, CurrFont->h,
			 width - 1, height - 1 - YGap + CurrFont->h,
			 Black);
*/
	draw_ifdata(image, ifdata_list, &ifdata_stat);

	gdImagePng(image, stdout);
	gdImageDestroy(image);

	exit(0);
}

void
read_ifdata(fp, fmt, list)
	FILE *fp;
	char *fmt;
	IfDataEntry **list;
{
	int i, ac;
	char *cp, *av[100], buf[1024];
	IfDataEntry ifdata, *curr, *next, *prev;
	extern int errno;

	/* sanity check */
	if (!fp || !fmt || !list) return;

next_line:
	while (fgets(buf, sizeof(buf), fp) != 0) {
		if ((cp = strpbrk(buf, "\r\n")) != 0)
			*cp = '\0';
		if (buf[0] == '\0' || buf[0] == '#')
			continue; /* skip empty lines and comments */

		memset(&ifdata, 0, sizeof(ifdata));

		/* parse input data */
		ac = make_argv(buf, (char ***)av, 100);
		for (cp = fmt, i = 0; *cp && i < ac; ++cp, ++i) {
			errno = 0;
			switch (*cp) {
			case 't':
				if (sscanf(av[i], "%u:%u:%u", &tim.tm_hour,
					   &tim.tm_min, &tim.tm_sec) != 3) {
					fprintf(stderr, "Wrong format time %s (HH:MM:SS)\n", av[i]);
					errno = EINVAL;
				} else	ifdata.absTime = mktime(&tim);
				break;
			case 'T':
				ifdata.absTime = strtoll(av[i], 0, 10);
				break;
			case 'o':
				ifdata.inOctets = strtoll(av[i], 0, 10);
				break;
			case 'O':
				ifdata.outOctets = strtoll(av[i], 0, 10);
				break;
			case 'p':
				ifdata.inPackets = strtoll(av[i], 0, 10);
				break;
			case 'P':
				ifdata.outPackets = strtoll(av[i], 0, 10);
				break;
			case 'd':
				ifdata.inDrops = strtoll(av[i], 0, 10);
				break;
			case 'D':
				ifdata.outDrops = strtoll(av[i], 0, 10);
				break;
			case 'e':
				ifdata.inErrors = strtoll(av[i], 0, 10);
				break;
			case 'E':
				ifdata.outErrors = strtoll(av[i], 0, 10);
				break;
			}
			if (errno) goto next_line;
		}
		if (!ifdata.absTime)
			continue;

		/* make sure the list sorted by time descending! */
		prev = 0;
		for (next = *list; next; next = next->next) {
			if (ifdata.absTime >= next->absTime)
				break;
			prev = next;
		}
		if (next && next->absTime == ifdata.absTime) {
			curr = next;
			next = curr->next;
			ifdata.inOctets += curr->inOctets;
			ifdata.outOctets += curr->outOctets;
			ifdata.inPackets += curr->inPackets;
			ifdata.outPackets += curr->outPackets;
			ifdata.inDrops += curr->inDrops;
			ifdata.outDrops += curr->outDrops;
			ifdata.inErrors += curr->inErrors;
			ifdata.outErrors += curr->outErrors;
		} else {
			curr = (IfDataEntry *)malloc(sizeof(IfDataEntry));
			if (!curr) {
				perror("malloc");
				exit(1);
			}
			if (prev)
				prev->next = curr;
			else	*list = curr;
		}
		memcpy(curr, &ifdata, sizeof(IfDataEntry));
		curr->next = next;
	}
}

int
make_argv(buf, avp, max)
	char *buf, ***avp;
	int max;
{
	int ac, op;
	char *cp, *ap, **av;

	ac = op = 0;
	cp = buf;
	av = (char **)avp;
	while (cp && *cp && ac < max) {
		switch (*cp) {
		case ' ':
			if (op == ' ') {
				*cp = '\0';
				*av++ = ap;
				ac++;
				op = 0;
			}
			break;
		case '"':
			if (!op) {
				if (cp == buf || cp[-1] == ' ' || cp[-1] == '\0') {
					ap = ++cp;
					op = '"';
					continue;
				}
			} else if (op == '"' && (cp[1] == ' ' || cp[1] == '\0')) {
				*cp = '\0';
				*av++ = ap;
				ac++;
				op = 0;
			}
			break;
		case '<':
		case '(':
			if (!op && (cp == buf || cp[-1] == ' ' || cp[-1] == '\0')) {
				op = *cp;
				ap = ++cp;
				continue;
			}
			break;
		case '>':
		case ')':
			if (((op == '<' && *cp == '>') ||
			     (op == '(' && *cp == ')')) &&
			    (cp[1] == ' ' || cp[1] == '\0')) {
				*cp = '\0';
				*av++ = ap;
				ac++;
				op = 0;
				break;
			}
		default:
			if (!op && (cp == buf || cp[-1] == ' ' || cp[-1] == '\0')) {
				ap = cp;
				op = ' ';
			}
		}
		cp++;
	}
	if (op) {
		*av++ = ap;
		ac++;
	}
	return ac;
}

int
stat_ifdata(list, stat)
	IfDataEntry **list, *stat;
{
	int sec, cnt = 0;
	IfDataEntry *curr;

	if (!list || !stat) return -1;

	memset(stat, 0, sizeof(IfDataEntry));

	for (curr = *list; curr; curr = curr->next) {
		cnt++;
		sec = 0;
		if (curr->next)
			sec = ABS(curr->absTime - curr->next->absTime);
		else	stat->absTime = curr->absTime;
		if (!sec) {
			curr->inOctets = 0;
			curr->outOctets = 0;
			curr->inPackets = 0;
			curr->outPackets = 0;
			curr->inDrops = 0;
			curr->outDrops = 0;
			curr->inErrors = 0;
			curr->outErrors = 0;
			continue;
		}
		/* make procent regarding packets */
/* comments it out because too little values
 *		curr->inDrops = curr->inDrops * 100 / (curr->inPackets + 1);
 *		curr->outDrops = curr->outDrops * 100 / (curr->outPackets + 1);
 *		curr->inErrors = curr->inErrors * 100 / (curr->inPackets + 1);
 *		curr->outErrors = curr->outErrors * 100 / (curr->outPackets + 1);
 */
		/* make thing per seconds */
		curr->inOctets = (double)curr->inOctets * 8.0 / (double)sec;
		curr->outOctets = (double)curr->outOctets * 8.0 / (double)sec;
		curr->inPackets /= sec;
		curr->outPackets /= sec;

		/* collect max values */
		if (curr->inOctets > stat->inOctets)
			stat->inOctets = curr->inOctets;
		if (curr->outOctets > stat->outOctets)
			stat->outOctets = curr->outOctets;
		if (curr->inPackets > stat->inPackets)
			stat->inPackets = curr->inPackets;
		if (curr->outPackets > stat->outPackets)
			stat->outPackets = curr->outPackets;
		if (curr->inDrops > stat->inDrops)
			stat->inDrops = curr->inDrops;
		if (curr->outDrops > stat->outDrops)
			stat->outDrops = curr->outDrops;
		if (curr->inErrors > stat->inErrors)
			stat->inErrors = curr->inErrors;
		if (curr->outErrors > stat->outErrors)
			stat->outErrors = curr->outErrors;
	}
	return cnt;
}

unsigned int
find_bandwidth(value)
	unsigned int value;
{
	if (value > 1024000)
		return (value - value % 1024000 + 1024000);
	if (value > 1024)
		return (value - value % 1024 + 1024);
	return value;
}

char *
time_text(abs_time)
	unsigned int abs_time;
{
	time_t clock = abs_time;
	static char text[100];
	strftime(text, sizeof(text), "%R %e %b %Y", localtime(&clock));
	return text;
}

char *
bps_text(bps)
	unsigned int bps;
{
	static char text[100];
	if (bps >= 10000000)
		sprintf(text, "%uMbit/s", bps / 1000000);
	else if (bps >= 1000)
		sprintf(text, "%uKbit/s", bps / 1000);
	else	sprintf(text, "%ubit/s", bps);
	return text;
}

char *
pps_text(pps)
	unsigned int pps;
{
	static char text[100];
	if (pps >= 10000000)
		sprintf(text, "%uMpkt/s", pps / 1000000);
	else if (pps >= 1000)
		sprintf(text, "%uKpkt/s", pps / 1000);
	else	sprintf(text, "%upkt/s", pps);
	return text;
}

char *
strippath(path)
	char *path;
{
	register char *cp;
	if ((cp = strrchr(path, '/')) != 0)
		cp++;
	else	cp = path;
	return cp;
}

void
draw_ifdata(img, list, stat)
	gdImagePtr img;
	IfDataEntry *list, *stat;
{
	int total_width, total_height, mid_line, idx;
	unsigned int total_time, bandwidth;
	unsigned int max_octets, max_packets, max_drops, max_errors;
	int x1, x2, y1, y2, color, style_dotted[2];
	IfDataEntry *curr, *array, ave, max;
	double width_scale;
	double octets_scale, packets_scale, drops_scale, errors_scale;
	char *cp, text[100];

	if (!img || !list || !stat)
		return;

	total_width = gdImageSX(img) - XGap;
	total_height = gdImageSY(img) - YGap;
	mid_line = total_height / 2;

	array = (IfDataEntry *)malloc(sizeof(IfDataEntry) * total_width);
	if (!array) {
		perror("malloc");
		return;
	}
	memset(array, 0, sizeof(IfDataEntry) * total_width);

	total_time = ABS(list->absTime - stat->absTime);
	if (total_time)
		width_scale = (double)(total_width - 1) / (double)total_time;
	else	width_scale = 0;

	/*
	 * Draw curves.
	 */

	/* reflect data to x coordinate */
	for (curr = list; curr; curr = curr->next) {
		idx = (int)((double)ABS(curr->absTime - stat->absTime) * width_scale);
		if (idx < 0 || idx > total_width-1)
			continue; /* should not happen */
		array[idx].absTime++;
		array[idx].inOctets += curr->inOctets;
		array[idx].outOctets += curr->outOctets;
		array[idx].inPackets += curr->inPackets;
		array[idx].outPackets += curr->outPackets;
		array[idx].inDrops += curr->inDrops;
		array[idx].outDrops += curr->outDrops;
		array[idx].inErrors += curr->inErrors;
		array[idx].outErrors += curr->outErrors;
	}
	memset(&max, 0, sizeof(IfDataEntry));
	memset(&ave, 0, sizeof(IfDataEntry));
	for (idx = 0; idx < total_width; ++idx) {
		if (!array[idx].absTime)
			continue;
		array[idx].inOctets /= array[idx].absTime;
		array[idx].outOctets /= array[idx].absTime;
		array[idx].inPackets /= array[idx].absTime;
		array[idx].outPackets /= array[idx].absTime;
		array[idx].inDrops /= array[idx].absTime;
		array[idx].outDrops /= array[idx].absTime;
		array[idx].inErrors /= array[idx].absTime;
		array[idx].outErrors /= array[idx].absTime;

		if (array[idx].inOctets > max.inOctets)
			max.inOctets = array[idx].inOctets;
		if (array[idx].outOctets > max.outOctets)
			max.outOctets = array[idx].outOctets;
		if (array[idx].inPackets > max.inPackets)
			max.inPackets = array[idx].inPackets;
		if (array[idx].outPackets > max.outPackets)
			max.outPackets = array[idx].outPackets;
		if (array[idx].inDrops > max.inDrops)
			max.inDrops = array[idx].inDrops;
		if (array[idx].outDrops > max.outDrops)
			max.outDrops = array[idx].outDrops;
		if (array[idx].inErrors > max.inErrors)
			max.inErrors = array[idx].inErrors;
		if (array[idx].outErrors > max.outErrors)
			max.outErrors = array[idx].outErrors;

		ave.absTime++;
		ave.inOctets += array[idx].inOctets;
		ave.outOctets += array[idx].outOctets;
		ave.inPackets += array[idx].inPackets;
		ave.outPackets += array[idx].outPackets;
		ave.inDrops += array[idx].inDrops;
		ave.outDrops += array[idx].outDrops;
		ave.inErrors += array[idx].inErrors;
		ave.outErrors += array[idx].outErrors;
	}
	if (ave.absTime) {
		ave.inOctets /= ave.absTime;
		ave.outOctets /= ave.absTime;
		ave.inPackets /= ave.absTime;
		ave.outPackets /= ave.absTime;
		ave.inDrops /= ave.absTime;
		ave.outDrops /= ave.absTime;
		ave.inErrors /= ave.absTime;
		ave.outErrors /= ave.absTime;
	}

	max_octets = MAX(max.inOctets, max.outOctets);
	bandwidth = find_bandwidth(max_octets);
	if (bandwidth)
		octets_scale = (double)(mid_line - 1) / (double)bandwidth;
	else	octets_scale = 0;

	max_packets = MAX(max.inPackets, max.outPackets);
	if (max_packets)
		packets_scale = (double)(mid_line - 1) / (double)max_packets;
	else	packets_scale = 0;

	max_drops = MAX(max.inDrops, max.outDrops);
	if (max_drops)
		drops_scale = (double)(mid_line - 1) / (double)max_drops;
	else	drops_scale = 0;

	max_errors = MAX(max.inErrors, max.outErrors);
	if (max_errors)
		errors_scale = (double)(mid_line - 1) / (double)max_errors;
	else	errors_scale = 0;


	/* draw curves */
	for (idx = 0; idx < total_width; ++idx) {
		x1 = x2 = idx;

		/* Input axis Up */
		y1 = mid_line - 1;

		if (array[idx].inOctets) {
			y2 = y1 - (int)((double)array[idx].inOctets * octets_scale);
			draw_line(img, x1, y1, x2, y2, Green);
		}
		if (array[idx].inPackets) {
			y2 = y1 - (int)((double)array[idx].inPackets * packets_scale);
			draw_point(img, x2, y2, Blue);
		}
		if (array[idx].inDrops) {
			y2 = y1 - (int)((double)array[idx].inDrops * drops_scale);
			draw_point(img, x2, y2, Brown);
		}
		if (array[idx].inErrors) {
			y2 = y1 - (int)((double)array[idx].inErrors * errors_scale);
			draw_point(img, x2, y2, Red);
		}

		/* Output axis Down */
		y1 = mid_line + 1;

		if (array[idx].outOctets) {
			y2 = y1 + (int)((double)array[idx].outOctets * octets_scale);
			draw_line(img, x1, y1, x2, y2, Green);
		}
		if (array[idx].outPackets) {
			y2 = y1 + (int)((double)array[idx].outPackets * packets_scale);
			draw_point(img, x2, y2, Blue);
		}
		if (array[idx].outDrops) {
			y2 = y1 + (int)((double)array[idx].outDrops * drops_scale);
			draw_point(img, x2, y2, Brown);
		}
		if (array[idx].outErrors) {
			y2 = y1 + (int)((double)array[idx].outErrors * errors_scale);
			draw_point(img, x2, y2, Red);
		}
	}

	/*
	 * Draw axis.
	 */

	style_dotted[0] = Black;
	style_dotted[1] = gdTransparent;
	gdImageSetStyle(img, style_dotted, 2);

	/* middle line horizontal axis */
	draw_line(img, 0, mid_line, total_width - 1, mid_line, Black);

	cp = time_text(MIN(list->absTime, stat->absTime));
	x1 = 0 - CurrFont->h - 1;
	y1 = mid_line + strlen(cp) * CurrFont->w / 2;
	draw_vtext(img, x1, y1, cp, Black);

	cp = time_text(MAX(list->absTime, stat->absTime));
	x2 = total_width - 1;
	y2 = mid_line + strlen(cp) * CurrFont->w / 2;
	draw_vtext(img, x2, y2, cp, Black);

	/* left vertical axis */
	draw_line(img, 0, -1, 0, total_height, Black);

	cp = "In";
	x1 = 0 - strlen(cp) * CurrFont->w / 2;
	y1 = 0 - CurrFont->h;
	draw_htext(img, x1, y1, cp, Black);
	x1 += (strlen(cp) + 1) * CurrFont->w;
	cp = "Out";
	x2 = 0 - strlen(cp) * CurrFont->w / 2;
	y2 = total_height - 1;
	draw_htext(img, x2, y2, cp, Black);
	x2 += (strlen(cp) + 1) * CurrFont->w;
	cp = bps_text(bandwidth);
	draw_htext(img, x1, y1, cp, Black);
	draw_htext(img, x2, y2, cp, Black);

	/* horizontal scale grid */
	idx = (int)((double)bandwidth * octets_scale);
	if (idx / 4 > CurrFont->h * 3 / 2)
		idx /= 4;
	else	idx /= 2;
	if (idx > CurrFont->h) {
		y1 = (mid_line - 1) - idx;
		y2 = (mid_line + 1) + idx;
		for (; y1 >= 0 && y2 < total_height; y1 -= idx, y2 += idx) {
			draw_line(img, 0, y1, total_width - 1, y1, gdStyled);
			draw_line(img, 0, y2, total_width - 1, y2, gdStyled);
		}
	}

	/* vertical time grid */
	idx = 0;
	color = gdStyled;
	if (draw_timescale(img, !idx, color, Hour, list, stat)) {
		color = Red;
		gdImageSetThickness(img, ++idx);
	}
	if (draw_timescale(img, !idx, color, Day, list, stat)) {
		color = Red;
		gdImageSetThickness(img, ++idx);
	}
	if (draw_timescale(img, !idx, color, Week, list, stat)) {
		color = Red;
		gdImageSetThickness(img, ++idx);
	}
	if (draw_timescale(img, !idx, color, Month, list, stat)) {
		color = Red;
		gdImageSetThickness(img, ++idx);
	}
	draw_timescale(img, !idx, color, Year, list, stat);
	gdImageSetThickness(img, 1);

	/* print statistics */
	x1 = total_width - CurrFont->w * 80;
	if (x1 < 0) x1 = 0;
	x2 = x1;
	y1 = 0 - CurrFont->h * 2;
	y2 = total_height - 1 + CurrFont->h;

	cp = "Max:";
	draw_htext(img, x1, y1, cp, Black);
	x1 += (strlen(cp) + 1) * CurrFont->w;
	draw_htext(img, x2, y2, cp, Black);
	x2 += (strlen(cp) + 1) * CurrFont->w;
	if (max_octets) {
		cp = bps_text(max.inOctets);
		draw_htext(img, x1, y1, cp, Green);
		x1 += (strlen(cp) + 1) * CurrFont->w;
		cp = bps_text(max.outOctets);
		draw_htext(img, x2, y2, cp, Green);
		x2 += (strlen(cp) + 1) * CurrFont->w;
	}
	if (max_packets) {
		cp = pps_text(max.inPackets);
		draw_htext(img, x1, y1, cp, Blue);
		x1 += (strlen(cp) + 1) * CurrFont->w;
		cp = pps_text(max.outPackets);
		draw_htext(img, x2, y2, cp, Blue);
		x2 += (strlen(cp) + 1) * CurrFont->w;
	}
	cp = "Ave:";
	draw_htext(img, x1, y1, cp, Black);
	x1 += (strlen(cp) + 1) * CurrFont->w;
	draw_htext(img, x2, y2, cp, Black);
	x2 += (strlen(cp) + 1) * CurrFont->w;
	if (max_octets) {
		cp = bps_text(ave.inOctets);
		draw_htext(img, x1, y1, cp, Green);
		x1 += (strlen(cp) + 1) * CurrFont->w;
		cp = bps_text(ave.outOctets);
		draw_htext(img, x2, y2, cp, Green);
		x2 += (strlen(cp) + 1) * CurrFont->w;
	}
	if (max_packets) {
		cp = pps_text(ave.inPackets);
		draw_htext(img, x1, y1, cp, Blue);
		x1 += (strlen(cp) + 1) * CurrFont->w;
		cp = pps_text(ave.outPackets);
		draw_htext(img, x2, y2, cp, Blue);
		x2 += (strlen(cp) + 1) * CurrFont->w;
	}
	cp = "New:";
	draw_htext(img, x1, y1, cp, Black);
	x1 += (strlen(cp) + 1) * CurrFont->w;
	draw_htext(img, x2, y2, cp, Black);
	x2 += (strlen(cp) + 1) * CurrFont->w;
	if (max_octets) {
		cp = bps_text(list->inOctets);
		draw_htext(img, x1, y1, cp, Green);
		x1 += (strlen(cp) + 1) * CurrFont->w;
		cp = bps_text(list->outOctets);
		draw_htext(img, x2, y2, cp, Green);
		x2 += (strlen(cp) + 1) * CurrFont->w;
	}
	if (max_packets) {
		cp = pps_text(list->inPackets);
		draw_htext(img, x1, y1, cp, Blue);
		x1 += (strlen(cp) + 1) * CurrFont->w;
		cp = pps_text(list->outPackets);
		draw_htext(img, x2, y2, cp, Blue);
		x2 += (strlen(cp) + 1) * CurrFont->w;
	}
}

int
draw_timescale(img, legend, color, time_scale, list, stat)
	gdImagePtr img;
	int legend, color;
	TimeScale time_scale;
	IfDataEntry *list, *stat;
{
	int total_width, total_height, ntimes, nlines;
	int x1, x2, y1, y2, xprev, range;
	unsigned int total_time, time_factor;
	time_t clock;
	double width_scale;
	char *time_fmt, text[100];

	if (!img || !list || !stat)
		return 0;

	switch (time_scale) {
	case Hour:
		time_fmt = "%k";
		time_factor = 3600;
		range = 0;
		break;
	case Day:
		time_fmt = "%a%e";
		time_factor = 86400;
		range = 1;
		break;
	case Week:
		time_fmt = "%e";
		time_factor = 7 * 86400;
		range = 0;
		break;
	case Month:
		time_fmt = "%b%y";
		time_factor = 28 * 86400;
		range = 1;
		break;
	case Year:
		time_fmt = "%Y";
		time_factor = 365 * 86400;
		range = 1;
		break;
	default:
		return 0; /* unknown scale factor */
	}

	total_width = gdImageSX(img) - XGap;
	total_height = gdImageSY(img) - YGap;

	total_time = ABS(list->absTime - stat->absTime);
	ntimes = total_time / time_factor;

	nlines = CurrFont->w * 2;
	if (time_scale != Hour)
		nlines *= 2;

	if (ntimes < 1 || total_width / ntimes < nlines)
		return 0;

	width_scale = (double)(total_width - 1) / (double)total_time;
	nlines = 0;
	xprev = -1;
	clock = MAX(list->absTime, stat->absTime);
	while (time_sharp(&clock, time_scale) != -1) {
		ntimes = clock - MIN(list->absTime, stat->absTime);
		if (ntimes < 0) break;
		x1 = x2 = (int)((double)ntimes * width_scale);
		y1 = 0;
		y2 = total_height - 1;
		if (legend && !range) {
			if (nlines % 2)
				y1 += CurrFont->h;
			else	y2 -= CurrFont->h;
		}
		draw_line(img, x1, y1, x2, y2, color);
		if (legend && xprev != -1) {
			strftime(text, sizeof(text), time_fmt, localtime(&clock));
			x1 -= strlen(text) * CurrFont->w / 2;
			if (range)
				x1 += ABS(xprev - x2) / 2;
			if (nlines % 2)
				y1 = 0;
			else	y1 = (total_height - 1) - CurrFont->h;
			draw_htext(img, x1, y1, text, Black);
		}
		nlines++;
		xprev = x2;
		clock -= time_factor;
	}
	return nlines;
}

int
time_sharp(clock, time_scale)
	time_t *clock;
	TimeScale time_scale;
{
	int rval = -1;
	struct tm *tp = localtime(clock);

	if (!tp) return rval;

	*clock -= tp->tm_sec;
	if (time_scale >= Hour) {
		*clock -= tp->tm_min * 60;
		rval = 0;
	}
	if (time_scale >= Day) {
		*clock -= tp->tm_hour * 3600;
		rval = 0;
	}
	if (time_scale == Week) {
		*clock -= (tp->tm_wday ? (tp->tm_wday - 1) : 6) * 86400;
		rval = 0;
	}
	if (time_scale == Month) {
		*clock -= (tp->tm_mday - 1) * 86400;
		rval = 0;
	}
	if (time_scale == Year) {
		*clock -= tp->tm_yday * 86400;
		/*if (tp->tm_mon > 1 && ((tp->tm_year + 1900) % 4) == 0)
			*clock -= 86400;*/
		rval = 0;
	}
	return rval;
}

void
draw_point(img, x, y, color)
	gdImagePtr img;
	int x, y, color;
{
	if (img) {
		gdImageSetPixel(img,
				x + XGap - CurrFont->w * 2,
				y + CurrFont->h * 2,
				color);
	}
}

void
draw_line(img, x1, y1, x2, y2, color)
	gdImagePtr img;
	int x1, y1, x2, y2, color;
{
	if (img) {
		gdImageLine(img,
			    x1 + XGap - CurrFont->w * 2,
			    y1 + CurrFont->h * 2,
			    x2 + XGap - CurrFont->w * 2,
			    y2 + CurrFont->h * 2,
			    color);
	}
}

void
draw_htext(img, x, y, text, color)
	gdImagePtr img;
	int x, y, color;
	char *text;
{
	if (img && text) {
		gdImageString(img, CurrFont,
			      x + XGap - CurrFont->w * 2,
			      y + CurrFont->h * 2,
			      text, color);
	}
}

void
draw_vtext(img, x, y, text, color)
	gdImagePtr img;
	int x, y, color;
	char *text;
{
	if (img && text) {
		gdImageStringUp(img, CurrFont,
				x + XGap - CurrFont->w * 2,
				y + CurrFont->h * 2,
				text, color);
	}
}

void
usage()
{
	fprintf(stderr, "usage:\n\
 ifgraph [-b color] [-f fmt] [-g size] [files...]\n");

	fprintf(stderr, "where:\n\
 -b color Set image background RGB color, use 0 for transparent\n\
 -f fmt   Input data format, default `%s' (o=ifInOctets,O=ifOutOctets...)\n\
 -g size  Set image geometry as WxH pixels, auto by default\n\
 files    List of input files, default `stdin'\n",
		IFDATA_FORMAT);
	exit(1);
}

#ifdef	DEBUG_IFGRAPH
void
dump_ifdata(list, stat)
	IfDataEntry *list, *stat;
{
	IfDataEntry *curr;

	if (stat) {
		printf("STAT: %u %llu %llu %llu %llu %llu %llu %llu %llu\n",
		       stat->absTime,
		       stat->inOctets,
		       stat->outOctets,
		       stat->inPackets,
		       stat->outPackets,
		       stat->inDrops,
		       stat->outDrops,
		       stat->inErrors,
		       stat->outErrors);
	}
	for (curr = list; curr; curr = curr->next) {
		printf("%u %llu %llu %llu %llu %llu %llu %llu %llu\n",
		       curr->absTime,
		       curr->inOctets,
		       curr->outOctets,
		       curr->inPackets,
		       curr->outPackets,
		       curr->inDrops,
		       curr->outDrops,
		       curr->inErrors,
		       curr->outErrors);
	}
}
#endif
