/*
 *	Copyright (c) 1999-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifndef	__MIB_H_
#define	__MIB_H_

#include "netmon.h"

/*
 * Built-in SNMP variables declaration.
 */

/*
 * System variables.
 */
extern const oid sysObjID[];
extern const int sysObjID_len;
extern const oid sysUpTime[];
extern const oid sysContact[];
extern const oid sysName[];
extern const oid sysLocation[];
extern const oid sysServices[];	/* bit-mapped OSI layers: */
				/* 1 - physical		*/
				/* 2 - data link	*/
				/* 4 - network		*/
				/* 8 - transport	*/
				/* 64 - application	*/

/*
 * Interface variables according to MIB-2 IF-MIB.txt
 */
extern const oid ifNumber[];		/* the Number of network interfaces */
extern const oid ifTableLastChange[];	/* sysUpTime at the time of the last creation
				or deletion of an entry in the ifTable or 0 */

/*
 * The Interfaces table contains information on the entity's
 * interfaces. Each sub-layer below the internetwork-layer
 * of a network interface is considered to be an interface.
 * An entry containing management information applicable to
 * a particular interface.
 */
extern const oid ifIndex[];	/* Interface Index */
extern const oid ifDescr[];	/* Display String */
extern const oid ifType[];	/* IANAifType -- the type of interface */
extern const oid ifMtu[];	/* MTU on the interface */
extern const oid ifSpeed[];	/* Interface bandwidth in bits per second */
extern const oid ifPhysAddress[];/* MAC address? */
extern const oid ifAdminStatus[];/* up(1), down(2), testing(3) */
extern const oid ifOperStatus[];/* up(1), down(2), testing(3), unknown(4),
				dormant(5), notPresent(6), lowerLayerDown(7) */
extern const oid ifLastChange[];/* sysUpTime the interface entered its
				current operational state */
extern const oid ifInOctets[];	/* The total number of octets received on the
				interface, including framing characters */
extern const oid ifInUcastPkts[];/* unicast packets */
extern const oid ifInNUcastPkts[];/* broadcast & multicast packets -- deprecated */
extern const oid ifInDrops[];	/* packets to be discarded even though no
				errors has been detected */
extern const oid ifInErrors[];	/* packets that contained errors */
extern const oid ifInUnknownProtos[];/* unknown or unsupported protocol packets */

extern const oid ifOutOctets[];	/* the same as input */
extern const oid ifOutUcastPkts[];/* the same as input */
extern const oid ifOutNUcastPkts[];/* deprecated */
extern const oid ifOutDrops[];	/* the same as input */
extern const oid ifOutErrors[];	/* the same as input */
extern const oid ifOutQLen[];	/* The length of the output packet queue (in packets) */

/*
 * Extension to the interface table.
 */
extern const oid ifName[];		/* The textual name of the interface */
extern const oid ifInMulticastPkts[];	/* multicas packets */
extern const oid ifInBroadcastPkts[];	/* broadcast packets */
extern const oid ifOutMulticastPkts[];	/* the same as input */
extern const oid ifOutBroadcastPkts[];	/* the same as input */
extern const oid ifHCInOctets[];	/* 64-bit version of ifInOctets */
extern const oid ifHCInUcastPkts[];	/* 64-bit version of ifInUcastPkts */
extern const oid ifHCInMulticastPkts[];	/* 64-bit version of ifInMulticastPkts */
extern const oid ifHCInBroadcastPkts[];	/* 64-bit version of ifInBroadcastPkts */
extern const oid ifHCOutOctets[];	/* the same as input */
extern const oid ifHCOutUcastPkts[];	/* the same as input */
extern const oid ifHCOutMulticastPkts[];/* the same as input */
extern const oid ifHCOutBroadcastPkts[];/* the same as input */
extern const oid ifLinkUpDownTrapEnable[];/* enabled(1), disabled(2) */
extern const oid ifHighSpeed[];		/* bandwidth in units of 1,000,000
					bits per second */
extern const oid ifPromiscuousMode[];	/* true(1), false(2) */
extern const oid ifConnectorPresent[];	/* cable attached? true(1), false(2) */
extern const oid ifAlias[];		/* 'alias' name for the interface as
					specified by a network manager */
extern const oid ifCounterDiscontinuityTime[];/* sysUpTime on the most recent occasion
					at which any one or more of this interface's
					counters suffered a discontinuity */

/*
 * Cisco specific variables.
 */
extern const oid ciscoProductID[];
extern const int ciscoProductID_len;
extern const oid ciscoWhyReload[];
extern const oid ciscoFreeMem[];
extern const oid ciscoAvgBusy[];
extern const oid ciscoIfReason[];
extern const oid ciscoIfDescr[];
extern const oid ciscoEnvTempDescr[];
extern const oid ciscoEnvTempValue[];
extern const oid ciscoEnvTempState[];

extern const oid catalystProductID[];
extern const int catalystProductID_len;

/*
 * BGP router variables (see BGP4-MIB: RFC-1657).
 */
extern const oid bgpLocalAs[];
extern const oid bgpPeerIdentifier[];
extern const oid bgpPeerState[];
extern const oid bgpPeerAdminStatus[];
extern const oid bgpPeerNegotiatedVersion[];
extern const oid bgpPeerLocalAddr[];
extern const oid bgpPeerLocalPort[];
extern const oid bgpPeerRemoteAddr[];
extern const int bgpPeerRemoteAddr_len;
extern const oid bgpPeerRemotePort[];
extern const oid bgpPeerRemoteAs[];
extern const int bgpPeerRemoteAs_len;
extern const oid bgpPeerInUpdates[];
extern const oid bgpPeerOutUpdates[];
extern const oid bgpPeerInTotalMessages[];
extern const oid bgpPeerOutTotalMessages[];
extern const oid bgpPeerLastError[];
extern const oid bgpPeerFsmEstablishedTime[];
extern const oid bgpPeerInUpdateElapsedTime[];
extern const oid bgpIdentifier[];
extern const oid bgpTrapEnterprise[];
extern const int bgpTrapEnterprise_len;

/* IANA interface types */
#define ifType_other 1
#define ifType_regular1822 2
#define ifType_hdh1822 3
#define ifType_ddnX25 4
#define ifType_rfc877x25 5
#define ifType_ethernetCsmacd 6
#define ifType_iso88023Csmacd 7
#define ifType_iso88024TokenBus 8
#define ifType_iso88025TokenRing 9
#define ifType_iso88026Man 10
#define ifType_starLan 11
#define ifType_proteon10Mbit 12
#define ifType_proteon80Mbit 13
#define ifType_hyperchannel 14
#define ifType_fddi 15
#define ifType_lapb 16
#define ifType_sdlc 17
#define ifType_ds1 18
#define ifType_e1 19
#define ifType_basicISDN 20
#define ifType_primaryISDN 21
#define ifType_propPointToPointSerial 22
#define ifType_ppp 23
#define ifType_softwareLoopback 24
#define ifType_eon 25
#define ifType_ethernet3Mbit 26
#define ifType_nsip 27
#define ifType_slip 28
#define ifType_ultra 29
#define ifType_ds3 30
#define ifType_sip 31
#define ifType_frameRelay 32
#define ifType_rs232 33
#define ifType_para 34
#define ifType_arcnet 35
#define ifType_arcnetPlus 36
#define ifType_atm 37
#define ifType_miox25 38
#define ifType_sonet 39
#define ifType_x25ple 40
#define ifType_iso88022llc 41
#define ifType_localTalk 42
#define ifType_smdsDxi 43
#define ifType_frameRelayService 44
#define ifType_v35 45
#define ifType_hssi 46
#define ifType_hippi 47
#define ifType_modem 48
#define ifType_aal5 49
#define ifType_sonetPath 50
#define ifType_sonetVT 51
#define ifType_smdsIcip 52
#define ifType_propVirtual 53
#define ifType_propMultiplexor 54
#define ifType_ieee80212 55
#define ifType_fibreChannel 56
#define ifType_hippiInterface 57
#define ifType_frameRelayInterconnect 58
#define ifType_aflane8023 59
#define ifType_aflane8025 60
#define ifType_cctEmul 61
#define ifType_fastEther 62
#define ifType_isdn 63
#define ifType_v11 64
#define ifType_v36 65
#define ifType_g703at64k 66
#define ifType_g703at2mb 67
#define ifType_qllc 68
#define ifType_fastEtherFX 69
#define ifType_channel 70
#define ifType_ieee80211 71
#define ifType_ibm370parChan 72
#define ifType_escon 73
#define ifType_dlsw 74
#define ifType_isdns 75
#define ifType_isdnu 76
#define ifType_lapd 77
#define ifType_ipSwitch 78
#define ifType_rsrb 79
#define ifType_atmLogical 80
#define ifType_ds0 81
#define ifType_ds0Bundle 82
#define ifType_bsc 83
#define ifType_async 84
#define ifType_cnr 85
#define ifType_iso88025Dtr 86
#define ifType_eplrs 87
#define ifType_arap 88
#define ifType_propCnls 89
#define ifType_hostPad 90
#define ifType_termPad 91
#define ifType_frameRelayMPI 92
#define ifType_x213 93
#define ifType_adsl 94
#define ifType_radsl 95
#define ifType_sdsl 96
#define ifType_vdsl 97
#define ifType_iso88025CRFPInt 98
#define ifType_myrinet 99
#define ifType_voiceEM 100
#define ifType_voiceFXO 101
#define ifType_voiceFXS 102
#define ifType_voiceEncap 103
#define ifType_voiceOverIp 104
#define ifType_atmDxi 105
#define ifType_atmFuni 106
#define ifType_atmIma 107
#define ifType_pppMultilinkBundle 108
#define ifType_ipOverCdlc 109
#define ifType_ipOverClaw 110
#define ifType_stackToStack 111
#define ifType_virtualIpAddress 112
#define ifType_mpc 113
#define ifType_ipOverAtm 114
#define ifType_iso88025Fiber 115
#define ifType_tdlc 116
#define ifType_gigabitEthernet 117
#define ifType_hdlc 118
#define ifType_lapf 119
#define ifType_v37 120
#define ifType_x25mlp 121
#define ifType_x25huntGroup 122
#define ifType_trasnpHdlc 123
#define ifType_interleave 124
#define ifType_fast 125
#define ifType_ip 126
#define ifType_docsCableMaclayer 127
#define ifType_docsCableDownstream 128
#define ifType_docsCableUpstream 129
#define ifType_a12MppSwitch 130
#define ifType_tunnel 131
#define ifType_coffee 132
#define ifType_ces 133
#define ifType_atmSubInterface 134
#define ifType_l2vlan 135
#define ifType_l3ipvlan 136
#define ifType_l3ipxvlan 137
#define ifType_digitalPowerline 138
#define ifType_mediaMailOverIp 139
#define ifType_dtm 140
#define ifType_dcn 141
#define ifType_ipForward 142
#define ifType_msdsl 143
#define ifType_ieee1394 144
#define ifType_if_gsn 145
#define ifType_dvbRccMacLayer 146
#define ifType_dvbRccDownstream 147
#define ifType_dvbRccUpstream 148
#define ifType_atmVirtual 149
#define ifType_mplsTunnel 150
#define ifType_srp 151
#define ifType_voiceOverAtm 152
#define ifType_voiceOverFrameRelay 153
#define ifType_idsl 154
#define ifType_compositeLink 155
#define ifType_ss7SigLink 156
#define ifType_propWirelessP2P 157
#define ifType_frForward 158
#define ifType_rfc1483 159
#define ifType_usb 160
#define ifType_ieee8023adLag 161
#define ifType_bgppolicyaccounting 162
#define ifType_frf16MfrBundle 163
#define ifType_h323Gatekeeper 164
#define ifType_h323Proxy 165
#define ifType_mpls 166
#define ifType_mfSigLink 167
#define ifType_hdsl2 168
#define ifType_shdsl 169
#define ifType_ds1FDL 170
#define ifType_pos 171
#define ifType_dvbAsiIn 172
#define ifType_dvbAsiOut 173
#define ifType_plc 174
#define ifType_nfas 175
#define ifType_tr008 176
#define ifType_gr303RDT 177
#define ifType_gr303IDT 178
#define ifType_isup 179
#define ifType_propDocsWirelessMaclayer 180
#define ifType_propDocsWirelessDownstream 181
#define ifType_propDocsWirelessUpstream 182
#define ifType_hiperlan2 183
#define ifType_propBWAp2Mp 184
#define ifType_sonetOverheadChannel 185
#define ifType_digitalWrapperOverheadChannel 186
#define ifType_aal2 187
#define ifType_radioMAC 188
#define ifType_atmRadio 189
#define ifType_imt 190
#define ifType_mvl 191
#define ifType_reachDSL 192
#define ifType_frDlciEndPt 193
#define ifType_atmVciEndPt 194
#define ifType_opticalChannel 195
#define ifType_opticalTransport 196
#define ifType_propAtm 197
#define ifType_voiceOverCable 198
#define ifType_infiniband 199
#define ifType_teLink 200
#define ifType_q2931 201
#define ifType_virtualTg 202
#define ifType_sipTg 203
#define ifType_sipSig 204
#define ifType_docsCableUpstreamChannel 205
#define ifType_econet 206
#define ifType_pon155 207
#define ifType_pon622 208
#define ifType_bridge 209
#define ifType_linegroup 210
#define ifType_voiceEMFGD 211
#define ifType_voiceFGDEANA 212
#define ifType_voiceDID 213
#define ifType_mpegTransport 214
#define ifType_sixToFour 215
#define ifType_gtp 216
#define ifType_pdnEtherLoop1 217
#define ifType_pdnEtherLoop2 218
#define ifType_opticalChannelGroup 219

/* length of SNMP variable */
#define	SV_LEN(v)		(sizeof(v) / sizeof(oid))

typedef	struct snmp_mib_entry {
	const char *name;	/* name of snmp variable */
	const oid *objid;	/* object ID of snmp variable */
	int objid_len;		/* length of object ID (in OIDs) */
	int mode;		/* set of modes, bit-mapped */
#define	MIB_GET_MODES		000007
#define	 GET_STATE		000001
#define  GET_INFO		000002
#define  GET_DATA		000004
#define	MIB_SCAN_MODES		000070
#define	 SCAN_MATCH		000010
#define	 SCAN_DETECT		000020
#define	MIB_SNMP_VERSION	000700
#define	 MIB_SNMP_V1		000100 /* if not set - any version */
#define	 MIB_SNMP_V2		000200

} SNMP_MIB_ENTRY;

extern const SNMP_MIB_ENTRY system_MIB[];
extern const SNMP_MIB_ENTRY interface_MIB[];
extern const SNMP_MIB_ENTRY ipaddr_MIB[];
extern const SNMP_MIB_ENTRY bgp_MIB[];
extern const SNMP_MIB_ENTRY cisco_system_MIB[];
extern const SNMP_MIB_ENTRY cisco_interface_MIB[];
extern const SNMP_MIB_ENTRY cisco_envtemp_MIB[];
extern const SNMP_MIB_ENTRY cisco_envvolt_MIB[];
extern const SNMP_MIB_ENTRY cisco_envfan_MIB[];
extern const SNMP_MIB_ENTRY cisco_envps_MIB[];

void dump_mib_table __P((const SNMP_MIB_ENTRY *table));
const SNMP_MIB_ENTRY *get_mib_entry __P((const char *name,
					 const SNMP_MIB_ENTRY *table));
const char *get_mib_name __P((const oid *objid, int *objid_len,
			      const SNMP_MIB_ENTRY *table));
int get_mib_mode __P((const char *name, const SNMP_MIB_ENTRY *table));
void add_mib_table __P((VARIABLE **to, const SNMP_MIB_ENTRY *from, int ver));
void free_mib_table __P((VARIABLE **to, const SNMP_MIB_ENTRY *from));

#endif	/* !__MIB_H_ */
