/*
 *	Copyright (c) 1999-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifndef	__NETMON_H_
#define	__NETMON_H_

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#ifdef	DEBUG_MEMORY
#include <assert.h>
#endif

#ifndef	__P
#ifdef	__STDC__
#define	__P(proto)	proto
#else
#define	__P(proto)	()
#endif
#endif

#if defined(htons) && defined(_BIG_ENDIAN) && defined(sparc)
#undef	htons
#define	htons(x)	((uint16_t)(x))
#endif
#if defined(ntohs) && defined(_BIG_ENDIAN) && defined(sparc)
#undef	ntohs
#define	ntohs(x)	((uint16_t)(x))
#endif

#ifdef	DEBUG
#ifdef	HAVE_PTHREAD
#define	dprintf(x)	\
	{\
		char Dbuf[50];\
		struct timeval Dtv;\
		gettimeofday(&Dtv, 0);\
		strftime(Dbuf, sizeof(Dbuf), "%T", localtime(&Dtv.tv_sec));\
		printf("%s.%03d(%d): ", Dbuf, Dtv.tv_usec / 1000, (int)pthread_self());\
		printf x;\
		fflush(stdout);\
	}
#else
#define	dprintf(x)	\
	{\
		char Dbuf[50];\
		struct timeval Dtv;\
		gettimeofday(&Dtv, 0);\
		strftime(Dbuf, sizeof(Dbuf), "%T", localtime(&Dtv.tv_sec));\
		printf("%s.%03d: ", Dbuf, Dtv.tv_usec / 1000);\
		printf x;\
		fflush(stdout);\
	}
#endif
#ifdef	TRACE
#define	trace(x)	{ printf x; }
#else
#define	trace(x)
#endif
#else
#define	dprintf(x)
#define	trace(x)
#endif

#ifndef	FALSE
#define	FALSE	0
#endif
#ifndef	TRUE
#define	TRUE	1
#endif

#define	NETMON			"netmon"
#define	DEFAULT_CONFIG		"/etc/netmon.conf"
#define	DEFAULT_WATCHDOG	600	/* 10 min */

#define	POLLING_MIN		30	/* 30 sec */
#define	TIMEOUT_DEFAULT		2	/* 2 sec */
#define	RETRIES_DEFAULT		3

typedef void (*sigfunc) ();
typedef	struct timeval TIMEVAL;
typedef	unsigned long oid;

#define	STATE_UP	1
#define	STATE_DOWN	2
#define	BGP_ESTABLISHED	6
#define	ENV_NOTPRESENT	5

#define	OBJ_STATE	0
#define	IF_STATE	1
#define	BGP_STATE	2
#define	ENV_STATE	3

/* object types */
#define	TYPE_ANY		0
#define	TYPE_OBJECT		1
#define	TYPE_SERVICE		2
#define	TYPE_INTERFACE		3
#define	TYPE_BGPAS		4
#define	TYPE_BGPPEER		5
#define	TYPE_ENVMON		6
#define	TYPE_ENVTEMP		7
#define	TYPE_ENVVOLT		8
#define	TYPE_ENVFAN		9
#define	TYPE_ENVPS		10

struct object_ent;
struct method_ent;

/*
 * Built-in variables
 */
typedef struct var_table_ent {
	char *name;		/* name of variable */
	int old;		/* curr or old value (flag) */
	char *(*func)();	/* value printing function */
} VAR_TABLE;

/*
 * User-defined variables
 */
typedef	struct statistic_ent {
	int count;		/* number of probes */
	u_int64_t sum, max, min, delta;
} STATISTIC;

typedef	struct variable_ent {
	char *name;		/* variable name (without leading '$') */
	oid *objid;		/* object identifier for SNMP variable */
	int objid_len;		/* number of subid's in the objid */
	oid *instance;		/* in/out instance buffer */
	int instance_len;	/* number of subid's in the instance */
	int mode;		/* set of modes; bit-mapped (see mib.h) */
	enum {
		INTEGER = 0,
		UNSIGNED,
		UNSIGNED64,
		IPADDRESS,
		FLOATING,
		STRING,
		OBJID
	} type;			/* type of variable */
	union {
		u_int32_t *integer;
		u_int64_t *integer64;
		double *floating;
		char *string;
		oid *objid;
	} value;		/* value of variable */
	int length;		/* length of value in bytes */

#ifdef	USE_VAR_MTIME
	TIMEVAL	mtime;		/* value modification time */
#endif
	STATISTIC *statistic;	/* statistic for save */
	u_int64_t delta;	/* need recover after counter overload */

	struct variable_ent *old;/* previous stamp of the variable */

	struct variable_ent *next;
} VARIABLE;

typedef	struct reference_ent {
	char *name;		/* name of reference */
	VARIABLE *var;		/* referenced variable */
	struct reference_ent *next;
} REFERENCE;

/*
 * Group stuff
 */
typedef	struct group_entry {
	int permit;		/* 1=permit, 0=deny */
	char *pattern;		/* regex pattern mask */

	struct group_entry *next;
} GROUP_ENTRY;

typedef	struct group_list {
	char *name;		/* name of the list */
	GROUP_ENTRY *entry;

	struct group_list *next;
} GROUP_LIST;

typedef	struct group_ref {
	char *name;		/* name of reference */
	struct group_ref *next;
} GROUP_REF;


/******************
 * Output methods *
 ******************/

typedef	struct save_ent {
	char *name;		/* save name */
	int method;		/* save method */
#define	SAVE_FILE	1
#define	SAVE_EXEC	2
#define	SAVE_PIPE	3
	char *file;		/* file name to write, execute or pipe */

	char *saving;		/* saving output format */
	char *state;		/* change state output format */
	char *when;		/* condition string */
	int delay;		/* condition delay in seconds */
	char *when_fmt;		/* when condition true output format */

	/* current owner */
	int type;
	void *object;

	struct save_ent *next;
} SAVE;


/***********************
 * SNMP specific stuff *
 ***********************/

/* IP address of interface */
typedef	struct ifaddr_ent {
	struct in_addr ip_addr;	/* address */

	struct ifaddr_ent *next;
} IFADDR;

/* interface descriptor */
typedef	struct interface_ent {
#ifdef	DEBUG_MEMORY
	int cksum;
#endif
	/* configuration data */
	char *name;		/* interface name or         */
	int index;		/*           interface index */
	char *descr;		/* interface description */
	char *datadir;		/* directory where store data */
	long bandwidth;		/* predefined interface bandwidth (bps) */
	VARIABLE *var_list;	/* list of variables */
	SAVE *save_list;	/* list of save methods */
	GROUP_REF *ns_acl;	/* netstate client access list */

	int state;		/* current operational status (UP/DOWN/...) */
	int prev_state;		/* previous operational status */
	TIMEVAL last_request;	/* last time interface requested */
	TIMEVAL prev_request;	/* previous time interface requested */
	TIMEVAL last_reply;	/* last time interface reply */
	TIMEVAL prev_reply;	/* previous time interface reply */
	TIMEVAL	last_change;	/* last time status change */

	int	mode;		/* SNMP current request mode (see mib.h) */
	int	reqid;		/* SNMP current request id */
	u_long	tabIdx;		/* SNMP ifTable index */
	IFADDR *ifaddr;		/* list of all interface IP addresses */

	struct object_ent *parent;
	struct interface_ent *next;
} INTERFACE;

struct bgp_as_ent;

/* bgp peer descriptor */
typedef	struct bgp_peer_ent {
	struct in_addr ip_addr;	/* IP address of peer */
	VARIABLE *var_list;	/* list of variables */

	int state;		/* current operational status (UP/DOWN/...) */
	int prev_state;		/* previous operational status */
	TIMEVAL last_request;	/* last time bgp peer requested */
	TIMEVAL prev_request;	/* previous time bgp peer requested */
	TIMEVAL last_reply;	/* last time bgp peer reply */
	TIMEVAL prev_reply;	/* previous time bgp peer reply */
	TIMEVAL	last_change;	/* last time status change */

	int	mode;		/* SNMP current request mode (see mib.h) */
	int	reqid;		/* SNMP current request id */
	u_long	tabIdx;		/* SNMP bgpPeerTable index */

	struct bgp_as_ent *parent;
	struct bgp_peer_ent *next;
} BGP_PEER;

/* bgp descriptor */
typedef	struct bgp_as_ent {
	/* configuration data */
	int asn;		/* AS number */
	char *descr;		/* AS description */
	char *datadir;		/* directory where store data */
	SAVE *save_list;	/* list of save methods */
	GROUP_REF *ns_acl;	/* netstate client access list */

	int handconf;		/* list of peers was configured by hand */
	BGP_PEER *peer;		/* list of bgp peers */

	struct object_ent *parent;
	struct bgp_as_ent *next;
} BGP_AS;

struct env_mon_ent;

/* Environment temperature descriptor */
typedef	struct env_gauge_ent {
	char *name;		/* description of envtemp gauge */
	u_long tabIdx;		/* SNMP ciscoEnvMonTable index */
	VARIABLE *var_list;	/* list of variables */

	int state;		/* current operational status (NORMAL...) */
	int prev_state;		/* previous operational status */

	struct env_mon_ent *parent;
	struct env_gauge_ent *next;
} ENV_GAUGE;

/* Environment monitor descriptor */
typedef	struct env_mon_ent {
	/* configuration data */
	int type;		/* type of gauge: TYPE_ENVTEMP, etc */
	char *descr;		/* env mon description */
	char *datadir;		/* directory where store data */
	SAVE *save_list;	/* list of save methods */
	GROUP_REF *ns_acl;	/* netstate client access list */

	int handconf;		/* list of gauges was configured by hand */
	ENV_GAUGE *gauge;	/* list of gauges */

	TIMEVAL last_request;	/* last time envmon gauge requested */
	TIMEVAL prev_request;	/* previous time envmon gauge requested */
	TIMEVAL last_reply;	/* last time envmon gauge reply */
	TIMEVAL prev_reply;	/* previous time envmon gauge reply */
	TIMEVAL	last_change;	/* last time status change */

	int	mode;		/* SNMP current request mode (see mib.h) */
	int	reqid;		/* SNMP current request id */

	struct object_ent *parent;
	struct env_mon_ent *next;
} ENV_MON;


/**********************
 * Monitoring methods *
 **********************/

/* Chat-Script magic cookie */
/* NOTE: private allocation for each object required */
typedef	struct chatscript_ent {
	char *send;		/* data to send */
	char *expect;		/* expected string */
	int matched;		/* boolean -- expect matched */
	REFERENCE *var_ref;	/* reference to extracted variables */

	struct chatscript_ent *next;
} CHATSCRIPT;

/* SNMP module part */
typedef	struct snmp_request_ent {
	int command;		/* SNMP command (i.e. type of request) */
	REFERENCE *var_ref;	/* reference to variables */
	int error_status;	/* last error status of response */
	int error_index;	/* last error index of response */
} SNMP_REQ;

/* handle multiply requests in an udp session */
typedef	struct request_ent {
	SNMP_REQ *snmpreq;	/* SNMP part of session request */
	int reqid;		/* request id */
	int sequence;		/* retry counter, while not zero -- busy */
	TIMEVAL expire;		/* time until first timeout */

	struct request_ent *next;
} REQUEST;

/* Asynchronous session task */
typedef	struct session_ent {
	/* this stuff must be initialized outside */
	struct object_ent *owner;	/* session owner */
	struct method_ent *method;	/* session method */
	int sock;		/* socket file descriptor */
	struct sockaddr peer;	/* address of peer */
	long timeout;		/* number of microseconds until first timeout */
	int retries;		/* number of retries before timeout */
	int (*connect) __P((struct session_ent *));
	int (*send) __P((struct session_ent *, REQUEST *));
	int (*recv) __P((struct session_ent *));
	void (*read) __P((int, struct session_ent *, int));

	/* returned values */
	int data_int;		/* data length or chat-script matchs */
	char *data_ptr;		/* pointer to resulting data if any */

	/* session internal data */
	CHATSCRIPT *cs;		/* current pair in chat-script */
	char *buf;		/* temporary buffer for a session */
	TIMEVAL tv_sum;
	int pkt_send;
	int pkt_recv;
	REQUEST *request;

	struct session_ent *next;
} SESSION;

/* Descriptor of monitoring method */
/* NOTE: private allocation for each object required */
typedef	struct method_ent {
	/* initialize method session on the object */
	int (*init) __P((struct object_ent *, struct method_ent *));

	/* start method session */
	void (*start) __P((struct method_ent *));

	/* stop method session */
	void (*stop) __P((struct method_ent *));

	/* configuration data */
	char *name;		/* method name */
	char *argument;		/* method argument (for ex. SNMP community) */
	char *address;		/* domain name or dotted IP address */
	struct in_addr ip_addr;	/* ip address of peer */
	int protocol;		/* method protocol */
	u_short	rport;		/* remote port number, 0=unused */
	int timeout;		/* number of seconds until first timeout */
	int retries;		/* number of retries before timeout */
	union {
		struct ping_param {
			short send;	/* ICMP echo request packets to send */
			short expect;	/* ICMP echo reply packets to expect */
		} ping_param;
		struct snmp_param {
			short version;	/* SNMP protocol version number */
			short batch;	/* ROUTER batch parameter */
		} snmp_param;
		struct bind_param {
			u_short lport;	/* lowest local port TCP/UDP binding */
			u_short hport;	/* highest local port TCP/UDP binding*/
		} bind_param;
		int reserved;	/* unused, just for padding */
	} method_param;
#define	ping_send	method_param.ping_param.send
#define	ping_expect	method_param.ping_param.expect
#define	snmp_version	method_param.snmp_param.version
#define	router_batch	method_param.snmp_param.batch
#define	lport_min	method_param.bind_param.lport
#define	lport_max	method_param.bind_param.hport

	/* method data */
	SESSION	*sd;		/* session descriptor */
	CHATSCRIPT *chatscript;	/* method chat-script */
	SNMP_REQ *snmpreq;	/* method snmp-request */

	struct method_ent *next;
} METHOD;

/* Traping method descriptor */
/* NOTE: private allocation for each object required */
typedef	struct trap_ent {
	char *name;		/* name of traping method */
	char *community;	/* community string */
	oid *objid;		/* enterprise object identifier */
	int objid_len;		/* number of subid's in the enterprise */
	int type;		/* generic trap type */
	int specific;		/* specific trap code */
	long uptime;		/* time stamp in 100*seconds */

	struct object_ent *owner;	/* trap owner */
	VARIABLE *var_list;	/* list of variables */

	struct trap_ent *next;
} TRAP;

typedef	struct nexthop_ent {
	int hop_count;
	struct object_ent *target;
	struct interface_ent *interface;

	struct nexthop_ent *prev;
	struct nexthop_ent *next;
} NEXTHOP;

/* SNMP magic cookie */
typedef	struct snmp_ent {
	int scan;		/* SNMP current scan mode; bit-mapped: */
	int pending_scan;
	int recover_scan;
#define	SCAN_IFTABLE	0x01	/* ifTable */
#define	SCAN_IFADDR	0x02	/* ipAddrTable */
#define	SCAN_BGPPEER	0x04	/* bgpPeerTable */
#define	SCAN_ENVMON	0x08	/* ciscoEnvMonTable */

	int mode;		/* SNMP current request mode (see mib.h) */
	int reqid;		/* SNMP current request id */
	int device;		/* device id */
#define	DEVICE_UNKNOWN		0
#define	DEVICE_CISCO		1
#define	DEVICE_CATALYST		2

	/* interface requests package (SNMP packets per time) */
	int batch_count;
	INTERFACE *next_batch;

	/* SCAN_IFTABLE */
	int ifPhase;		/* 0-match , 1-detect */
	VARIABLE *ifTable;

	/* SCAN_IFADDR */
	VARIABLE *ipTable;

	/* SCAN_BGPPEER */
	int bgpPhase;		/* 0-match , 1-detect */
	VARIABLE *bgpTable;

	/* SCAN_ENVMON */
	int envPhase;		/* 0-match , 1-detect */
	VARIABLE *envTable;
} SNMP_DATA;

typedef	struct object_ent {
#ifdef	DEBUG_MEMORY
	int cksum;
#endif
	/* configuration data */
	char *name;		/* object name */
	char *descr;		/* object description */
	char *datadir;		/* directory where store data */
	char *address;		/* domain name or dotted IP address */
	struct in_addr ip_addr;	/* ip address of peer */
	int polling;		/* polling period in seconds */
	int saving;		/* saving period in seconds */
	int sync;		/* polling counter to synchronize saving */
	int timeout;		/* default object timeout in seconds */
	int retries;		/* default object number of retries */
	TRAP *trap_list;	/* list of traping methods */
	METHOD *method_list;	/* list of monitoring methods */
	VARIABLE *var_list;	/* list of variables */
	SAVE *save_list;	/* list of save methods */
	GROUP_REF *ns_acl;	/* netstate client access list */

	int state;		/* current operational status (UP/DOWN/...) */
	int prev_state;		/* previous operational status */
	TIMEVAL last_request;	/* last time method requested */
	TIMEVAL prev_request;	/* previous time method requested */
	TIMEVAL last_reply;	/* last time method reply */
	TIMEVAL prev_reply;	/* previous time method reply */
	TIMEVAL	last_change;	/* last time status change */

	/* discovery stuff */
	int	hop_count;	/* hop count */
	NEXTHOP	*nexthop;	/* next hop */
	IFADDR	*ifaddr;	/* list of router IP addresses */

	SNMP_DATA *snmpdata;	/* SNMP data for SNMP ROUTER method */

	INTERFACE *interface;		/* interface list */
	INTERFACE *ifgroup;		/* interface group prototype */
	BGP_AS *bgp;			/* bgp list */
	ENV_MON *env;			/* list of environment monitors */
	struct object_ent *service;	/* service list */

	struct object_ent *parent;	/* parent object or nil */
	struct object_ent *next;	/* next object in this chain */
} OBJECT;


/********************
 * Main config file *
 ********************/

typedef	struct config_ent {
	char *rootdir;		/* default work directory */
	char *timefmt;		/* strftime format of currtime for logging */
	int polling;		/* default polling interval in seconds */
	int saving;		/* default saving interval in seconds */
	int timeout;		/* default timeout in seconds */
	int retries;		/* default number of retries */

	int enable_traps;	/* enable SNMP traps */
	int source_traps;	/* match src-addr and agent-addr of traps */

	/* netstate server */
	int ns_port;		/* server port number */
	int ns_timo;		/* client timeout in seconds */
	GROUP_REF *ns_acl;	/* netstate client access list */

	GROUP_LIST *group_list;	/* list of configured groups */

	OBJECT *target;		/* list of configured target objects */
} CONFIG;


/*
 * Event scheduler.
 */

typedef	struct event_ent {
	TIMEVAL tv;		/* system time in [micro]seconds from UTC */
	void (*func) __P((void *));	/* function to call at the time */
	void *arg;		/* function argument pointer */

	struct event_ent *next;
} EVENT;


/*
 * Global varibales
 */
extern char package[], version[], target[], compiled[];
extern char *program_name, *config_file, *pid_file;
extern int trap_sock;
extern int netstate_sock;
extern int Cflag;
extern int syslog_initialized;

/*
 * Function prototypes
 */
/* netmond.c */
int set_timer __P((int, void (*f)()));
void object_init __P((OBJECT *));
void object_stop __P((OBJECT *));
int object_busy __P((OBJECT *));
void start_object_list __P((OBJECT *));
void start_method_list __P((void *));	/* OBJECT * */
void method_finished __P((OBJECT *, METHOD *, char *, int));
void execfile __P((int, char *, char *));

/* netstate.c */
int netstate_init __P((int));
void *netstate_serve __P((void *));
int get_type_by_name __P((char *));
char *get_name_by_type __P((int));

/* event.c */
void tv_sub __P((TIMEVAL *, TIMEVAL *));
void tv_add __P((TIMEVAL *, TIMEVAL *));
u_long tv_round __P((TIMEVAL *));
int select_event __P((TIMEVAL *));
void add_event __P((TIMEVAL *, void (*f)(void *), void *));
int remove_event __P((void (*f)(void *), void *));
int change_event __P((void (*f)(void *), void *, void *));
#define	find_event(f, a)	change_event(f, a, a)
void free_events __P((void));
int object_offset __P((OBJECT *, int));
void tv_sharp __P((TIMEVAL *, int, int));

/* util.c */
void report __P((int, const char *, ...));
char *strippath __P((char *));
char *stripquotes __P((char *));
int match_ifdescr __P((const char *name, const char *descr));
char *oid2str __P((char *, const oid *, int));
void str2oid __P((oid *, int *, char *));
char *var2str __P((char *, VARIABLE *));
void str2var __P((VARIABLE *, char *));
void free_var __P((VARIABLE *));
void free_var_list __P((VARIABLE *));
VARIABLE *dup_var __P((const VARIABLE *));
REFERENCE *dup_var_ref __P((REFERENCE *));
SNMP_REQ *dup_snmp_request __P((SNMP_REQ *));
void free_var_ref __P((REFERENCE *));
void free_chatscript __P((CHATSCRIPT *));
void free_save __P((SAVE *));
void free_save_list __P((SAVE *));
SAVE *dup_save __P((const SAVE *));
void free_method __P((METHOD *));
void free_method_list __P((METHOD *));
void free_trap_list __P((TRAP *));
void free_group_list __P((GROUP_LIST **chain, const char *name));
void free_group_ref __P((GROUP_REF **chain));
GROUP_REF *add_group_ref_copy __P((GROUP_REF **chain, const GROUP_REF *from));
int check_group __P((char *ref, char *str, char *altstr));
int check_group_ref __P((GROUP_REF *ref, char *str, char *altstr));
void set_var __P((VARIABLE *, int, void *, int));
void set_objid __P((VARIABLE *, const oid *, int));
void set_instance __P((VARIABLE *, const oid *, int));
VARIABLE *get_var __P((VARIABLE *, const char *));
VARIABLE *add_var __P((VARIABLE **, const char *));
VARIABLE *make_var __P((const char *name, const oid *objid, int objid_len, int mode));
VARIABLE *add_var_copy __P((VARIABLE **to, const VARIABLE *from));
SAVE *add_save_copy __P((SAVE **to, const SAVE *from));
int is_var_changed __P((VARIABLE *, const char *));
void dump_config __P((CONFIG *));
int make_argv __P((char *, char ***, int));
char *my_strsep __P((char **, char *));
u_short in_cksum __P((u_short *, int));
char *intoa __P((char *buf, struct in_addr in));
char *itoa __P((int));
#ifdef	TRACE
void dump_var __P((VARIABLE *));
void dump_var_list __P((VARIABLE *));
#else
#define	dump_var(v)
#define	dump_var_list(v)
#endif
#ifdef	DEBUG_MEMORY
void set_cksum __P((void));
int get_cksum __P((void));
#endif

/* parseconf.y */
void *parse_config __P((void *));

/* calc.y */
double calculate __P((char *));

/* session.c */
SESSION *session_create __P((SESSION *));
int session_offset __P((SESSION *, int));
SESSION *session_busy __P((SESSION *));
void session_free __P((SESSION *));
int session_start __P((SESSION *, SNMP_REQ *));
void session_stop __P((SESSION *, int));
int session_select __P((int *, fd_set *, fd_set *, TIMEVAL *, int *));
REQUEST *get_request __P((SESSION *, int));
void session_operate __P((fd_set *, fd_set *));
void session_timeout __P((void));
int get_socket_error __P((int s));
int set_socket_async __P((int s, int on));

/* ping.c */
void make_netpath __P((OBJECT *, in_addr_t *, int));
void free_netpath __P((NEXTHOP *));
int check_netpath __P((OBJECT *));

/* tcp.c */
int tcp_init __P((OBJECT *, METHOD *));
void tcp_start __P((METHOD *));
void tcp_stop __P((METHOD *));
int match_expect __P((SESSION *, CHATSCRIPT *, char *));

/* udp.c */
int udp_init __P((OBJECT *, METHOD *));
void udp_start __P((METHOD *));
void udp_stop __P((METHOD *));

/* ping.c */
int echo_init __P((OBJECT *, METHOD *));
void echo_start __P((METHOD *));
void echo_stop __P((METHOD *));

/* snmp.c */
u_long oid2long __P((const oid *from, int len));
oid *long2oid __P((oid *to, int len, u_long from));
int snmp_send __P((SESSION *, REQUEST *));
int snmp_recv __P((SESSION *));
TRAP *snmp_trap __P((int, int));

/* trap.c */
int trap_init __P((int));
OBJECT *trap_auth __P((struct in_addr *, char *, oid *, int, int, int));
void trap_parse __P((TRAP *));
void trap_free __P((TRAP *));

/* router.c */
void free_ifaddr __P((IFADDR **));
void free_bgp_peers __P((BGP_AS *, int));
void free_env_gauges __P((ENV_MON *, int));

/* variables.c */
char *state2str __P((int, int));
char *get_bgppeer_name __P((char *, struct in_addr in));
char *insert_variables __P((char *, int, char *, int, void *));

/* save.c */
const char *get_envstate_name __P((int type)); /* XXX fake */
void reset_statistic __P((VARIABLE *, int));
void set_none_state __P((OBJECT *));
void save_state __P((SAVE *, int, void *));
void save_when __P((SAVE *, int, void *));
void save_object_data __P((void *));	/* OBJECT * */

/* reconfig.c */
void soft_reconfig __P((CONFIG *));
void hard_reconfig __P((CONFIG *));

#endif	/* !__NETMON_H_ */
