/*
 *	Copyright (c) 1999-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

/*
 * For Linux lamers:
 * Force to use traditional `struct ip' and `struct icmp' instead of
 * Linux shit named by someone as `struct iphdr' and `struct icmphdr'.
 */
#ifndef	__USE_BSD
#define	__USE_BSD
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <time.h>
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"

/*
 * Define it if you dont want to receive ICMP errors on the RAW socket.
 */
/*#define NO_ICMP_ERRORS */

extern int errno;

#ifndef	IP_HDRINCL
#warning the IP_HDRINCL is unsupported on this system
#define	IP_HDRINCL	2
#endif

/* IP packet Type of service */
#ifndef	IPTOS_LOWDELAY
#define IPTOS_LOWDELAY			0x10
#endif
#ifndef	IPTOS_PREC_INTERNETCONTROL
#define IPTOS_PREC_INTERNETCONTROL	0xc0
#endif

/* IP Dont Fragment flag */
#ifndef	IP_DF
#define	IP_DF		0x4000
#endif

/* IP packet Time to live */
#ifndef	IPDEFTTL
#define	IPDEFTTL	64
#endif

/* IP packet Options */
#ifndef	MAX_IPOPTLEN
#define	MAX_IPOPTLEN	40	/* mandatory */
#endif
#ifndef	IPOPT_OPTVAL
#define IPOPT_OPTVAL	0	/* option ID */
#endif
#ifndef	IPOPT_OLEN
#define IPOPT_OLEN	1	/* option length */
#endif
#ifndef	IPOPT_OFFSET
#define IPOPT_OFFSET	2	/* offset within option */
#endif
#ifndef	IPOPT_MINOFF
#define IPOPT_MINOFF	4	/* min value of above */
#endif
#ifndef	IPOPT_RR
#define IPOPT_RR	7	/* record packet route */
#endif

/* ICMP packet minimum size */
#ifndef	ICMP_MINLEN
#define	ICMP_MINLEN	8
#endif

/* ICMP packet types */
#ifndef	ICMP_UNREACH
#define	ICMP_UNREACH	3
#endif
#ifndef	ICMP_SOURCEQUENCH
#define	ICMP_SOURCEQUENCH	4
#endif
#ifndef	ICMP_REDIRECT
#define	ICMP_REDIRECT	5
#endif
#ifndef	ICMP_TIMXCEED
#define	ICMP_TIMXCEED	11
#endif
#ifndef	ICMP_PARAMPROB
#define	ICMP_PARAMPROB	12
#endif

#define	MIN_PACKETSZ	(sizeof(struct ip) + ICMP_MINLEN + sizeof(TIMEVAL *))
#define	MAX_PACKETSZ	4096	/* max reasonable size in bytes */

static char *icmp_error __P((int));
static void echo_reply __P((int, SESSION *, int));

#define	ICMP_UNREACH_CODES	16
static char *unreach_error[ICMP_UNREACH_CODES] = {
	"Network unreachable",		/* ICMP_UNREACH_NET */
	"Host unreachable",		/* ICMP_UNREACH_HOST */
	"Protocol unreachable",		/* ICMP_UNREACH_PROTOCOL */
	"Port unreachable",		/* ICMP_UNREACH_PORT */
	"Fragmentation needed but DF set",/* ICMP_UNREACH_NEEDFRAG */
	"Source route failed",		/* ICMP_UNREACH_SRCFAIL */
	"Network is unknown",		/* ICMP_UNREACH_NET_UNKNOWN */
	"Host is unknown",		/* ICMP_UNREACH_HOST_UNKNOWN */
	"Source host isolated",		/* ICMP_UNREACH_ISOLATED */
	"Network access prohibited",	/* ICMP_UNREACH_NET_PROHIB */
	"Host access prohibited",	/* ICMP_UNREACH_HOST_PROHIB */
	"Bad TOS for network",		/* ICMP_UNREACH_TOSNET */
	"Bad TOS for host",		/* ICMP_UNREACH_TOSHOST */
	"Packet filtered",		/* ICMP_UNREACH_FILTER_PROHIB */
	"Host precedence violation",	/* ICMP_UNREACH_HOST_PRECEDENCE */
	"Precedence cut off",		/* ICMP_UNREACH_PRECEDENCE_CUTOFF */
};

#define	ICMP_REDIRECT_CODES	4
static char *redirect_error[ICMP_REDIRECT_CODES] = {
	"Redirect network",		/* ICMP_REDIRECT_NET */
	"Redirect host",		/* ICMP_REDIRECT_HOST */
	"Redirect network for TOS",	/* ICMP_REDIRECT_TOSNET */
	"Redirect host for TOS",	/* ICMP_REDIRECT_TOSHOST */
};

#define	ICMP_TIMXCEED_CODES	2
static char *timxceed_error[ICMP_TIMXCEED_CODES] = {
	"TTL exceeded in transit",	/* ICMP_TIMXCEED_INTRANS */
	"TTL exceeded in reassemble",	/* ICMP_TIMXCEED_REASS */
};

#define	ICMP_PARAMPROB_CODES	3
static char *paramprob_error[ICMP_PARAMPROB_CODES] = {
	"Bad IP header",		/* 0 */
	"Required IP option absent",	/* ICMP_PARAMPROB_OPTABSENT */
	"Bad length",			/* ICMP_PARAMPROB_BADLENGTH */
};

static char *
icmp_error(err)
	int err;
{
	u_char code = err & 0xff;
	u_char type = (err >> 8) & 0xff;
	if (type == ICMP_UNREACH && code < ICMP_UNREACH_CODES)
		return unreach_error[code];
	if (type == ICMP_SOURCEQUENCH)
		return "Packet lost, slow down";
	if (type == ICMP_REDIRECT && code < ICMP_REDIRECT_CODES)
		return redirect_error[code];
	if (type == ICMP_TIMXCEED && code < ICMP_TIMXCEED_CODES)
		return timxceed_error[code];
	if (type == ICMP_PARAMPROB && code < ICMP_PARAMPROB_CODES)
		return paramprob_error[code];
	return "Unknown error";
}

#ifdef	TRACE
void
dump_ip(ip)
	register struct ip *ip;
{
	char buf[100];
	printf("\t IP: hlen=%d tos=0x%x tlen=%d id=%d off=0x%x ttl=%d proto=%d cksum=0x%x\n",
	       (int)(ip->ip_hl << 2),
	       (int)ip->ip_tos,
	       (int)ip->ip_len,
	       (int)ip->ip_id,
	       (int)ip->ip_off,
	       (int)ip->ip_ttl,
	       (int)ip->ip_p,
	       (int)ip->ip_sum);
	printf("\t     src=%s", intoa(buf, ip->ip_src));
	printf(" dst=%s\n", intoa(buf, ip->ip_dst));
}

void
dump_icmp(icmp)
	register struct icmp *icmp;
{
	printf("\t ICMP: type=%d code=%d cksum=0x%x\n",
	       (int)icmp->icmp_type,
	       (int)icmp->icmp_code,
	       (int)icmp->icmp_cksum);
}
#endif

static int hop_count;

static NEXTHOP *
get_nexthop(ip_addr, addr, rest)
	struct in_addr *ip_addr;
	in_addr_t *addr;
	int rest;
{
	OBJECT *tp;
	INTERFACE *ifp;
	IFADDR *ifa;
	NEXTHOP *nexthop;
#ifdef	TRACE
	char ipaddr[20];
#endif
	extern CONFIG *cf;

	do {
		hop_count++;
		if (!memcmp(ip_addr, addr, sizeof(in_addr_t)))
			return NULL;

		for (tp = cf->target; tp; tp = tp->next) {
			/* search for interface dependency first */
			for (ifp = tp->interface; ifp; ifp = ifp->next) {
				for (ifa = ifp->ifaddr; ifa; ifa = ifa->next) {
					if (!memcmp(&ifa->ip_addr, addr, sizeof(in_addr_t))) {
						trace(("\t ifa=%s rest=%d\n",
						       intoa(ipaddr, ifa->ip_addr), rest));

						if ((nexthop = (NEXTHOP *)malloc(sizeof(NEXTHOP))) == NULL)
							return NULL;
						memset(nexthop, 0, sizeof(NEXTHOP));
						nexthop->hop_count = hop_count;
						nexthop->target = tp;
						nexthop->interface = ifp;
						if (rest - 1 > 0)
							nexthop->prev = get_nexthop(ip_addr, addr+1, rest-1);
						return nexthop;
					}
				}
			}
			for (ifa = tp->ifaddr; ifa; ifa = ifa->next) {
				if (!memcmp(&ifa->ip_addr, addr, sizeof(in_addr_t))) {
					trace(("\t ifa=%s rest=%d\n",
					       intoa(ipaddr, ifa->ip_addr), rest));

					if ((nexthop = (NEXTHOP *)malloc(sizeof(NEXTHOP))) == NULL)
						return NULL;
					memset(nexthop, 0, sizeof(NEXTHOP));
					nexthop->hop_count = hop_count;
					nexthop->target = tp;
					if (rest - 1 > 0)
						nexthop->prev = get_nexthop(ip_addr, addr+1, rest-1);
					return nexthop;
				}
			}
		}
		addr++;
	} while (--rest);

	return NULL;
}

void
free_netpath(nexthop)
	NEXTHOP *nexthop;
{
	NEXTHOP *next;

	while (nexthop) {
		next = nexthop->next;
		free(nexthop);
		nexthop = next;
	}
}

void
make_netpath(target, addr, len)
	OBJECT *target;
	in_addr_t *addr;
	int len;
{
	NEXTHOP *nexthop, *first;
#ifdef	DEBUG
	INTERFACE *iface;
	int i;
	struct in_addr ip_addr;
	char ipaddr[20];
#endif

	if (!target || !addr || len < 1) return;

	dprintf(("make_netpath(%s)...\n", target->name));

#ifdef	DEBUG
	for (i = 0; i < len; i++) {
		memcpy(&ip_addr, &addr[i], sizeof(in_addr_t));
		trace(("\t %s\n", intoa(ipaddr, ip_addr)));
	}
#endif
	/* free current chain */
	free_netpath(target->nexthop);

	/* make new chain */
	hop_count = 0;
	nexthop = get_nexthop(&target->ip_addr, addr, len);

	/* make reverse chain, back to top */
	first = NULL;
	while (nexthop) {
		if (nexthop->prev) {
			first = nexthop->prev;
			first->next = nexthop;
		} else	first = nexthop;
		nexthop = nexthop->prev;
	}
	target->hop_count = hop_count;
	target->nexthop = first;
	trace(("\t %d %s\n", hop_count, target->name));

#ifdef	DEBUG
	while (first) {
		target = first->target;
		iface = first->interface;
		trace(("\t %d %s/%s\n", first->hop_count,
		       target ? target->name : "NULL",
		       iface ? iface->name : "NULL"));
		first = first->next;
	}
#endif
}

int
check_netpath(target)
	OBJECT *target;
{
	NEXTHOP *nexthop;

	/* sanity check */
	if (!target) return 0;

	dprintf(("check_netpath(%s)\n", target->name));

	for (nexthop = target->nexthop; nexthop; nexthop = nexthop->next) {
		if ((nexthop->target && nexthop->target->state != STATE_UP) ||
		    (nexthop->interface && nexthop->interface->state != STATE_UP))
			return 0;
	}
	return 1;
}

static int
echo_send(sd, request)
	SESSION *sd;
	REQUEST *request;
{
	METHOD *method = sd->method;
	u_char buf[MAX_PACKETSZ];
	struct ip *ip;
	struct icmp *icmp;
	struct sockaddr_in *to = (struct sockaddr_in *)&sd->peer;
	int header_len = sizeof(struct ip);
	int total_len = method->rport ? method->rport : MIN_PACKETSZ;
	int rr_opt = (!method->rport && request->sequence == 1);

	dprintf(("echo_send: entered\n"));

	if (rr_opt) {
		header_len += MAX_IPOPTLEN;
		total_len += MAX_IPOPTLEN;
	}
	if (total_len < MIN_PACKETSZ)
		total_len = MIN_PACKETSZ;
	else if (total_len > MAX_PACKETSZ)
		total_len = MAX_PACKETSZ;
	memset(buf, 0, total_len);

	/* IP header draft */
	ip = (struct ip *)buf;
	ip->ip_v = IPVERSION;
	ip->ip_hl = header_len >> 2; /* make 32-bit words */
	if (!method->rport) /* set TOS only for built-in method */
		ip->ip_tos = IPTOS_PREC_INTERNETCONTROL | IPTOS_LOWDELAY;
#ifdef	linux
	ip->ip_len = htons(total_len);
	ip->ip_off = htons(IP_DF);
#else	/* BSD */
	ip->ip_len = total_len;
	ip->ip_off = IP_DF;
#endif
	ip->ip_ttl = IPDEFTTL;
	ip->ip_p = IPPROTO_ICMP;
	/* ip->ip_src <-- filled by kernel (hopefulness) */
	ip->ip_dst = to->sin_addr;

	if (rr_opt) { /* IP Option: Record Route */
		u_char *cp = buf + sizeof(struct ip);
		cp[IPOPT_OPTVAL] = IPOPT_RR;
		cp[IPOPT_OLEN] = MAX_IPOPTLEN - 1; /* 3 head + 36 addr data */
		cp[IPOPT_OFFSET] = IPOPT_MINOFF;
	}
	/* replaced by kernel (hopefulness) */
	ip->ip_sum = in_cksum((u_short *)ip, header_len);

	/* ICMP header */
	icmp = (struct icmp *)(buf + header_len);
	icmp->icmp_type = ICMP_ECHO;
	icmp->icmp_id = htons(request->reqid);
	icmp->icmp_seq = htons(request->sequence);

	/* address of timestamp buffer as ICMP data */
	gettimeofday((TIMEVAL *)sd->buf, NULL);
	memcpy(icmp->icmp_data, &sd->buf, sizeof(TIMEVAL *));

	icmp->icmp_cksum = in_cksum((u_short *)icmp, total_len - header_len);
#ifdef	NO_ICMP_ERRORS
	total_len = send(sd->sock, (char *)buf, total_len, 0);
#else
	total_len = sendto(sd->sock, (char *)buf, total_len, 0, &sd->peer, sizeof(struct sockaddr));
#endif
	return total_len;
}

static int
ident_icmp_packet(icmp, request)
	register struct icmp *icmp;
	REQUEST *request;
{
	int seq;

	/* check for ID */
	if (icmp->icmp_id != htons(request->reqid)) {
		trace(("\t ident_icmp_packet: bad icmp_id %d\n",
		       ntohs(icmp->icmp_id)));
		return 0;
	}

	/* check for SEQ */
	seq = ntohs(icmp->icmp_seq);
	if (seq < 1 || seq > request->sequence) {
		trace(("\t ident_icmp_packet: bad icmp_seq %d\n",
		       ntohs(icmp->icmp_seq)));
		return 0;
	}
	return 1;
}

static int
echo_recv(sd)
	SESSION *sd;
{
	register struct ip *ip;
	register struct icmp *icmp;
	struct sockaddr_in from;
	socklen_t total_len, header_len;
	REQUEST *request;
	u_char buf[MAX_PACKETSZ];

	dprintf(("echo_recv: entered\n"));

	if ((request = get_request(sd, 0 /* no reqid yet */)) == NULL) {
		trace(("\t echo_recv: internal error?\n"));
		return -1;
	}

	header_len = sizeof(from);
	if ((total_len = recvfrom(sd->sock, (char *)buf, sizeof(buf), 0,
				  (struct sockaddr *)&from, &header_len)) < 0) {
		echo_reply(errno, sd, request->reqid);
		return -1;
	}
	if (total_len < MIN_PACKETSZ) {
		trace(("\t echo_recv: truncated packet\n"));
		return -1;
	}

	/* check the IP header first */
	ip = (struct ip *)buf;
#ifdef	TRACE
	dump_ip(ip);
#endif
	header_len = ip->ip_hl << 2;
	if (header_len < sizeof(struct ip) ||
	    header_len > sizeof(struct ip) + MAX_IPOPTLEN) {
		trace(("\t echo_recv: wrong ip header length\n"));
		return -1;
	}
	if (ip->ip_p != IPPROTO_ICMP) {
		trace(("\t echo_recv: not icmp packet\n"));
		return -1;
	}

	total_len -= header_len;
	if (total_len < ICMP_MINLEN + sizeof(TIMEVAL *)) {
		trace(("\t echo_recv: truncated packet\n"));
		return -1;
	}

	/* check the ICMP header */
	icmp = (struct icmp *)(buf + header_len);
#ifdef	TRACE
	dump_icmp(icmp);
#endif
	if (icmp->icmp_type == ICMP_ECHOREPLY) {
		TIMEVAL *tvp, tv_buf;

		if (!ident_icmp_packet(icmp, request))
			return -1; /* not my packet */

		/* ICMP data must match our session buffer address */
		memcpy(&tvp, icmp->icmp_data, sizeof(TIMEVAL *));
		if (tvp != (TIMEVAL *)sd->buf) {
			trace(("\t echo_recv: bad session id\n"));
			return -1;
		}
		gettimeofday(&tv_buf, 0);
		tv_sub(&tv_buf, tvp);
		*tvp = tv_buf;

		/* extract IP Record Route Options if any */
		if (sd->data_ptr && header_len > sizeof(struct ip)) {
			u_char *cp = buf + sizeof(struct ip);
			int len = cp[IPOPT_OLEN] - 3; /* 3 header octets */
			if (cp[IPOPT_OPTVAL] == IPOPT_RR &&
			    len > 0 && len < MAX_IPOPTLEN - 1) {
				len = (cp[IPOPT_OFFSET] - IPOPT_MINOFF) / sizeof(in_addr_t);
				if (len > 0 && len < 10) {
					sd->data_int = len;
					cp += IPOPT_OFFSET + 1;
					memcpy(sd->data_ptr, cp, len * sizeof(in_addr_t));
				}
			}
		}
		return 0;
	}
	if (icmp->icmp_type == ICMP_UNREACH ||
	    icmp->icmp_type == ICMP_TIMXCEED ||
	    icmp->icmp_type == ICMP_PARAMPROB ||
	    icmp->icmp_type == ICMP_SOURCEQUENCH) {
		struct sockaddr_in *to = (struct sockaddr_in *)&sd->peer;
		int err = icmp->icmp_type;
		err = (err << 8) | icmp->icmp_code;

		/* extract original IP header caused error */
		ip = &icmp->icmp_ip;
#ifdef	TRACE
		dump_ip(ip);
#endif
		/* check the original IP header */
		header_len = ip->ip_hl << 2;
		if (header_len < sizeof(struct ip) ||
		    header_len > sizeof(struct ip) + MAX_IPOPTLEN)
			return -1;
		if (ip->ip_p != IPPROTO_ICMP ||
		    memcmp(&ip->ip_dst, &to->sin_addr, sizeof(to->sin_addr)))
			return -1;

		/* check the original ICMP header */
		icmp = (struct icmp *)((u_char *)ip + header_len);
#ifdef	TRACE
		dump_icmp(icmp);
#endif
		if (icmp->icmp_type != ICMP_ECHO ||
		    !ident_icmp_packet(icmp, request))
			return -1;

		sd->data_int = err;
		if (sd->data_ptr) {
			free(sd->data_ptr);
			sd->data_ptr = 0;
		}
		return 0;
	}
	if (icmp->icmp_type == ICMP_REDIRECT) {
		; /* noop */
	}
	return -1;
}

void
echo_stop(method)
	METHOD *method;
{
	session_stop(method->sd, 1);
}

void
echo_start(method)
	METHOD *method;
{
	SESSION *sd = method->sd;
	int tmpval;

	/* sanity check */
	if (!sd) return;

	echo_stop(method); /* stop the method if still running */

	if (!method->ping_send) {
		echo_reply(0, sd, 0);
		return;
	}

	/* open RAW socket near before ICMP session */
	if ((sd->sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
		echo_reply(errno, sd, 0);
		return;
	}
#ifdef	SO_BSDCOMPAT
	/* The following option is only necessary on Linux machines because
	 * they have the unusual behavior of returning some ICMP errors to
	 * the process that generated them, which means that when we get a
	 * PORT UNREACHABLE error, it causes the next sendto to fail and
	 * return ECONNREFUSED. Switching to BSD_COMPAT mode solves that
	 * problem.
	 */
	tmpval = 1;
	if (setsockopt(sd->sock, SOL_SOCKET, SO_BSDCOMPAT, &tmpval, sizeof(int)) < 0) {
		echo_reply(errno, sd, 0);
		return;
	}
#endif
	tmpval = 1;
	if (setsockopt(sd->sock, IPPROTO_IP, IP_HDRINCL, &tmpval, sizeof(int)) < 0) {
		echo_reply(errno, sd, 0);
		return;
	}

	/* prepare buffer for timestamp */
	if ((sd->buf = (char *)malloc(sizeof(TIMEVAL))) == NULL) {
		echo_reply(errno, sd, 0);
		return;
	}
	memset(sd->buf, 0, sizeof(TIMEVAL));

	if (!method->rport) { /* built-in method used RR */
		/* prepare us for record routes */
		if ((sd->data_ptr = (char *)malloc(MAX_IPOPTLEN)) == NULL) {
			echo_reply(errno, sd, 0);
			return;
		}
		memset(sd->data_ptr, 0, MAX_IPOPTLEN);
	}

#ifdef	NO_ICMP_ERRORS
	if (set_socket_async(sd->sock, TRUE) < 0) {
		echo_reply(errno, sd, 0);
		return;
	}
	if (connect(sd->sock, &sd->peer, sizeof(struct sockaddr)) < 0) {
		echo_reply(errno, sd, 0);
		return;
	}
#endif
	if ((tmpval = session_start(sd, NULL)) != 0) {
		dprintf(("echo_start(%s/%s): reqid=%d\n",
			 sd->owner->name, method->name, tmpval));
	}
}

static void
echo_reply(op, sd, reqid)
	int op;
	SESSION *sd;
	int reqid;
{
	OBJECT *target = sd->owner;
	METHOD *method = sd->method;
	VARIABLE *var;
	char *diag, buf[50], varname[100];

	/* sanity check */
	if (!target || !method) return;

	if (!op) { /* reply was received */
		if (sd->data_int && sd->data_ptr) {
			make_netpath(sd->owner, (in_addr_t *)sd->data_ptr, sd->data_int);
			sd->data_int = 0;
		}
		if (!sd->data_int) {
			if (sd->buf)
				tv_add(&sd->tv_sum, (TIMEVAL *)sd->buf);
			if (sd->pkt_send < method->ping_send) {
				session_start(sd, NULL);
				return;
			}
			/* calculate RTT */
			if (timerisset(&sd->tv_sum)) {
				double msec = ((double)sd->tv_sum.tv_sec) * 1000.0 +
						((double)sd->tv_sum.tv_usec) / 1000.0;
				if (sd->pkt_recv > 1) msec /= (double)sd->pkt_recv;
				sprintf(buf, "%g", msec);
				diag = buf;
			} else	diag = "0.000";
		} else {
			op = -1;
			diag = icmp_error(sd->data_int);
		}
	} else if (op == ETIMEDOUT) {
		if (sd->pkt_send < method->ping_send) {
			session_start(sd, NULL);
			return;
		} else	diag = strerror(op);
	} else	diag = strerror(op);

	dprintf(("echo_reply(%s/%s): reqid=%d diag=\"%s\"\n",
		 target->name, method->name, reqid, diag));

	/* save method statistics */
	if (op != -1) op = (sd->pkt_recv < method->ping_expect);

	(void)strcpy(varname, method->name);
	(void)strcat(varname, ".send");
	if ((var = get_var(target->var_list, varname)) != NULL)
		set_var(var, INTEGER, &sd->pkt_send, sizeof(sd->pkt_send));

	(void)strcpy(varname, method->name);
	(void)strcat(varname, ".recv");
	if ((var = get_var(target->var_list, varname)) != NULL)
		set_var(var, INTEGER, &sd->pkt_recv, sizeof(sd->pkt_recv));

	echo_stop(method);

	method_finished(target, method, diag, !op);
}

int
echo_init(target, method)
	OBJECT *target;
	METHOD *method;
{
	SESSION template;
	struct sockaddr_in *to;
	char varname[100];

	dprintf(("echo_init(%s/%s)\n", target->name, method->name));

	if (method->sd)
		session_free(method->sd);

	if (!method->ping_expect || method->ping_expect > method->ping_send)
		method->ping_expect = method->ping_send;

	memset(&template, 0, sizeof(template));
	template.owner = target;
	template.method = method;
	template.sock = -1;	/* not yet opened */
	to = (struct sockaddr_in *)&template.peer;
	to->sin_family = AF_INET;
	to->sin_addr = method->address ? method->ip_addr : target->ip_addr;
	template.timeout = method->timeout * 1000000L; /* make microseconds */
	template.retries = method->retries;
	template.send = echo_send;
	template.recv = echo_recv;
	template.read = echo_reply;

	if ((method->sd = session_create(&template)) == NULL) {
		errno = ENOMEM;
		return -1;
	}

	free_netpath(target->nexthop);
	target->nexthop = 0;

	(void)strcpy(varname, method->name);
	(void)strcat(varname, ".send");
	add_var(&target->var_list, varname);

	(void)strcpy(varname, method->name);
	(void)strcat(varname, ".recv");
	add_var(&target->var_list, varname);

	return 0;
}

/*
 * Built-in method initiator block.
 */

METHOD	method_ping = {
	echo_init,	/* initialize icmp session */
	echo_start,	/* start icmp session */
	echo_stop,	/* stop icmp session */

	"PING",		/* method name */
	NULL,		/* no argument by default */
	NULL, { 0 },	/* use target address by default */
	IPPROTO_ICMP,	/* network protocol */
	0,		/* no packet size for built-in method */
	0, 0,		/* timeout and retries undefined yet */
	{ 1, 1 },	/* send/expect packet counter */

	/* Non-initialized data */
	NULL,		/* session descriptor */
	NULL,		/* chat-script */
	NULL,		/* snmp-request */
	NULL		/* next method */
};
