/*
 *	Copyright (c) 1999-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"
#include "snmp.h"
#include "mib.h"

extern int errno;
static void router_reply __P((int, SESSION *, int));

#define	BATCH_DEFAULT	20 /* default value for the router batch */

/* add variable/instance to the list of references */
static int
add_var_ref(ref, var, index)
	REFERENCE **ref;
	VARIABLE *var;
	long index;
{
	REFERENCE *curr, *last;

	/* sanity check */
	if (!ref || !var || !var->objid)
		return 0;

	if (index != -1 && var->instance_len >= 0) {
		oid oidbuf[MAX_OID_LEN];
		set_instance(var, long2oid(oidbuf, var->instance_len, index),
			     var->instance_len);
	}
	if (var->old) {
		if (var->old->instance) {
			free(var->old->instance);
			var->old->instance = 0;
		}
		var->old->instance_len = 0;
	}

	last = 0;
	for (curr = *ref; curr; curr = curr->next) {
		if (curr->var == var) /* prevent dups */
			return 1;
		last = curr;
	}
	/* append */
	if ((curr = (REFERENCE *)malloc(sizeof(REFERENCE))) == NULL)
		return 0;
	memset(curr, 0, sizeof(REFERENCE));
	if ((curr->name = strdup(var->name)) == NULL) {
		free(curr);
		return 0;
	}
	curr->var = var;
	if (last)
		last->next = curr;
	else	*ref = curr;
	return 1;
}

static int
is_var_advanced(var)
	VARIABLE *var;
{
	u_long curr = oid2long(var->instance, var->instance_len);
	u_long prev = curr;
	if (var->old && var->old->instance && var->old->instance_len)
		prev = oid2long(var->old->instance, var->old->instance_len);
	return (curr > prev);
}

static INTERFACE *
get_iface_by_reqid(reqid, router)
	int reqid;
	OBJECT *router;
{
	INTERFACE *iface;

	for (iface = router->interface; iface; iface = iface->next) {
		if (iface->tabIdx && iface->reqid == reqid)
			return iface;
	}
	return NULL;
}

static BGP_PEER *
get_bgppeer_by_reqid(reqid, router)
	int reqid;
	OBJECT *router;
{
	BGP_AS *bgp;
	BGP_PEER *peer;

	for (bgp = router->bgp; bgp; bgp = bgp->next) {
		if (bgp->asn < 1) continue;
		for (peer = bgp->peer; peer; peer = peer->next) {
			if (peer->tabIdx && peer->reqid == reqid)
				return peer;
		}
	}
	return NULL;
}

static ENV_MON *
get_envmon_by_reqid(reqid, router)
	int reqid;
	OBJECT *router;
{
	ENV_MON *env;

	for (env = router->env; env; env = env->next) {
		if (env->gauge && env->reqid == reqid)
			return env;
	}
	return 0;
}

#ifdef	unused
/* convert OID address to IP address (network byte order) */
void
oid2addr(oid_addr, ip_addr)
	oid *oid_addr;
	struct in_addr *ip_addr;
{
	int i;
	in_addr_t addr = 0;

	for (i = 0; i < sizeof(addr); i++) {
		addr = addr << 8;
		addr |= oid_addr[i] & 0xff;
	}
	addr = htonl(addr);
	memcpy(ip_addr, &addr, sizeof(addr));
}

/* convert IP address to OID address (network byte order) */
void
addr2oid(ip_addr, oid_addr)
	struct in_addr *ip_addr;
	oid *oid_addr;
{
	int i;
	in_addr_t addr;

	memcpy(&addr, ip_addr, sizeof(addr));
	addr = ntohl(addr);
	for (i = sizeof(addr)-1; i >= 0; i--) {
		oid_addr[i] = addr & 0xff;
		addr = addr >> 8;
	}
}
#endif /* unused */

void
free_ifaddr(chain)
	IFADDR **chain;
{
	IFADDR *curr, *next;

	curr = *chain;
	while (curr) {
		next = curr->next;
		free(curr);
		curr = next;
	}
	*chain = NULL;
}

static IFADDR *
get_ifaddr(list, ip_addr)
	register IFADDR *list;
	struct in_addr ip_addr;
{
	while (list) {
		if (!memcmp(&list->ip_addr, &ip_addr, sizeof(ip_addr)))
			return list;
		list = list->next;
	}
	return 0;
}

/* add any IP address to addr list */
static void
add_ifaddr(chain, ip_addr)
	IFADDR **chain;
	struct in_addr ip_addr;
{
	IFADDR *curr, *prev;

	/* prevent dups */
	if (!chain || get_ifaddr(*chain, ip_addr))
		return;

	if ((curr = (IFADDR *)malloc(sizeof(IFADDR))) == NULL) {
		report(LOG_ERR, "add_ifaddr: Out of memory");
		return;
	}
	memset(curr, 0, sizeof(IFADDR));
	memcpy(&curr->ip_addr, &ip_addr, sizeof(ip_addr));
	if (*chain) {
		for (prev = *chain; prev->next; prev = prev->next);
		prev->next = curr;
	} else	*chain = curr;
}

/* add interface to list of object interfaces and return it */
static INTERFACE *
add_interface(object, proto, name, index)
	OBJECT *object;
	INTERFACE *proto;
	char *name;
	int index;
{
	INTERFACE *iface, *last = 0;
	VARIABLE *var;
	SAVE *save;
	GROUP_REF *group_ref;

	/* sanity check */
	if (!target || !proto || !name || !*name)
		return 0;

	/* search for duplicated or previously configured interface  */
	for (iface = object->interface; iface; iface = iface->next) {
		if (iface->tabIdx && iface->tabIdx == index)
			break;
		if (iface->name) {
			if (!strcasecmp(iface->name, name))
				break;
		} else if (iface->index > 0 && iface->index == index)
			break;

		last = iface;
	}
	if (iface) { /* already in the list; append new group name if any */
		if (proto->name && (var = add_var(&iface->var_list, "IFGROUP")) != NULL) {
			char buf[1024];
			buf[0] = '\0';
			if (var->type == STRING && var->value.string) {
				char *cp, *next;
				next = strncpy(buf, var->value.string, sizeof(buf)/2);
				buf[sizeof(buf)/2-1] = '\0';
				while ((cp = my_strsep(&next, ",")) != NULL) {
					if (!strcasecmp(cp, proto->name))
						return iface;
				}
				(void)strncpy(buf, var->value.string, sizeof(buf)/2);
				buf[sizeof(buf)/2-1] = '\0';
			}
			if (buf[0]) (void)strcat(buf, ",");
			(void)strcat(buf, proto->name);
			set_var(var, STRING, buf, strlen(buf));
		}
		return iface;
	}
	/* append as full copy of prototype */
	if ((iface = (INTERFACE *)malloc(sizeof(INTERFACE))) == NULL) {
		report(LOG_ERR, "add_interface: Out of memory");
		return 0;
	}
	memcpy(iface, proto, sizeof(INTERFACE));
	iface->next = 0;
	if ((iface->name = strdup(name)) == NULL) {
		report(LOG_ERR, "add_interface: Out of memory");
		return 0;
	}
	if (proto->descr && (iface->descr = strdup(proto->descr)) == NULL) {
		report(LOG_ERR, "add_interface: Out of memory");
		return 0;
	}
	if (proto->datadir && (iface->datadir = strdup(proto->datadir)) == NULL) {
		report(LOG_ERR, "add_interface: Out of memory");
		return 0;
	}

	iface->var_list = 0;
	if (proto->name && (var = add_var(&iface->var_list, "IFGROUP")) != NULL)
		set_var(var, STRING, proto->name, strlen(proto->name));
	for (var = proto->var_list; var; var = var->next)
		add_var_copy(&iface->var_list, var);

	iface->save_list = 0;
	for (save = proto->save_list; save; save = save->next)
		add_save_copy(&iface->save_list, save);

	iface->ns_acl = 0;
	for (group_ref = proto->ns_acl; group_ref; group_ref = group_ref->next)
		add_group_ref_copy(&iface->ns_acl, group_ref);

        if (last)
                last->next = iface;
        else    object->interface = iface;
	return iface;
}

/* add bgp peer to list of AS's peers */
static BGP_PEER *
add_bgp_peer(bgp, ip_addr)
	BGP_AS *bgp;
	struct in_addr ip_addr;
{
	BGP_PEER *peer, *last = 0;

	if (!bgp) return 0;

	/* search for duplicated or previously configured peer */
	for (peer = bgp->peer; peer; peer = peer->next) {
		if (!memcmp(&peer->ip_addr, &ip_addr, sizeof(ip_addr))) {
			peer->parent = bgp; /* just for certainty */
			return peer;
		}
		last = peer;
	}
	if (bgp->handconf)
		return 0; /* the address of peer was not configured manualy */
 
	if ((peer = (BGP_PEER *)malloc(sizeof(BGP_PEER))) == NULL) {
		report(LOG_ERR, "add_bgp_peer: Out of memory");
		return 0;
	}
	memset(peer, 0, sizeof(BGP_PEER));
	memcpy(&peer->ip_addr, &ip_addr, sizeof(ip_addr));
	peer->parent = bgp;
	if (last)
		last->next = peer;
	else    bgp->peer = peer;
	return peer;
}

void
free_bgp_peers(bgp, all)
	BGP_AS *bgp;
	int all;
{
	BGP_PEER *peer, *next;

	if (!bgp) return;

	if (all || !bgp->handconf) {
		peer = bgp->peer;
		while (peer) {
			next = peer->next;
			free_var_list(peer->var_list);
			free(peer);
			peer = next;
		}
		bgp->peer = NULL;
	}
}

/* add env gauge to list */
static ENV_GAUGE *
add_env_gauge(env, name)
	ENV_MON *env;
	const char *name;
{
	ENV_GAUGE *gauge, *last = 0;

	if (!env || !name || *name == '\0')
		return 0;

	/* search for duplicated or previously configured gauge name */
	for (gauge = env->gauge; gauge; gauge = gauge->next) {
		if (!strcasecmp(gauge->name, name)) {
			gauge->parent = env; /* just for certainty */
			return gauge;
		}
		last = gauge;
	}
	if (env->handconf)
		return 0; /* the name of gauge was not configured manualy */

	if ((gauge = (ENV_GAUGE *)malloc(sizeof(ENV_GAUGE))) == NULL) {
		report(LOG_ERR, "add_env_gauge: Out of memory");
		return 0;
	}
	memset(gauge, 0, sizeof(ENV_GAUGE));
	if ((gauge->name = strdup(name)) == 0) {
		free(gauge);
		return 0;
	}
	gauge->parent = env;
	if (last)
		last->next = gauge;
	else	env->gauge = gauge;
	return gauge;
}

void
free_env_gauges(env, all)
	ENV_MON *env;
	int all;
{
	ENV_GAUGE *gauge, *next;

	if (!env) return;

	if (all || !env->handconf) {
		gauge = env->gauge;
		while (gauge) {
			next = gauge->next;
			free_var_list(gauge->var_list);
			free(gauge);
			gauge = next;
		}
		env->gauge = NULL;
	}
}

static SNMP_REQ *
create_request(cmd, ref)
	int cmd;
	REFERENCE *ref;
{
	SNMP_REQ *snmpreq;

	if ((snmpreq = (SNMP_REQ *)malloc(sizeof(SNMP_REQ))) == NULL) {
		free_var_ref(ref);
		report(LOG_ERR, "create_request: Out of memory");
		return 0;
	}
	memset(snmpreq, 0, sizeof(SNMP_REQ));
	snmpreq->command = cmd;
	snmpreq->var_ref = ref;
	return snmpreq;
}

static void
clear_requests(target)
	OBJECT *target;
{
	if (target->snmpdata)
		target->snmpdata->reqid = 0;
	if (target->interface) {
		INTERFACE *iface = target->interface;
		for (; iface; iface = iface->next) {
			iface->reqid = 0;
		}
	}
	if (target->bgp) {
		BGP_AS *bgp = target->bgp;
		for (; bgp; bgp = bgp->next) {
			BGP_PEER *peer = bgp->peer;
			for (; peer; peer = peer->next) {
				peer->reqid = 0;
			}
		}
	}
	if (target->env) {
		ENV_MON *env = target->env;
		for (; env; env = env->next) {
			env->reqid = 0;
		}
	}
}

static void
send_system_request(sd, target)
	SESSION *sd;
	OBJECT *target;
{
	METHOD *method = sd->method;
	SNMP_DATA *snmp = target->snmpdata;
	VARIABLE *var;
	REFERENCE *ref = NULL;

	/* sanity check */
	if (!method || !snmp || !target->var_list) {
		trace(("\t DATA ERROR at send_system_request()\n"));
		report(LOG_ERR, "send_system_request(%s): Invalid arguments: method=%p snmp=%p var_list=%p",
		       target->name, method, snmp, target->var_list);
		return;
	}

	if (!snmp->mode) /* initial state */
		snmp->mode = GET_STATE;

	if (snmp->reqid) {
		snmp->reqid = 0; /* force request next time */
		report(LOG_WARNING, "Has lost %s request at \"%s\"",
		       method->name, target->name);
		return;
	}
	for (var = target->var_list; var; var = var->next) {
		if (var->objid && (var->mode & snmp->mode) != 0)
			add_var_ref(&ref, var, 0);
	}
	if (ref) {
		SNMP_REQ *req = create_request(GET_REQUEST, ref);
		if (ref && (snmp->reqid = session_start(sd, req)) != 0) {
			/*target->prev_request = target->last_request;*/
			gettimeofday(&target->last_request, 0);

			dprintf(("send_system_request(%s): mode=%06o reqid=%d\n",
				 target->name, snmp->mode, snmp->reqid));
		}
	}
}

static int
send_interface_request(sd, target, iface)
	SESSION *sd;
	OBJECT *target;
	INTERFACE *iface;
{
	METHOD *method = sd->method;
	VARIABLE *var;
	REFERENCE *ref = NULL;

	/* sanity check */
	if (!method || !iface->var_list || !iface->tabIdx) {
		trace(("\t DATA ERROR at send_interface_request()\n"));
		report(LOG_ERR, "send_interface_request(%s/%s): Invalid arguments: method=%p var_list=%p tabIdx=%d",
		       target->name, iface->name ? iface->name : itoa(iface->index),
		       method, iface->var_list, iface->tabIdx);
		return 0;
	}

	if (!iface->mode) /* initial state */
		iface->mode = GET_STATE;

	if (iface->reqid) {
		iface->reqid = 0; /* force request next time */
		report(LOG_WARNING, "Has lost %s request on \"%s\" at \"%s\"",
		       method->name, iface->name ? iface->name : itoa(iface->index),
		       target->name);
		return 0;
	}
	for (var = iface->var_list; var; var = var->next) {
		if (var->objid && (var->mode & iface->mode) != 0)
			add_var_ref(&ref, var, iface->tabIdx);
	}
	if (ref) {
		SNMP_REQ *req = create_request(GET_REQUEST, ref);
		if (req && (iface->reqid = session_start(sd, req)) != 0) {
			iface->prev_request = iface->last_request;
			gettimeofday(&iface->last_request, 0);

			dprintf(("send_interface_request(%s/%s): mode=%06o reqid=%d\n",
				 target->name, iface->name ? iface->name : itoa(iface->index),
				 iface->mode, iface->reqid));
			return 1;
		}
	}
	return 0;
}

static void
send_bgppeer_request(sd, target, peer)
	SESSION *sd;
	OBJECT *target;
	BGP_PEER *peer;
{
	METHOD *method = sd->method;
	VARIABLE *var;
	REFERENCE *ref = NULL;

	/* sanity check */
	if (!method || !peer->var_list || !peer->tabIdx) {
		char tmpbuf[100];
		trace(("\t DATA ERROR at send_bgppeer_request()\n"));
		report(LOG_ERR, "send_bgppeer_request(%s/%s): Invalid arguments: method=%p var_list=%p tabIdx=%d",
		       target->name, intoa(tmpbuf, peer->ip_addr),
		       method, peer->var_list, peer->tabIdx);
		return;
	}

	if (!peer->mode) /* initial state */
		peer->mode = GET_STATE;

	if (peer->reqid) {
		char tmpbuf[100];
		peer->reqid = 0; /* force request next time */
		report(LOG_WARNING, "Has lost %s request on \"%s\" at \"%s\" (AS%d)",
		       method->name, intoa(tmpbuf, peer->ip_addr),
		       target->name, peer->parent ? peer->parent->asn : 0);
		return;
	}
	for (var = peer->var_list; var; var = var->next) {
		if (var->objid && (var->mode & peer->mode) != 0)
			add_var_ref(&ref, var, peer->tabIdx);
	}
	if (ref) {
		SNMP_REQ *req = create_request(GET_REQUEST, ref);
		if (req && (peer->reqid = session_start(sd, req)) != 0) {
#ifdef	DEBUG
			char tmpbuf[100];
#endif
			peer->prev_request = peer->last_request;
			gettimeofday(&peer->last_request, 0);

			dprintf(("send_bgppeer_request(%s/%s): mode=%06o reqid=%d\n",
				 target->name, intoa(tmpbuf, peer->ip_addr),
				 peer->mode, peer->reqid));
		}
	}
}

static void
send_envmon_request(sd, target, env)
	SESSION *sd;
	OBJECT *target;
	ENV_MON *env;
{
	METHOD *method = sd->method;
	VARIABLE *var;
	ENV_GAUGE *gauge;
	REFERENCE *ref = NULL;

	/* sanity check */
	if (!method)
		return;

	if (!env->mode) /* initial state */
		env->mode = GET_STATE;

	if (env->reqid) {
		env->reqid = 0; /* force request next time */
		report(LOG_WARNING, "Has lost %s request on \"%s\" at \"%s\"",
		       method->name, get_name_by_type(env->type), target->name);
		return;
	}
	for (gauge = env->gauge; gauge; gauge = gauge->next) {
		if (!gauge->tabIdx || !gauge->var_list)
			continue;
		for (var = gauge->var_list; var; var = var->next) {
			if (var->objid && (var->mode & env->mode) != 0)
				add_var_ref(&ref, var, gauge->tabIdx);
		}
	}
	if (ref) {
		SNMP_REQ *req = create_request(GET_REQUEST, ref);
		if (req && (env->reqid = session_start(sd, req)) != 0) {
			env->prev_request = env->last_request;
			gettimeofday(&env->last_request, 0);

			dprintf(("send_envmon_request(%s/%s): mode=%06o reqid=%d\n",
				 target->name, get_name_by_type(env->type),
				 env->mode, env->reqid));
		}
	}
}

/*
 * Send SNMP periodic request.
 */
static void
send_router_request(sd, target)
	SESSION *sd;
	OBJECT *target;
{
	SNMP_DATA *snmp;

	/* sanity check */
	if (!sd || !sd->method || !target || !target->snmpdata) {
		trace(("\t DATA ERROR at send_souter_request()\n"));
		report(LOG_ERR, "send_router_request(%s): Invalid arguments",
		       target ? target->name : "null");
		return;
	}
	snmp = target->snmpdata;
	if (snmp->scan) {
		if (snmp->scan & SCAN_IFTABLE)
			report(LOG_NOTICE, "Has lost ifTable scan at \"%s\"", target->name);

		if (snmp->scan & SCAN_IFADDR)
			report(LOG_NOTICE, "Has lost ipTable scan at \"%s\"", target->name);

		if (snmp->scan & SCAN_BGPPEER)
			report(LOG_NOTICE, "Has lost bgpTable scan at \"%s\"", target->name);

		if (snmp->scan & SCAN_ENVMON)
			report(LOG_NOTICE, "Has lost envTable scan at \"%s\"", target->name);

		snmp->pending_scan |= snmp->scan;
		snmp->scan = 0;
	}
	send_system_request(sd, target);
}

static int
advance_iftable(ref, target, ver)
	REFERENCE **ref;
	OBJECT *target;
	int ver;
{
	SNMP_DATA *snmp = target->snmpdata;
	VARIABLE *var;
	INTERFACE *iface;
	int first_time = 0, nref = 0;
	long index = 0, instance = 0;

	if (!snmp->ifTable) { /* first time init */
		/* remove all MIB variables first time before scan */
		for (iface = target->interface; iface; iface = iface->next) {
			iface->state = 0;
			iface->mode = 0;
			iface->reqid = 0;
			iface->tabIdx = 0;
			free_mib_table(&iface->var_list, 0);
		}
		add_mib_table(&snmp->ifTable, interface_MIB, ver);
		if (snmp->device == DEVICE_CISCO)
			add_mib_table(&snmp->ifTable, cisco_interface_MIB, ver);

		first_time = 1;
		snmp->ifPhase = SCAN_MATCH;

	} else if (snmp->recover_scan) {
		snmp->recover_scan &= ~SCAN_IFTABLE;
		first_time = 1;
		index = -1;
	}

	if (snmp->ifPhase == SCAN_MATCH) {
		VARIABLE *av[100];
		int ac = 0;

		/* collect variables to match */
		for (var = snmp->ifTable; var; var = var->next) {
			if (!var->objid || (var->mode & SCAN_MATCH) == 0)
				continue;

			if (first_time) {
				nref += add_var_ref(ref, var, index);
				continue;
			}
			if (!is_var_advanced(var)) {
				var->mode &= ~SCAN_MATCH; /* exclude it */
				trace(("\t Phase0: EOT at \"%s\" detected\n",
				       var->name));
				continue;
			}
			instance = oid2long(var->instance, var->instance_len);
			if (index == 0 || index == -1)
				index = instance;
			if (instance == 0 || instance == -1 || instance != index) {
				var->mode &= ~SCAN_MATCH; /* exclude it */
				trace(("\t Phase0: index %d at \"%s\"???\n",
				       instance, var->name));
				continue;
			}
			if (ac < 100) av[ac++] = var;
		}

		/* match collected variables against configured interfaces */
		if (ac) {
			int i, matched = -1;

			/* first match for directly specified interfaces */
			for (iface = target->interface; iface; iface = iface->next) {
				if (iface->tabIdx)
					continue; /* already matched */
				for (i = 0; i < ac; i++) {
					var = av[i];
					if (var->type == INTEGER && var->value.integer) {
						if (iface->index == *var->value.integer) {
							matched = i;
							break;
						}
					} else if (var->type == STRING && var->value.string) {
						if (match_ifdescr(iface->name, var->value.string)) {
							matched = i;
							break;
						}
					}
				}
				if (matched != -1) break;
			}

			/* if not matched above try group prototypes */
			if (matched == -1 && target->ifgroup) {
				char *cp, buf[1024];
				for (i = 0; i < ac; i++) {
					cp = stripquotes(var2str(buf, av[i]));
					for (iface = target->ifgroup; iface; iface = iface->next) {
						if (check_group(iface->name, cp, 0)) {
							matched = i;
							break;
						}
					}
					if (matched != -1) {
						iface = add_interface(target, iface,
								      cp, index);
						break;
					}
				}
			}

			/* if matched anything above than bind variables */
			if (matched != -1 && iface) {
				if (iface->tabIdx && iface->tabIdx != index) {
					var = av[matched];
					if (iface->name) {
						report(LOG_INFO, "$%s of \"%s\" rematched (index %d->%d) at \"%s\"",
						       var->name, iface->name,
						       iface->tabIdx, index,
						       target->name);
					} else if (iface->index > 0) {
						report(LOG_INFO, "$%s of Interface%d rematched (index %d->%d) at \"%s\"",
						       var->name, iface->index,
						       iface->tabIdx, index,
						       target->name);
					}
					trace(("\t Rematch: \"%s\" of \"%s\" index %d->%d at \"%s\"\n",
					       var->name,
					       iface->name ? iface->name : itoa(iface->index),
					       iface->tabIdx, index,
					       target->name));
				}
				iface->tabIdx = index;
				for (i = 0; i < ac; i++) {
					var = add_var_copy(&iface->var_list, av[i]);
					if (i == matched)
						var->mode = SCAN_MATCH | GET_STATE;
					else	var->mode = SCAN_DETECT| GET_INFO;

					trace(("\t Phase0: %s \"%s\" to \"%s\" on \"%s\"\n",
					       i == matched ? "bind" : "add",
					       var->name,
					       iface->name ? iface->name : itoa(iface->index),
					       target->name));
				}
			}

			/* is there everything matched? */
			matched = 0;
			if (!target->ifgroup) {
				i = 0;
				for (iface = target->interface; iface; iface = iface->next) {
					if (iface->tabIdx) matched++;
					i++;
				}
				matched = (i == matched);
			}

			/* advance in the table if required */
			if (!matched) {
				for (i = 0; i < ac; i++)
					nref += add_var_ref(ref, av[i], index);
			}
		}
		if (nref) return 1; /* advance to next */

		/* find lowest interface index for next phase */
		index = 0;
		for (iface = target->interface; iface; iface = iface->next) {
			if (iface->tabIdx && (!index || iface->tabIdx < index))
				index = iface->tabIdx;
		}
		if (index != 0 && index != -1) {
			index--;	/* just before it */
			first_time = 1;
			snmp->ifPhase = SCAN_DETECT;
		}
	}

	if (snmp->ifPhase == SCAN_DETECT) {
		for (var = snmp->ifTable; var; var = var->next) {
			if (!var->objid || (var->mode & SCAN_DETECT) == 0)
				continue;

			if (first_time) {
				nref += add_var_ref(ref, var, index);
				continue;
			}
			if (!is_var_advanced(var)) {
				var->mode &= ~SCAN_DETECT; /* exclude it */
				trace(("\t Phase1: EOT at \"%s\" detected\n",
				       var->name));
				continue;
			}

			/* check for reply */
			instance = oid2long(var->instance, var->instance_len);
			if (instance == 0 || instance == -1) {
				var->mode &= ~SCAN_DETECT; /* exclude it */
				trace(("\t Phase1: index %d at \"%s\"???\n",
				       instance, var->name));
				continue;
			}

			/* add variable to an interface and find next index */
			index = 0;
			for (iface = target->interface; iface; iface = iface->next) {
				if (iface->tabIdx > instance &&
				    (!index || iface->tabIdx < index))
					index = iface->tabIdx;

				if (iface->tabIdx == instance && var->length) {
					add_var_copy(&iface->var_list, var);

					trace(("\t Phase1: add \"%s\" to \"%s\" on \"%s\"\n",
					       var->name,
					       iface->name ? iface->name : itoa(iface->index),
					       target->name));
				}
			}
			if (index != 0 && index != -1)
				nref += add_var_ref(ref, var, index - 1);
			else	var->mode &= ~SCAN_DETECT; /* exclude it */
		}
		if (nref) return 1; /* advance to next */
	}

	free_var_list(snmp->ifTable);
	snmp->ifTable = 0;
	return 0;
}

static int
advance_iptable(ref, target, ver)
	REFERENCE **ref;
	OBJECT *target;
	int ver;
{
	SNMP_DATA *snmp = target->snmpdata;
	VARIABLE *var;
	INTERFACE *iface;
	int first_time = 0, nref = 0;
	long index = 0, instance = 0;
	struct in_addr ip_addr;
	memset(&ip_addr, 0, sizeof(ip_addr));

	if (!snmp->ipTable) { /* first time init */
		free_ifaddr(&target->ifaddr);
		for (iface = target->interface; iface; iface = iface->next) {
			free_ifaddr(&iface->ifaddr);
		}
		add_mib_table(&snmp->ipTable, ipaddr_MIB, ver);

		first_time = 1;

	} else if (snmp->recover_scan) {
		snmp->recover_scan &= ~SCAN_IFADDR;
		first_time = 1;
		index = -1;
	}

	for (var = snmp->ipTable; var; var = var->next) {
		if (first_time) {
			nref += add_var_ref(ref, var, index);
			continue;
		}
		if (!is_var_advanced(var)) {
			trace(("\t EOT at \"%s\" detected\n", var->name));
			continue;
		}

		/* check for reply */
		instance = oid2long(var->instance, var->instance_len);
		if (index == 0 || index == -1)
			index = instance;
		if (instance == 0 || instance == -1 || instance != index) {
			trace(("\t index %d at \"%s\"???\n", instance, var->name));
			continue;
		}

		if (var->type == IPADDRESS && var->value.string) {
			memcpy(&ip_addr, var->value.string,
			       var->length > sizeof(ip_addr) ?
			       sizeof(ip_addr) : var->length);

			nref += add_var_ref(ref, var, index);

		} else if (var->type == INTEGER && var->value.integer) {
#ifdef	TRACE
			char tmpbuf[100];
			intoa(tmpbuf, ip_addr);
			trace(("\t add [%s] to \"%s\"\n", tmpbuf, target->name));
#endif
			add_ifaddr(&target->ifaddr, ip_addr);

			/* match it against tabIdx (XXX must be ifIndex) */
			for (iface = target->interface; iface; iface = iface->next) {
				if (iface->tabIdx &&
				    iface->tabIdx == *var->value.integer) {
					trace(("\t add [%s] to \"%s\" on \"%s\"\n",
					       tmpbuf, iface->name ? iface->name : itoa(iface->index),
					       target->name));
					add_ifaddr(&iface->ifaddr, ip_addr);
					break;
				}
			}
			nref += add_var_ref(ref, var, index);
		}
	}
	if (nref) return 1;

	free_var_list(snmp->ipTable);
	snmp->ipTable = 0;
	return 0;
}

static int
advance_bgptable(ref, target, ver)
	REFERENCE **ref;
	OBJECT *target;
	int ver;
{
	SNMP_DATA *snmp = target->snmpdata;
	VARIABLE *var;
	BGP_AS *bgp;
	BGP_PEER *peer;
	int first_time = 0, nref = 0;
	long index = 0, instance = 0;
#ifdef	TRACE
	char tmpbuf[100];
#endif

	if (!snmp->bgpTable) { /* first time init */
		first_time = 1;
		/* remove all MIB variables first time before scan */
		for (bgp = target->bgp; bgp; bgp = bgp->next) {
			for (peer = bgp->peer; peer; peer = peer->next) {
				peer->state = 0;
				peer->mode = 0;
				peer->reqid = 0;
				peer->tabIdx = 0;
				free_mib_table(&peer->var_list, 0);
			}
		}
		add_mib_table(&snmp->bgpTable, bgp_MIB, ver);

		first_time = 1;
		snmp->bgpPhase = SCAN_MATCH;

	} else if (snmp->recover_scan) {
		snmp->recover_scan &= ~SCAN_BGPPEER;
		first_time = 1;
		index = -1;
	}

	if (snmp->bgpPhase == SCAN_MATCH) {
		/* last matched ASN */
		VARIABLE *matched_var = 0;
		BGP_AS *matched_bgp = 0;

		for (var = snmp->bgpTable; var; var = var->next) {
			if (!var->objid || (var->mode & SCAN_MATCH) == 0)
				continue;

			if (first_time) {
				nref += add_var_ref(ref, var, index);
				continue;
			}
			if (!is_var_advanced(var)) {
				var->mode &= ~SCAN_MATCH; /* exclude it */
				trace(("\t Phase0: EOT at \"%s\" detected\n",
				       var->name));
				continue;
			}

			/* check for reply */
			instance = oid2long(var->instance, var->instance_len);
			if (index == 0 || index == -1)
				index = instance;
			if (instance == 0 || instance == -1 || instance != index) {
				var->mode &= ~SCAN_MATCH; /* exclude it */
				trace(("\t Phase0: index %d at \"%s\"???\n",
				       instance, var->name));
				continue;
			}

			/* match for configured ASN */
			if (var->type == INTEGER && var->value.integer) {
				for (bgp = target->bgp; bgp; bgp = bgp->next) {
					if (bgp->asn == *var->value.integer) {
						matched_bgp = bgp;
						matched_var = var;
						break;
					}
				}
			} else if (matched_bgp &&
				   var->type == IPADDRESS && var->value.string) {
				struct in_addr ip_addr;	/* peer address */
				memcpy(&ip_addr, var->value.string,
				       var->length > sizeof(ip_addr) ?
				       sizeof(ip_addr) : var->length);
				peer = add_bgp_peer(matched_bgp, ip_addr);
				if (peer) {
					peer->tabIdx = index;
					if (matched_var) {
						add_var_copy(&peer->var_list, matched_var);

						trace(("\t Phase0: add \"%s\" to [%s] on \"%s\" (AS%d)\n",
						       matched_var->name,
						       intoa(tmpbuf, peer->ip_addr),
						       target->name,
						       peer->parent->asn));
					}
					add_var_copy(&peer->var_list, var);

					trace(("\t Phase0: add \"%s\" to [%s] on \"%s\" (AS%d)\n",
					       var->name,
					       intoa(tmpbuf, peer->ip_addr),
					       target->name,
					       peer->parent->asn));
				}
			}
			nref += add_var_ref(ref, var, index);
		}
		if (nref) return 1; /* advance to next */

		/* find lowest peer index for next phase */
		index = 0;
		for (bgp = target->bgp; bgp; bgp = bgp->next) {
			for (peer = bgp->peer; peer; peer = peer->next) {
				if (peer->tabIdx &&
				    (!index || peer->tabIdx < index))
					index = peer->tabIdx;
			}
		}
		if (index != 0 && index != -1) {
			index--;	/* just before it */
			first_time = 1;
			snmp->bgpPhase = SCAN_DETECT;
		}
	}

	if (snmp->bgpPhase == SCAN_DETECT) {
		for (var = snmp->bgpTable; var; var = var->next) {
			if (!var->objid || (var->mode & SCAN_DETECT) == 0)
				continue;

			if (first_time) {
				nref += add_var_ref(ref, var, index);
				continue;
			}
			if (!is_var_advanced(var)) {
				var->mode &= ~SCAN_DETECT; /* exclude it */
				trace(("\t Phase1: EOT at \"%s\" detected\n",
				       var->name));
				continue;
			}

			/* check for reply */
			instance = oid2long(var->instance, var->instance_len);
			if (instance == 0 || instance == -1) {
				var->mode &= ~SCAN_DETECT; /* exclude it */
				trace(("\t Phase1: index %d at \"%s\"???\n",
				       instance, var->name));
				continue;
			}

			/* add variable to bgp peer and find next index */
			index = 0;
			for (bgp = target->bgp; bgp; bgp = bgp->next) {
				for (peer = bgp->peer; peer; peer = peer->next) {
					if (peer->tabIdx > instance &&
					    (!index || peer->tabIdx < index))
						index = peer->tabIdx;

					if (peer->tabIdx == instance && var->length) {
						add_var_copy(&peer->var_list, var);

						trace(("\t Phase1: add \"%s\" to [%s] on \"%s\" (AS%d)\n",
						       var->name,
						       intoa(tmpbuf, peer->ip_addr),
						       target->name,
						       peer->parent->asn));
					}
				}
			}
			if (index != 0 && index != -1)
				nref += add_var_ref(ref, var, index - 1);
			else	var->mode &= ~SCAN_DETECT; /* exclude it */
		}
		if (nref) return 1; /* advance to next */
	}

	free_var_list(snmp->bgpTable);
	snmp->bgpTable = 0;
	return 0;
}

static ENV_MON *
match_envmib(env_list, name)
	ENV_MON *env_list;
	const char *name;
{
	ENV_MON *env = env_list;
	while (env) {
		if (env->type == TYPE_ENVTEMP &&
		    get_mib_entry(name, cisco_envtemp_MIB))
			return env;
		if (env->type == TYPE_ENVVOLT &&
		    get_mib_entry(name, cisco_envvolt_MIB))
			return env;
		if (env->type == TYPE_ENVFAN &&
		    get_mib_entry(name, cisco_envfan_MIB))
			return env;
		if (env->type == TYPE_ENVPS &&
		    get_mib_entry(name, cisco_envps_MIB))
			return env;
		env = env->next;
	}
	return 0;
}

static int
advance_envtable(ref, target, ver)
	REFERENCE **ref;
	OBJECT *target;
	int ver;
{
	SNMP_DATA *snmp = target->snmpdata;
	VARIABLE *var;
	ENV_MON *env;
	ENV_GAUGE *gauge;
	int first_time = 0, nref = 0;
	long index = 0, instance = 0;

	if (!snmp->envTable) { /* first time init */
		/* remove all MIB variables first time before scan */
		for (env = target->env; env; env = env->next) {
			env->mode = 0;
			env->reqid = 0;
			for (gauge = env->gauge; gauge; gauge = gauge->next) {
				gauge->state = 0;
				gauge->tabIdx = 0;
				free_mib_table(&gauge->var_list, 0);
			}
			if (snmp->device == DEVICE_CISCO) {
				if (env->type == TYPE_ENVTEMP)
					add_mib_table(&snmp->envTable, cisco_envtemp_MIB, ver);
				else if (env->type == TYPE_ENVVOLT)
					add_mib_table(&snmp->envTable, cisco_envvolt_MIB, ver);
				else if (env->type == TYPE_ENVFAN)
					add_mib_table(&snmp->envTable, cisco_envfan_MIB, ver);
				else if (env->type == TYPE_ENVPS)
					add_mib_table(&snmp->envTable, cisco_envps_MIB, ver);
			}
		}
		first_time = 1;
		snmp->envPhase = SCAN_MATCH;

	} else if (snmp->recover_scan) {
		snmp->recover_scan &= ~SCAN_ENVMON;
		first_time = 1;
		index = -1;
	}

	if (snmp->envPhase == SCAN_MATCH) {
		for (var = snmp->envTable; var; var = var->next) {
			if (!var->objid || (var->mode & SCAN_MATCH) == 0)
				continue;

			if (first_time) {
				nref += add_var_ref(ref, var, index);
				continue;
			}
			if (!is_var_advanced(var)) {
				var->mode &= ~SCAN_MATCH; /* exclude it */
				trace(("\t Phase0: EOT at \"%s\" detected\n",
				       var->name));
				continue;
			}

			/* check for reply */
			index = oid2long(var->instance, var->instance_len);
			if (index == 0 || index == -1) {
				var->mode &= ~SCAN_MATCH; /* exclude it */
				trace(("\t Phase0: index %d at \"%s\"???\n",
				       index, var->name));
				continue;
			}

			/* match for configured set of gauges */
			env = match_envmib(target->env, var->name);
			if (!env) {
				trace(("\t Phase0: no env for \"%s\"\n", var->name));
				continue;
			}
			if (var->type == STRING && var->value.string) {
				gauge = add_env_gauge(env, var->value.string);
				if (gauge) {
					gauge->tabIdx = index;
					add_var_copy(&gauge->var_list, var);

					trace(("\t Phase0: add \"%s\" (%d) to %s on \"%s\"\n",
					       var->name, gauge->tabIdx,
					       get_name_by_type(env->type),
					       target->name));
				}
			}
			nref += add_var_ref(ref, var, index);
		}
		if (nref) return 1; /* advance to next */

		/* find lowest gauge index for next phase */
		index = 0;
		for (env = target->env; env; env = env->next) {
			for (gauge = env->gauge; gauge; gauge = gauge->next) {
				if (gauge->tabIdx &&
				    (!index || gauge->tabIdx < index))
					index = gauge->tabIdx;
			}
		}
		if (index != 0 && index != -1) {
			index--;	/* just before it */
			first_time = 1;
			snmp->envPhase = SCAN_DETECT;
		}
	}

	if (snmp->envPhase == SCAN_DETECT) {
		for (var = snmp->envTable; var; var = var->next) {
			if (!var->objid || (var->mode & SCAN_DETECT) == 0)
				continue;

			if (first_time) {
				nref += add_var_ref(ref, var, index);
				continue;
			}
			if (!is_var_advanced(var)) {
				var->mode &= ~SCAN_DETECT; /* exclude it */
				trace(("\t Phase1: EOT at \"%s\" detected\n",
				       var->name));
				continue;
			}

			/* check for reply */
			instance = oid2long(var->instance, var->instance_len);
			if (instance == 0 || instance == -1) {
				var->mode &= ~SCAN_DETECT; /* exclude it */
				trace(("\t Phase1: index %d at \"%s\"???\n",
				       instance, var->name));
				continue;
			}

			env = match_envmib(target->env, var->name);
			if (!env) {
				trace(("\t Phase1: no env for \"%s\"\n", var->name));
				continue;
			}

			/* add variable to env gauge and find next index */
			index = 0;
			for (gauge = env->gauge; gauge; gauge = gauge->next) {
				if (gauge->tabIdx > instance &&
				    (!index || gauge->tabIdx < index))
					index = gauge->tabIdx;

				if (gauge->tabIdx == instance && var->length) {
					add_var_copy(&gauge->var_list, var);

					trace(("\t Phase1: add \"%s\" to %s on \"%s\"\n",
					       var->name,
					       get_name_by_type(env->type),
					       target->name));
				}
			}
			if (index != 0 && index != -1)
				nref += add_var_ref(ref, var, index - 1);
			else	var->mode &= ~SCAN_DETECT; /* exclude it */
		}
		if (nref) return 1; /* advance to next */
	}

	free_var_list(snmp->envTable);
	snmp->envTable = 0;
	return 0;
}

static void
parse_table_reply(sd, target)
	SESSION *sd;
	OBJECT *target;
{
	METHOD *method = sd->method;
	SNMP_DATA *snmp = target->snmpdata;
	REFERENCE *ref = NULL;

	if (snmp->reqid) {
		dprintf(("parse_table_reply(%s): scan=0x%x reqid=%d\n",
			 target->name, snmp->scan, snmp->reqid));

		snmp->reqid = 0; /* system request done */
		/*target->prev_reply = target->last_reply;*/
		gettimeofday(&target->last_reply, 0);
	}

	if (snmp->scan & SCAN_IFTABLE) {
		/*
		 * Retrieve the list of interface variables.
		 */
		if (!advance_iftable(&ref, target, method->snmp_version)) {
			INTERFACE *iface;
			snmp->scan &= ~SCAN_IFTABLE;
			/*
			 * End of interface tables.
			 * Make sure all interfaces was matched.
			 */
			for (iface = target->interface; iface; iface = iface->next) {
				if (!iface->tabIdx) {
					if (iface->name) {
						report(LOG_WARNING, "No \"%s\" found at \"%s\"",
						       iface->name, target->name);
					} else if (iface->index > 0) {
						report(LOG_WARNING, "No Interface%d found at \"%s\"",
						       iface->index, target->name);
					}
				}
			}
			free_var_list(snmp->ipTable);
			snmp->ipTable = 0;
			snmp->scan |= SCAN_IFADDR;
		}
	}

	if (snmp->scan & SCAN_IFADDR) {
		/*
		 * Retrieve the list of interface IP addresses.
		 */
		if (!advance_iptable(&ref, target, method->snmp_version)) {
			snmp->scan &= ~SCAN_IFADDR;
		}
	}

	if (snmp->scan & SCAN_BGPPEER) {
		/*
		 * Retrieve the list of bgp peer IP addresses.
		 */
		if (!advance_bgptable(&ref, target, method->snmp_version)) {
			BGP_AS *bgp;
			snmp->scan &= ~SCAN_BGPPEER;
			/*
			 * End of bgpPeer table.
			 * Make sure all ASN and Peers was matched.
			 */
			for (bgp = target->bgp; bgp; bgp = bgp->next) {
				BGP_PEER *peer = bgp->peer;
				if (!peer) {
					report(LOG_WARNING, "No BGP AS%d found at \"%s\"",
					       bgp->asn, target->name);
					continue;
				}
				for (; peer; peer = peer->next) {
					if (!peer->tabIdx) {
						char tmpbuf[100];
						report(LOG_WARNING, "No BGP Peer \"%s\" found at \"%s\" (AS%d)",
						       intoa(tmpbuf, peer->ip_addr),
						       target->name, bgp->asn);
						free_bgp_peers(bgp, 0);
						break;
					}
				}
			}
		}
	}

	if (snmp->scan & SCAN_ENVMON) {
		/*
		 * Retrieve the list of environment monitors.
		 */
		if (!advance_envtable(&ref, target, method->snmp_version)) {
			ENV_MON *env;
			snmp->scan &= ~SCAN_ENVMON;
			/*
			 * End of envMon table.
			 * Make sure all EnvMon Gauges was matched.
			 */
			for (env = target->env; env; env = env->next) {
				ENV_GAUGE *gauge = env->gauge;
				if (!gauge) {
					report(LOG_WARNING, "No %s found at \"%s\"",
					       get_name_by_type(env->type), target->name);
					continue;
				}
				for (; gauge; gauge = gauge->next) {
					if (!gauge->tabIdx) {
						report(LOG_WARNING,
						       "No %s Gauge \"%s\" found at \"%s\"",
						       get_name_by_type(env->type),
						       gauge->name,
						       target->name);
						free_env_gauges(env, 0);
						break;
					}
				}
			}
		}
	}

	if (ref) {
		SNMP_REQ *req = create_request(GETNEXT_REQUEST, ref);
		if (req && (snmp->reqid = session_start(sd, req)) != 0) {
			dprintf(("send_table_request(%s): scan=0x%x reqid=%d\n",
				 target->name, snmp->scan, snmp->reqid));
		}
	}
}

static void
batch_interface_requests(sd, target)
	SESSION *sd;
	OBJECT *target;
{
	METHOD *method = sd->method;
	SNMP_DATA *snmp = target->snmpdata;
	INTERFACE *iface;

	for (iface = snmp->next_batch; iface; iface = iface->next) {
		if (!iface->tabIdx)
			continue;
		if (method && snmp->batch_count >= method->router_batch)
			break;
		snmp->batch_count += send_interface_request(sd, target, iface);
	}
	snmp->next_batch = iface;
}

static int
get_device_id(objid, objid_len)
	oid *objid;
	int objid_len;
{
	/* null objid must be ignored */
	if (objid && objid_len > 2) {
		if (objid_len == ciscoProductID_len &&
		    !memcmp(objid, ciscoProductID, (objid_len - 1) * sizeof(oid))) {
			trace(("\t DeviceID: Cisco\n"));
			return DEVICE_CISCO;
		}

		if (objid_len == catalystProductID_len &&
		    !memcmp(objid, catalystProductID, (objid_len - 1) * sizeof(oid))) {
			trace(("\t DeviceID: Catalyst\n"));
			return DEVICE_CATALYST;
		}

		/* other device vendors would be matched here */
	}
	return DEVICE_UNKNOWN;
}

static void
replace_device_mib(var_list, prev_id, curr_id, ver)
	VARIABLE **var_list;
	int prev_id, curr_id;
	int ver;
{
	if (!var_list || prev_id == curr_id)
		return;

	/* remove old MIB */
	if (prev_id == DEVICE_CISCO) {
		free_mib_table(var_list, cisco_system_MIB);
	} else if (prev_id == DEVICE_CATALYST) {
		/* XXX not implemented yet!
		free_mib_table(var_list, catalyst_system_MIB);
		*/
	}

	/* add new MIB */
	if (curr_id == DEVICE_CISCO) {
		add_mib_table(var_list, cisco_system_MIB, ver);
	} else if (curr_id == DEVICE_CATALYST) {
		/* XXX not implemented yet!
		add_mib_table(var_list, catalyst_system_MIB, ver);
		*/
	}
}

static void
parse_system_reply(sd, target)
	SESSION *sd;
	OBJECT *target;
{
	METHOD *method = sd->method;
	SNMP_DATA *snmp = target->snmpdata;
	VARIABLE *var;
	int rescan = 0;

	/* sanity check */
	if (!method)
		return;

	if (snmp->reqid) {
		dprintf(("parse_system_reply(%s): mode=%06o reqid=%d\n",
			 target->name, snmp->mode, snmp->reqid));

		snmp->reqid = 0; /* system request done */
		/*target->prev_reply = target->last_reply;*/
		gettimeofday(&target->last_reply, 0);
	}

	snmp->mode &= ~GET_INFO;
	if ((snmp->mode & GET_STATE) == 0) {
		snmp->mode |= GET_STATE; /* get state next time */
		report(LOG_WARNING, "Has lost %s mode at \"%s\", recover...",
		       method->name, target->name);
		return;
	}

	/*
	 * Check for sysObjID changes.
	 */
	var = get_var(target->var_list, "sysObjID");
	if (var && var->type == OBJID && var->value.objid) {
		oid prev[MAX_OID_LEN];
		int prev_len = 0; /* bytes */
		if (var->old && var->old->type == OBJID &&
		    var->old->value.objid) {
			prev_len = var->old->length;
			if (prev_len > sizeof(prev)) prev_len = sizeof(prev);
			memcpy(prev, var->old->value.objid, prev_len);
		}
		if (var->length != prev_len ||
		    memcmp(var->value.objid, prev, var->length)) {
			char curr_str[512], prev_str[512];
			oid2str(curr_str, var->value.objid, var->length / sizeof(oid));
			oid2str(prev_str, prev, prev_len / sizeof(oid));
			trace(("\t sysObjID changed (%s->%s), rescan...\n",
			       prev_str, curr_str));
			if (!rescan && prev_len) {
				report(LOG_NOTICE, "$sysObjID changed (%s->%s) at \"%s\", rescan...",
				       prev_str, curr_str, target->name);
			}
			rescan++;
		}

		/*
		 * Further check for device id changes.
		 * Replace vendor-specific MIBs if required.
		 */
		prev_len = snmp->device;
		snmp->device = get_device_id(var->value.objid, var->length / sizeof(oid));
		if (snmp->device != prev_len) {
			trace(("\t DeviceID changed (%d->%d), rescan...\n",
			       prev_len, snmp->device));
			if (!rescan && prev_len != DEVICE_UNKNOWN) {
				report(LOG_NOTICE, "DeviceID changed (%d->%d) at \"%s\", rescan...",
				       prev_len, snmp->device, target->name);
			}
			replace_device_mib(&target->var_list, prev_len,
					   snmp->device, method->snmp_version);
			rescan++;
		}
	} else report(LOG_WARNING, "Cant get $sysObjID at \"%s\"", target->name);

	/*
	 * Check for sysUpTime to detect system reloading.
	 */
	var = get_var(target->var_list, "sysUpTime");
	if (var && var->type == UNSIGNED && var->value.integer) {
		u_long curr = *var->value.integer;
		u_long prev = 0;
		if (var->old && var->old->type == UNSIGNED &&
		    var->old->value.integer)
			prev = *var->old->value.integer;
		if (curr < prev) {
			trace(("\t sysUpTime lowered (%u->%u), rescan...\n"));
			if (!rescan && prev) {
				report(LOG_NOTICE, "$sysUpTime lowered (%u->%u) at \"%s\" (reloaded?), rescan...",
				       prev, curr, target->name);
			}
			rescan++;
		}
	} else report(LOG_WARNING, "Cant get $sysUpTime at \"%s\"", target->name);

	/*
	 * Check for ifNumber to detect ifTable changes.
	 */
	var = get_var(target->var_list, "ifNumber");
	if (var && var->type == INTEGER && var->value.integer) {
		int curr = *var->value.integer;
		int prev = 0;
		if (var->old && var->old->type == INTEGER &&
		    var->old->value.integer)
			prev = *var->old->value.integer;
		if (curr != prev) {
			trace(("\t ifNumber changed (%d->%d), rescan...\n",
			       prev, curr));
			if (!rescan && prev) {
				report(LOG_NOTICE, "$ifNumber changed (%d->%d) at \"%s\", rescan...",
				       prev, curr, target->name);
			}
			rescan++;
		}
	} else report(LOG_WARNING, "Cant get $ifNumber at \"%s\"", target->name);

	if (rescan) {
		session_stop(sd, 0);
		clear_requests(target);

		snmp->mode |= GET_INFO;
		snmp->mode |= GET_DATA;

		if (target->interface || target->ifgroup)
			snmp->scan |= SCAN_IFTABLE;
		if (target->bgp)
			snmp->scan |= SCAN_BGPPEER;
		if (target->env)
			snmp->scan |= SCAN_ENVMON;
	}
	snmp->scan |= snmp->pending_scan;
	snmp->pending_scan = 0;

	if (snmp->scan & SCAN_IFTABLE) {
		free_var_list(snmp->ifTable);
		snmp->ifTable = 0;
	} else {
		snmp->batch_count = 0;
		snmp->next_batch = target->interface;
		batch_interface_requests(sd, target);
	}
	if (snmp->scan & SCAN_BGPPEER) {
		free_var_list(snmp->bgpTable);
		snmp->bgpTable = 0;
	} else {
		BGP_AS *bgp;
		for (bgp = target->bgp; bgp; bgp = bgp->next) {
			if (bgp->asn > 0) {
				BGP_PEER *peer = bgp->peer;
				for (; peer; peer = peer->next) {
					if (peer->tabIdx)
						send_bgppeer_request(sd, target, peer);
				}
			}
		}
	}
	if (snmp->scan & SCAN_ENVMON) {
		free_var_list(snmp->envTable);
		snmp->envTable = 0;
	} else {
		ENV_MON *env;
		for (env = target->env; env; env = env->next) {
			send_envmon_request(sd, target, env);
		}
	}
	if (snmp->scan) /* start scan */
		parse_table_reply(sd, target);
}

static void
parse_interface_reply(sd, target, iface)
	SESSION *sd;
	OBJECT *target;
	INTERFACE *iface;
{
	SNMP_DATA *snmp = target->snmpdata;
	VARIABLE *var;

	if (iface->reqid) {
		dprintf(("parse_interface_reply(%s/%s): mode=%06o reqid=%d\n",
			 target->name, iface->name ? iface->name : itoa(iface->index),
			 iface->mode, iface->reqid));

		iface->reqid = 0; /* interface request done */
		iface->prev_reply = iface->last_reply;
		gettimeofday(&iface->last_reply, 0);
	}

	iface->mode &= ~GET_INFO;
	if ((iface->mode & GET_STATE) == 0) {
		iface->mode |= GET_STATE; /* get state next time */
		return;
	}
	iface->mode &= ~GET_DATA;

	/*
	 * First, make sure ifTable was not changed.
	 */
	for (var = iface->var_list; var; var = var->next) {
		if (!var->objid || (var->mode & SCAN_MATCH) == 0)
			continue;
		if (var->type == INTEGER && var->value.integer) {
			if (iface->index > 0 && iface->index != *var->value.integer)
				break;
		} else if (var->type == STRING && var->value.string) {
			if (iface->name && !match_ifdescr(iface->name, var->value.string))
				break;
		}
	}
	if (var) {
		trace(("\t %s not matched, rescan next time...\n", var->name));
		if (iface->name) {
			report(LOG_NOTICE, "$%s of \"%s\" not matched at \"%s\", rescan next time...",
			       var->name, iface->name, target->name);
		} else if (iface->index > 0) {
			report(LOG_NOTICE, "$%s of Interface%d not matched at \"%s\", rescan next time...",
			       var->name, iface->index, target->name);
		}
		iface->state = 0;
		iface->tabIdx = 0;
		snmp->pending_scan |= SCAN_IFTABLE;
		return;
	}

	/*
	 * Check for interface operation status.
	 */
	var = get_var(iface->var_list, "ifOperStatus");
	if (var && var->type == INTEGER && var->value.integer) {
		iface->prev_state = iface->state;
		iface->state = *var->value.integer;

		if (iface->state != iface->prev_state)
			iface->last_change = iface->last_reply;

		if (iface->state == STATE_UP)
			iface->mode |= GET_DATA;

		if (iface->prev_state && iface->state &&
		    (iface->prev_state != STATE_UP && iface->state == STATE_UP) ||
		    (iface->prev_state == STATE_UP && iface->state != STATE_UP)) {
			iface->mode &= ~GET_STATE;
			iface->mode |= GET_INFO;
			send_interface_request(sd, target, iface);
		}
	} else {
		report(LOG_WARNING, "Cant get $ifOperStatus of \"%s\" at \"%s\"",
		       iface->name ? iface->name : itoa(iface->index), target->name);
	}
}

static void
parse_bgppeer_reply(sd, target, peer)
	SESSION *sd;
	OBJECT *target;
	BGP_PEER *peer;
{	
	SNMP_DATA *snmp = target->snmpdata;
	BGP_AS *bgp = peer->parent;
	VARIABLE *var;
	char tmpbuf[100];

	/* sanity check */
	if (!snmp || !bgp) return;

	if (peer->reqid) {
		dprintf(("parse_bgppeer_reply(%s/AS%d): mode=%06o reqid=%d\n",
			 target->name, bgp->asn, peer->mode, peer->reqid));

		peer->reqid = 0; /* bgp peer request done */
		peer->prev_reply = peer->last_reply;
		gettimeofday(&peer->last_reply, 0);
	}

	peer->mode &= ~GET_INFO;
	if ((peer->mode & GET_STATE) == 0) {
		peer->mode |= GET_STATE; /* get state next time */
		return;
	}
	peer->mode &= ~GET_DATA;

	/*
	 * First, make sure ASN still present at the BGP peer.
	 */
	var = get_var(peer->var_list, "bgpPeerRemoteAs");
	if (var && var->type == INTEGER && var->value.integer) {
		if (bgp->asn != *var->value.integer) {
			trace(("\t bgpPeerRemoteAs not matched, rescan next time...\n"));
			report(LOG_NOTICE, "$bgpPeerRemoteAs of \"%s\" not matched at \"%s\", rescan next time...",
			       intoa(tmpbuf, peer->ip_addr), target->name);
			peer->state = 0;
			peer->tabIdx = 0;
			snmp->pending_scan |= SCAN_BGPPEER;
			return;
		}
	}

	/*
	 * Check for BGP Peer operation status.
	 */
	var = get_var(peer->var_list, "bgpPeerState");
	if (var && var->type == INTEGER && var->value.integer) {
		peer->prev_state = peer->state;
		peer->state = *var->value.integer;

		if (peer->state != peer->prev_state)
			peer->last_change = peer->last_reply;

		if (peer->state == BGP_ESTABLISHED)
			peer->mode |= GET_DATA;

		if (peer->prev_state && peer->state &&
		    (peer->prev_state != BGP_ESTABLISHED && peer->state == BGP_ESTABLISHED) ||
		    (peer->prev_state == BGP_ESTABLISHED && peer->state != BGP_ESTABLISHED)) {
			peer->mode &= ~GET_STATE;
			peer->mode |= GET_INFO;
			send_bgppeer_request(sd, target, peer);
		}
	} else {
		report(LOG_WARNING, "Cant get $bgpPeerState of \"%s\" at \"%s\" (AS%d)",
		       intoa(tmpbuf, peer->ip_addr), target->name, bgp->asn);
	}
}

static void
parse_envmon_reply(sd, target, env)
	SESSION *sd;
	OBJECT *target;
	ENV_MON *env;
{	
	VARIABLE *var;
	ENV_GAUGE *gauge;

	if (env->reqid) {
		dprintf(("parse_envmon_reply(%s/%s): mode=%06o reqid=%d\n",
			 target->name, get_name_by_type(env->type), env->mode, env->reqid));

		env->reqid = 0; /* envmon request done */
		env->prev_reply = env->last_reply;
		gettimeofday(&env->last_reply, 0);
	}

	env->mode &= ~GET_INFO;
	if ((env->mode & GET_STATE) == 0) {
		env->mode |= GET_STATE;
		return;
	}
	env->mode &= ~GET_DATA;

	for (gauge = env->gauge; gauge; gauge = gauge->next) {
		for (var = gauge->var_list; var; var = var->next) {
			if (!var->objid || (var->mode & GET_STATE) == 0)
				continue;
			if (var->type == INTEGER && var->value.integer) {
				gauge->prev_state = gauge->state;
				gauge->state = *var->value.integer;

				if (gauge->state != gauge->prev_state)
					env->last_change = env->last_reply;

				env->mode |= GET_DATA;
			}
		}
	}
}

static VARIABLE *
get_last_var(vlist, mode)
	VARIABLE *vlist;
	int mode;
{
	VARIABLE *var, *last = 0;
	for (var = vlist; var; var = var->next) {
		if ((var->mode & mode) == 0)
			continue;
		if (!var->mode)
			break;
		last = var;
	}
	return last;
}

static int
recover_after_error(sd, target, status)
	SESSION *sd;
	OBJECT *target;
	int status;
{
	SNMP_DATA *snmp = target->snmpdata;
	VARIABLE *var;

	/* is there variable name caused error? */
	if (status == NOSUCHNAME && sd && sd->data_ptr) {
		char *mibname = "", *badvar = (char *)sd->data_ptr;

		if ((var = get_var(snmp->ifTable, badvar)) != 0) {
			snmp->recover_scan |= SCAN_IFTABLE;
			mibname = "ifTable";
		} else if ((var = get_var(snmp->ipTable, badvar)) != 0) {
			snmp->recover_scan |= SCAN_IFADDR;
			mibname = "ipAddr";
		} else if ((var = get_var(snmp->bgpTable, badvar)) != 0) {
			snmp->recover_scan |= SCAN_BGPPEER;
			mibname = "bgpPeer";
		} else if ((var = get_var(snmp->envTable, badvar)) != 0) {
			snmp->recover_scan |= SCAN_ENVMON;
			mibname = "envMon";
		} else if ((var = get_var(target->var_list, badvar)) != 0) {
			snmp->recover_scan = 0;
			mibname = "system";
		}
		if (var) { /* cutoff at the variable */
			var->mode = 0;
			report(LOG_INFO, "No $%s found in %s MIB at \"%s\"",
			       var->name, mibname, target->name);
			if (snmp->recover_scan) {
				while (var) {
					var->mode = 0;
					var = var->next;
				}
			}
			return 1;
		}
	} else if (snmp->scan) {
		int cutoff = 0;
		if ((snmp->scan & SCAN_IFTABLE) != 0 &&
		    (var = get_last_var(snmp->ifTable, SCAN_DETECT)) != 0) {
			snmp->recover_scan |= SCAN_IFTABLE;
			var->mode = 0;
			report(LOG_INFO, "Cutoff $%s detecting at \"%s\"",
			       var->name, target->name);
			cutoff++;
		}
		if ((snmp->scan & SCAN_IFADDR) != 0 &&
		    (var = get_last_var(snmp->ipTable, SCAN_DETECT)) != 0) {
			snmp->recover_scan |= SCAN_IFADDR;
			var->mode = 0;
			report(LOG_INFO, "Cutoff $%s detecting at \"%s\"",
			       var->name, target->name);
			cutoff++;
		}
		if ((snmp->scan & SCAN_BGPPEER) != 0 &&
		    (var = get_last_var(snmp->bgpTable, SCAN_DETECT)) != 0) {
			snmp->recover_scan |= SCAN_BGPPEER;
			var->mode = 0;
			report(LOG_INFO, "Cutoff $%s detecting at \"%s\"",
			       var->name, target->name);
			cutoff++;
		}
		if ((snmp->scan & SCAN_ENVMON) != 0 &&
		    (var = get_last_var(snmp->envTable, SCAN_DETECT)) != 0) {
			snmp->recover_scan |= SCAN_ENVMON;
			var->mode = 0;
			report(LOG_INFO, "Cutoff $%s detecting at \"%s\"",
			       var->name, target->name);
			cutoff++;
		}
		return cutoff;
	}
	if ((var = get_last_var(target->var_list, MIB_GET_MODES)) != 0) {
		var->mode = 0;
		report(LOG_INFO, "Cutoff $%s collecting at \"%s\"",
		       var->name, target->name);
		return 1;
	}
	return 0;
}

static int
parse_router_reply(sd, reqid, target)
	SESSION *sd;
	int reqid;
	OBJECT *target;
{
	int status = sd->data_int;
	SNMP_DATA *snmp = target->snmpdata;
	INTERFACE *iface;
	BGP_PEER *peer;
	ENV_MON *env;
	VARIABLE *var;

	/* sanity check */
	if (!snmp) return -1;

	if (snmp->reqid == reqid) {
		if (status == NOERROR) {
			if (snmp->scan) {
				parse_table_reply(sd, target);
				return status;
			}
			if (snmp->mode & MIB_GET_MODES) {
				parse_system_reply(sd, target);
				return status;
			}
		} else if (recover_after_error(sd, target, status)) {
			snmp->reqid = 0; /* dummy */
			if (snmp->scan) {
				parse_table_reply(sd, target);
				return status;
			}
			if (snmp->mode & MIB_GET_MODES) {
				send_system_request(sd, target);
				return status;
			}
		}
	} else if ((iface = get_iface_by_reqid(reqid, target)) != 0) {
		if (status == NOERROR) {
			if (iface->mode & MIB_GET_MODES) {
				parse_interface_reply(sd, target, iface);

				snmp->batch_count--;
				if (snmp->batch_count < 1) {
					snmp->batch_count = 0;
					batch_interface_requests(sd, target);
				}
				return status;
			}
		} else if (status == NOSUCHNAME) {
			iface->state = 0;
			iface->tabIdx = 0;
			snmp->pending_scan |= SCAN_IFTABLE;

			if (iface->name) {
				report(LOG_WARNING, "No \"%s\" detected at \"%s\", rescan next time...",
				       iface->name, target->name);
			} else if (iface->index > 0) {
				report(LOG_WARNING, "No Interface%d detected at \"%s\", rescan next time...",
				       iface->index, target->name);
			}
			snmp->batch_count--;
			if (snmp->batch_count < 1) {
				snmp->batch_count = 0;
				batch_interface_requests(sd, target);
			}
			return status;
		} else if (status == TOOBIG &&
			   (var = get_last_var(iface->var_list, SCAN_DETECT)) != 0) {
			var->mode = 0;

			if (iface->name) {
				report(LOG_WARNING, "Cutoff $%s collecting on \"%s\" at \"%s\"",
				       var->name, iface->name, target->name);
			} else if (iface->index > 0) {
				report(LOG_WARNING, "Cutoff $%s collecting on Interface%d at \"%s\"",
				       var->name, iface->index, target->name);
			}
			return status;
		}
	} else if ((peer = get_bgppeer_by_reqid(reqid, target)) != 0) {
		char tmpbuf[100];
		if (status == NOERROR) {
			if (peer->mode & MIB_GET_MODES) {
				parse_bgppeer_reply(sd, target, peer);
				return status;
			}
		} else if (status == NOSUCHNAME) {
			peer->state = 0;
			peer->tabIdx = 0;
			snmp->pending_scan |= SCAN_BGPPEER;

			report(LOG_WARNING, "No BGP Peer \"%s\" detected at \"%s\" (AS%d), rescan next time...",
			       intoa(tmpbuf, peer->ip_addr), target->name,
			       peer->parent ? peer->parent->asn : 0);
			return status;
		} else if (status == TOOBIG &&
			   (var = get_last_var(peer->var_list, SCAN_DETECT)) != 0) {
			var->mode = 0;

			report(LOG_WARNING, "Cutoff $%s collecting on \"%s\" at \"%s\" (AS%d)",
			       var->name, intoa(tmpbuf, peer->ip_addr), target->name,
			       peer->parent ? peer->parent->asn : 0);
			return status;
		}
	} else if ((env = get_envmon_by_reqid(reqid, target)) != 0) {
		if (status == NOERROR) {
			if (env->mode & MIB_GET_MODES) {
				parse_envmon_reply(sd, target, env);
				return status;
			}
		} else if (status == NOSUCHNAME) {
			ENV_GAUGE *gauge = env->gauge;
			for (; gauge; gauge = gauge->next) {
				gauge->state = 0;
				gauge->tabIdx = 0;
			}
			snmp->pending_scan |= SCAN_ENVMON;

			report(LOG_WARNING, "No %s detected at \"%s\", rescan next time...",
			       get_name_by_type(env->type), target->name);
			return status;
		}
	} else {
		dprintf(("parse_router_reply(%s): Unexpected reqid=%d?\n",
			 target->name, reqid));
		return status;
	}
	dprintf(("Cant parse_router_reply(%s): mode=%06o reqid=%d status=%d\n",
		 target->name, snmp->mode, reqid, status));

	if (status < 0 || status >= NUM_ERRORS)
		status = GENERR;
	report(LOG_ERR, "Unexpected SNMP reply from \"%s\" (mode=%06o reqid=%d): %s",
	       target->name, snmp->mode, reqid, snmp_error[status]);
	return status;
}

static void
router_stop(method)
	METHOD *method;
{
	SESSION *sd = method->sd;

	/* sanity check */
	if (!sd) return;

	session_stop(sd, 1);

	/* clear all requests on the router */
	if (sd->owner)
		clear_requests(sd->owner);
}

static void
router_start(method)
	METHOD *method;
{
	SESSION *sd = method->sd;

	/* sanity check */
	if (!sd) {
		trace(("\t DATA ERROR at router_start()\n"));
		report(LOG_ERR, "router_start(): internal error!?");
		return;
	}

	router_stop(method); /* stop the method if still running */

	/* open datagram socket near before UDP session */
	if ((sd->sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		router_reply(errno, sd, 0);
		return;
	}

	/* turn on non-blocking I/O */
	if (set_socket_async(sd->sock, TRUE) < 0) {
		router_reply(errno, sd, 0);
		return;
	}

	/*
	 * Make datagram socket `connected', so error on this socket
	 * will be returned asynchronously without timing out.
	 */
	if (connect(sd->sock, &sd->peer, sizeof(struct sockaddr)) < 0) {
		router_reply(errno, sd, 0);
		return;
	}

	send_router_request(sd, sd->owner);
}

static void
router_reply(op, sd, reqid)
	int op;
	SESSION *sd;
	int reqid;
{
	OBJECT *target = sd->owner;
	METHOD *method = sd->method;
	char *diag, buf[1024];

	/* sanity check */
	if (!method || !target) {
		trace(("\t DATA ERROR at router_reply()\n"));
		report(LOG_ERR, "router_reply(): internal error!?");
		return;
	}

	if (!op) { /* reply was received */
		op = parse_router_reply(sd, reqid, target);
		if (session_busy(sd))
			return; /* awainting session to be finished */
		if (op >= 0 && op < NUM_ERRORS)
			diag = snmp_error[op];
		else	diag = "Unknown error";
		if (op != NOERROR &&
		    sd->data_ptr && strlen(sd->data_ptr) < 100) {
			(void)strcpy(buf, sd->data_ptr);
			(void)strcat(buf, ": ");
			(void)strcat(buf, diag);
			diag = buf;
		}
	} else {
		INTERFACE *iface = get_iface_by_reqid(reqid, target);
		diag = strerror(op);
		if (iface) {
			if (iface->name) {
				report(LOG_INFO, "\"%s\" at \"%s\": %s",
				       iface->name, target->name, diag);
			} else if (iface->index > 0) {
				report(LOG_INFO, "Interface%d at \"%s\": %s",
				       iface->index, target->name, diag);
			}
		}
	}
	dprintf(("router_reply(%s): reqid=%d diag=\"%s\"\n",
		 target->name, reqid, diag));

	router_stop(method);

	method_finished(target, method, diag, !op);
}

static int
router_init(target, method)
	OBJECT *target;
	METHOD *method;
{
	SESSION template;
	struct sockaddr_in *to;

	dprintf(("router_init(%s/%s)\n", target->name, method->name));

	if (method->sd)
		session_free(method->sd);

	memset(&template, 0, sizeof(template));
	template.owner = target;
	template.method = method;
	template.sock = -1;	/* not yet opened */
	to = (struct sockaddr_in *)&template.peer;
	to->sin_family = AF_INET;
	to->sin_port = htons(method->rport);
	to->sin_addr = method->address ? method->ip_addr : target->ip_addr;
	template.timeout = method->timeout * 1000000L; /* make microseconds */
	template.retries = method->retries;
	template.send = snmp_send;
	template.recv = snmp_recv;
	template.read = router_reply;

	if ((method->sd = session_create(&template)) == NULL)
		return -1;

	if (!target->snmpdata) {
		if ((target->snmpdata = (SNMP_DATA *)malloc(sizeof(SNMP_DATA))) == NULL)
			return -1;
		memset(target->snmpdata, 0, sizeof(SNMP_DATA));
	} else {
		target->snmpdata->scan = 0;
		target->snmpdata->mode = 0;
		target->snmpdata->reqid = 0;
	}
	add_mib_table(&target->var_list, system_MIB, method->snmp_version);

	return 0;
}

/*
 * Built-in method initiator block.
 */

METHOD	method_router = {
	router_init,	/* initialize router session */
	router_start,	/* start SNMP session on the router */
	router_stop,	/* stop SNMP session on the router */

	"ROUTER",	/* method name */
	"public",	/* SNMP community by default */
	NULL, { 0 },	/* use target address by default */
	IPPROTO_UDP,	/* network protocol */
	SNMPSERVER_PORT,/* server port */
	0, 0,		/* timeout and retries undefined yet */
	{ SNMP_VERSION_1, /* version number */
	  BATCH_DEFAULT },/* batch value */

	/* Non-initialized data */
	NULL,		/* session descriptor */
	NULL,		/* chat-script */
	NULL,		/* snmp-request */
	NULL		/* next method */
};
