/*
 *	Copyright (c) 2000-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <syslog.h>

#include "threads.h"
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif
#include "netmon.h"

#define	SAVE_WHEN_VAR		"WHEN"

SAVE save_router = {
	"ROUTER",		/* save name */
	SAVE_FILE,		/* save method */
	"%Y.%m.%d",		/* file name to write */
"$time\
 $name\
 $sysUpTime\
 $ifNumber\
 $ciscoFreeMem.ave\
 $ciscoAvgBusy.ave",		/* data saving output format */
"$time\
 $name\
 $state\
 $ciscoWhyReload",		/* change state output format */
	NULL, 0, NULL,		/* when condition true output format */
	0, NULL, NULL		/* unused remainder (non-initialized data) */
};

SAVE save_interface = {
	"INTERFACE",
	SAVE_FILE,
	"%Y.%m.%d",
"$time\
 $name\
 $ifInOctets.delta\
 $ifOutOctets.delta\
 $ifInUcastPkts.delta\
 $ifOutUcastPkts.delta\
 $ifInDrops.delta\
 $ifOutDrops.delta\
 $ifInErrors.delta\
 $ifOutErrors.delta",
"$time\
 $name\
 $state\
 $ciscoIfReason",
"$ifSpeed.old && $ifSpeed != $ifSpeed.old", 0,
"$time\
 $name\
 BW $ifSpeed.old -> $ifSpeed",
	0, NULL, NULL		/* unused remainder (non-initialized data) */
};

SAVE save_bgppeer = {
	"BGPPEER",
	SAVE_FILE,
	"%Y.%m.%d",
	NULL,
"$time\
 $name\
 $state",
	NULL, 0, NULL,
	0, NULL, NULL		/* unused remainder (non-initialized data) */
};

SAVE save_envtemp = {
	"ENVTEMP",
	SAVE_FILE,
	"%Y.%m.%d",
	NULL,
"$time\
 $name\
 $state\
 $ciscoEnvTempValue.ave",
	NULL, 0, NULL,
	0, NULL, NULL		/* unused remainder (non-initialized data) */
};

SAVE save_envvolt = {
	"ENVVOLT",
	SAVE_FILE,
	"%Y.%m.%d",
	NULL,
"$time\
 $name\
 $state\
 $ciscoEnvVoltValue.ave",
	NULL, 0, NULL,
	0, NULL, NULL		/* unused remainder (non-initialized data) */
};

void
reset_statistic(var, need_free)
	register VARIABLE *var;
	int need_free;
{
	for (; var; var = var->next) {
		if (var->statistic) {
			if (need_free) {
				free(var->statistic);
				var->statistic = NULL;
			} else	memset(var->statistic, 0, sizeof(STATISTIC));
		}
	}
}

void
set_none_state(target)
	OBJECT *target;
{
	INTERFACE *iface;
	BGP_AS *bgp;
	BGP_PEER *peer;
	ENV_MON *env;
	ENV_GAUGE *gauge;
	OBJECT *service;
	VARIABLE *var;

	for (iface = target->interface; iface; iface = iface->next) {
		iface->state = 0;
		reset_statistic(iface->var_list, 1);
	}

	for (bgp = target->bgp; bgp; bgp = bgp->next) {
		for (peer = bgp->peer; peer; peer = peer->next) {
			peer->state = 0;
			reset_statistic(peer->var_list, 1);
		}
	}

	for (env = target->env; env; env = env->next) {
		for (gauge = env->gauge; gauge; gauge = gauge->next) {
			gauge->state = 0;
			reset_statistic(gauge->var_list, 1);
		}
	}

	for (service = target->service; service; service = service->next) {
		service->state = 0;
		reset_statistic(service->var_list, 1);
	}
}

static void
make_save(save, args)
	SAVE *save;
	char *args;
{
	char *file, buf[1024];

	if (strchr(save->file, '%')) {
		TIMEVAL tv;
		gettimeofday(&tv, NULL);
#ifdef	HAVE_PTHREAD
		pthread_mutex_lock(&localtime_lock);
#endif
		(void)strftime(buf, sizeof(buf), save->file, localtime((time_t *)&tv.tv_sec));
#ifdef	HAVE_PTHREAD
		pthread_mutex_unlock(&localtime_lock);
#endif
		file = buf;
	} else	file = save->file;

	if (save->method == SAVE_FILE) {
		FILE *fp;
		if ((fp = fopen(file, "a")) != NULL) {
			fprintf(fp, "%s\n", args);
			fclose(fp);
		} else	report(LOG_ERR, "fopen: %s: %m", file);
	} else	execfile(save->method == SAVE_PIPE, file, args);
}

static void
save_data(save, type, object)
	register SAVE *save;
	int type;
	void *object;
{
	char buf[4096];

	for (; save; save = save->next) {
		if (save->file && save->saving) {
			insert_variables(buf, sizeof(buf), save->saving, type, object);
			make_save(save, buf);
		}
	}
}

static void *
save_task_state(arg)
	void *arg;
{
	SAVE *save = (SAVE *)arg;
	char buf[4096];

#ifdef	HAVE_PTHREAD
	busy_threads++;
#endif
	insert_variables(buf, sizeof(buf), save->state, save->type, save->object);
	make_save(save, buf);
#ifdef	HAVE_PTHREAD
	busy_threads--;
#endif
}

void
save_state(save, type, object)
	register SAVE *save;
	int type;
	void *object;
{
#ifdef	HAVE_PTHREAD
	pthread_t tid;
#endif
	for (; save; save = save->next) {
		if (save->file && save->state) {
			save->type = type;
			save->object = object;
#ifdef	HAVE_PTHREAD
			if (type == TYPE_BGPPEER ||
			    type == TYPE_ENVTEMP ||
			    type == TYPE_ENVVOLT ||
			    type == TYPE_ENVFAN ||
			    type == TYPE_ENVPS) {
				pthread_create(&tid, &joinable_attr, save_task_state, (void *)save);
				pthread_join(tid, NULL);
			} else	pthread_create(&tid, &detached_attr, save_task_state, (void *)save);
#else
			save_task_state(save);
#endif
		}
	}
}

static void
save_when_var(type, object, str)
	int type;
	void *object;
	char *str;
{
	char *name = SAVE_WHEN_VAR;
	VARIABLE *var = 0;

	if (!object) return;

	switch (type) {
	case TYPE_OBJECT:
	case TYPE_SERVICE:
		var = add_var(&((OBJECT *)object)->var_list, name);
		break;
	case TYPE_INTERFACE:
		var = add_var(&((INTERFACE *)object)->var_list, name);
		break;
	case TYPE_BGPPEER:
		var = add_var(&((BGP_PEER *)object)->var_list, name);
		break;
	case TYPE_ENVTEMP:
	case TYPE_ENVVOLT:
	case TYPE_ENVFAN:
	case TYPE_ENVPS:
		var = add_var(&((ENV_GAUGE *)object)->var_list, name);
		break;
	}

	if (var && (str || var->value.integer))
		set_var(var, STRING, str, str ? strlen(str) : 0);
}

static void *
_save_task_when(arg)
	void *arg;
{
	SAVE *save = (SAVE *)arg;
	char buf[4096];

#ifdef	HAVE_PTHREAD
	busy_threads++;
#endif
	insert_variables(buf, sizeof(buf), save->when_fmt, save->type, save->object);
	save_when_var(save->type, save->object, buf);
	make_save(save, buf);
#ifdef	HAVE_PTHREAD
	busy_threads--;
#endif
}

static void
save_task_when(arg)
	void *arg;
{
#ifdef	HAVE_PTHREAD
	pthread_t tid;

	pthread_create(&tid, &detached_attr, _save_task_when, arg);
#else
	_save_task_when(arg);
#endif
}

void
save_when(save, type, object)
	register SAVE *save;
	int type;
	void *object;
{
	int matched, scheduled, save_cnt = 0, match_cnt = 0;
	TIMEVAL tv_buf, tv_task, *tv = NULL;
	char buf[4096];
#ifdef	HAVE_PTHREAD
	pthread_t tid;
#endif
	for (; save; save = save->next) {
		if (!save->file || !save->when) continue;
		save_cnt++;

		/* calculate condition */
		insert_variables(buf, sizeof(buf), save->when, type, object);
		matched = (calculate(buf) ? 1 : 0);
		match_cnt += matched;

		if (save->delay > 0) {
			scheduled = find_event(save_task_when, save);

			if (matched && !scheduled) {
				/* schedule new task */
				if (!tv) {
					gettimeofday(&tv_buf, NULL);
					tv = &tv_buf;
				}
				tv_task = *tv;
				tv_task.tv_sec += save->delay;
				save->type = type;
				save->object = object;
				add_event(&tv_task, save_task_when, save);

			} else if (!matched && scheduled) {
				/* condition has changed */
				remove_event(save_task_when, save);
			} /* no such conditions or condition still alive */

		} else if (matched) {
			save->type = type;
			save->object = object;
#ifdef	HAVE_PTHREAD
			pthread_create(&tid, &detached_attr, _save_task_when, (void *)save);
#else
			_save_task_when(save);
#endif
		}
	}
	if (save_cnt && !match_cnt) /* reset variable to empty value */
		save_when_var(type, object, NULL);
}

static void *
_save_object_data(arg)
	void *arg;
{
	OBJECT *target = (OBJECT *)arg;
	OBJECT *service;
	INTERFACE *iface;
	BGP_AS *bgp;
	BGP_PEER *peer;
	ENV_MON *env;
	ENV_GAUGE *gauge;

	/* sanity check */
	if (!target) return;

	dprintf(("save_object_data(%s)\n", target->name));

#ifdef	HAVE_PTHREAD
	busy_threads++;
#endif
	if (target->state == STATE_UP) {
		save_data(target->save_list, target->parent ? TYPE_SERVICE : TYPE_OBJECT, target);
		reset_statistic(target->var_list, 0);
	}

	for (iface = target->interface; iface; iface = iface->next) {
		if (iface->state == STATE_UP) {
			save_data(iface->save_list, TYPE_INTERFACE, iface);
			reset_statistic(iface->var_list, 0);
		}
	}

	for (bgp = target->bgp; bgp; bgp = bgp->next) {
		for (peer = bgp->peer; peer; peer = peer->next) {
			if (peer->state == BGP_ESTABLISHED) {
				save_data(bgp->save_list, TYPE_BGPPEER, peer);
				reset_statistic(peer->var_list, 0);
			}
		}
	}

	for (env = target->env; env; env = env->next) {
		for (gauge = env->gauge; gauge; gauge = gauge->next) {
			if (gauge->state && gauge->state != ENV_NOTPRESENT) {
				save_data(env->save_list, env->type, gauge);
				reset_statistic(gauge->var_list, 0);
			}
		}
	}

	for (service = target->service; service; service = service->next) {
		if (service->state == STATE_UP) {
			save_data(service->save_list, TYPE_SERVICE, service);
			reset_statistic(service->var_list, 0);
		}
	}
#ifdef	HAVE_PTHREAD
	busy_threads--;
#endif
}

void
save_object_data(arg)
	void *arg;
{
	OBJECT *target = (OBJECT *)arg;
	TIMEVAL tv;
#ifdef	HAVE_PTHREAD
	pthread_t tid;

	pthread_create(&tid, &detached_attr, _save_object_data, arg);
#else
	_save_object_data(arg);
#endif
	/* keep saving if required & non synchronized with polling */
	if (target->saving > 0 && !target->sync) {
		gettimeofday(&tv, NULL);
		tv_sharp(&tv, target->saving, object_offset(target, target->saving));
		add_event(&tv, save_object_data, target);
	}
}

void
save_object_state(arg)
	void *arg;
{
	OBJECT *target = (OBJECT *)arg;
	INTERFACE *iface;
	BGP_AS *bgp;
	BGP_PEER *peer;
	ENV_MON *env;
	ENV_GAUGE *gauge;

	/* sanity check */
	if (!target) return;

#ifdef	DONT_SAVE_ON_STARTUP
	if (!timerisset(&target->prev_request)) return;
#endif

	dprintf(("save_object_state(%s)\n", target->name));

	if (target->parent) { /* subobject? */
		if (target->state && target->state != target->prev_state) {
			save_state(target->save_list, TYPE_SERVICE, target);
		}
		save_when(target->save_list, TYPE_SERVICE, target);

		return;
	}

	if (target->state && target->state != target->prev_state) {
		save_state(target->save_list, TYPE_OBJECT, target);
	}
	save_when(target->save_list, TYPE_OBJECT, target);

	for (iface = target->interface; iface; iface = iface->next) {
		if (iface->state && iface->state != iface->prev_state) {
			save_state(iface->save_list, TYPE_INTERFACE, iface);
		}
		save_when(iface->save_list, TYPE_INTERFACE, iface);
	}

	for (bgp = target->bgp; bgp; bgp = bgp->next) {
		for (peer = bgp->peer; peer; peer = peer->next) {
			if (peer->state && peer->state != peer->prev_state) {
				save_state(bgp->save_list, TYPE_BGPPEER, peer);
			}
			save_when(bgp->save_list, TYPE_BGPPEER, peer);
		}
	}

	for (env = target->env; env; env = env->next) {
		for (gauge = env->gauge; gauge; gauge = gauge->next) {
			if (gauge->state && gauge->state != gauge->prev_state) {
				save_state(env->save_list, env->type, gauge);
			}
			save_when(env->save_list, env->type, gauge);
		}
	}
}
