%{
/*
 *	Copyright (c) 1999,2000 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#ifdef	__STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"
#include "y.tab.h"

#define	yylval	parseconf_lval
#define	yylex	parseconf_lex
#define	yyerror	parseconf_error

extern int errno;

#define MAX_DEPTH 10

static struct {
	FILE *fp;
	char *name;
	int lineno;
	YY_BUFFER_STATE stack;
} include[MAX_DEPTH];
static int depth = 0;

static char buf[1024];

void
#ifdef	__STDC__
yyerror(const char *fmt, ...)
#else
yyerror(fmt, va_alist)
	char *fmt;
	va_dcl
#endif
{
	char *file;
	va_list ap;
#ifdef	__STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	if (depth)
		file = include[depth-1].name;
	else	file = config_file;
	if (!syslog_initialized) {
		fprintf(stderr, "%s: line %d: ", file, yylineno);
		vfprintf(stderr, fmt, ap);
		fputc('\n', stderr);
		fflush(stderr);
	} else {
		char ebuf[1024];
		vsprintf(ebuf, fmt, ap);
		syslog(LOG_ERR, "%s: line %d: %s", file, yylineno, ebuf);
	}
	va_end(ap);
}

%}

%option never-interactive
%option yylineno

/* token names */
ROOTDIR		[Rr]oot[Dd]ir
TIMEFMT		[Tt]ime[Ff]mt
POLLING		[Pp]olling
SAVING		[Ss]aving
TIMEOUT		[Tt]imeout
RETRIES		[Rr]etries

GROUP		[Gg]roup
PERMIT		[Pp]ermit
DENY		[Dd]eny

NETSTATE	[Nn]et[Ss]tate
PORT		[Pp]ort

SAVE		[Ss]ave
FILE		[Ff]ile
EXEC		[Ee]xec
PIPE		[Pp]ipe
DATA		[Dd]ata
STATE		[Ss]tate
WHEN		[Ww]hen

OBJECT		[Oo]bject
ADDRESS		[Aa]ddress
DESCRIPTION	[Dd]escription|[Cc]omment
SERVICE		[Ss]ervice
INTERFACE	[Ii]nterface
DATADIR		[Dd]ata[Dd]ir
BANDWIDTH	[Bb]and[Ww]idth
BGP		[Bb][Gg][Pp]
PEER		[Pp]eer
ENV		[Ee]nv
TEMP		[Tt]emp
VOLT		[Vv]olt
FAN		[Ff]an
PS		PS|ps
GAUGE		[Gg]auge

METHOD		[Mm]ethod
TCP		[Tt][Cc][Pp]
UDP		[Uu][Dd][Pp]
ICMP		[Ii][Cc][Mm][Pp]
LOCALPORT	[Ll]ocal[Pp]ort
CHATSCRIPT	[Cc]hat[Ss]cript
SEND		[Ss]end
EXPECT		[Ee]xpect
SIZE		[Ss]ize
BATCH		[Bb]atch
PING		PING|[Pp]ing
SNMP		SNMP|[Ss]nmp
ROUTER		ROUTER|[Rr]outer
DNS		DNS|[Dd]ns
RADIUS		RADIUS|[Rr]adius
TACACS		TACACS|[Tt]acacs
V2		[Vv]2

TRAP		[Tt]rap
SOURCECHECK	[Ss]ource[Cc]heck
COMMUNITY	[Cc]ommunity
ENTERPRISE	[Ee]nterprise
SPECIFIC	[Ss]pecific
GENERIC		GENERIC|[Gg]eneric

LETTER		[a-zA-Z]
DIGIT		[0-9]

%x QUOTEDSTR ESCAPEDCHAR INCLUDE

%%

[ \t\f\r\n]+	/* ignore whitespaces */ ;

"/*"		{
			register int ch;
			for (;;) {
				while ((ch = input()) != '*') {
					if (ch == EOF) break;
#ifdef	OLD_LEX
					else if (ch == '\n') yylineno++;
#endif
				}
				if (ch == '*') {
					while ((ch = input()) == '*') ;
					if (ch == '/') break;
#ifdef	OLD_LEX
					else if (ch == '\n') yylineno++;
#endif
				}
				if (ch == EOF) {
					yyerror("unexpected EOF in comment");
					break;
				}
			}
		}

#.*		/* ignore comments */ ;

{ROOTDIR}	{	return TOKEN_ROOTDIR;	}

{TIMEFMT}	{	return TOKEN_TIMEFMT;	}

{POLLING}	{	return TOKEN_POLLING;	}

{SAVING}	{	return TOKEN_SAVING;	}

{TIMEOUT}	{	return TOKEN_TIMEOUT;	}

{RETRIES}	{	return TOKEN_RETRIES;	}

{GROUP}		{	return TOKEN_GROUP;	}

{PERMIT}	{	return TOKEN_PERMIT;	}

{DENY}		{	return TOKEN_DENY;	}

{NETSTATE}	{	return TOKEN_NETSTATE;	}

{PORT}		{	return TOKEN_PORT;	}

{SAVE}		{	return TOKEN_SAVE;	}

{FILE}		{	return TOKEN_FILE;	}

{EXEC}		{	return TOKEN_EXEC;	}

{PIPE}		{	return TOKEN_PIPE;	}

{DATA}		{	return TOKEN_DATA;	}

{STATE}		{	return TOKEN_STATE;	}

{WHEN}		{	return TOKEN_WHEN;	}

{OBJECT}	{	return TOKEN_OBJECT;	}

{ADDRESS}	{	return TOKEN_ADDRESS;	}

{DESCRIPTION}	{	return TOKEN_DESCRIPTION; }

{SERVICE}	{	return TOKEN_SERVICE;	}

{INTERFACE}	{	return TOKEN_INTERFACE;	}

{DATADIR}	{	return TOKEN_DATADIR;	}

{BANDWIDTH}	{	return TOKEN_BANDWIDTH;	}

{BGP}		{	return TOKEN_BGP;	}

{PEER}		{	return TOKEN_PEER;	}

{ENV}		{	return TOKEN_ENV;	}

{TEMP}		{	return TOKEN_TEMP;	}

{VOLT}		{	return TOKEN_VOLT;	}

{FAN}		{	return TOKEN_FAN;	}

{PS}		{	return TOKEN_PS;	}

{GAUGE}		{	return TOKEN_GAUGE;	}

{METHOD}	{	return TOKEN_METHOD;	}

{TCP}		{	return TOKEN_TCP;	}

{UDP}		{	return TOKEN_UDP;	}

{ICMP}		{	return TOKEN_ICMP;	}

{LOCALPORT}	{	return TOKEN_LOCALPORT;	}

{CHATSCRIPT}	{	return TOKEN_CHATSCRIPT; }

{SEND}		{	return TOKEN_SEND;	}

{EXPECT}	{	return TOKEN_EXPECT;	}

{SIZE}		{	return TOKEN_SIZE;	}

{BATCH}		{	return TOKEN_BATCH;	}

{PING}		{	return METHOD_PING;	}

{SNMP}		{	return METHOD_SNMP;	}

{ROUTER}	{	return METHOD_ROUTER;	}

{DNS}		{	return METHOD_DNS;	}

{RADIUS}	{	return METHOD_RADIUS;	}

{TACACS}	{	return METHOD_TACACS;	}

{V2}		{	return TOKEN_V2;	}

{TRAP}		{	return TOKEN_TRAP;	}

{SOURCECHECK}	{	return TOKEN_SOURCECHECK; }

{COMMUNITY}	{	return TOKEN_COMMUNITY;	}

{ENTERPRISE}	{	return TOKEN_ENTERPRISE; }

{SPECIFIC}	{	return TOKEN_SPECIFIC;	}

{GENERIC}	{	return TRAP_GENERIC;	}

\${LETTER}({LETTER}|{DIGIT}|_)+(\.{DIGIT}+)*	{
			yylval.string = &yytext[1];
			return VAR_NAME;
		}

{DIGIT}+(\.{DIGIT}+){4,}	{
			yylval.string = yytext;
			return VAR_OID;
		}

{DIGIT}+	{
			yylval.number = atoi(yytext);
			return TOKEN_NUMBER;
		}

\"		{
			buf[0] = '\0';
			BEGIN QUOTEDSTR;
		}

<QUOTEDSTR>[^"\n\\]*["\n\\]	{
			switch(yytext[yyleng-1]) {
				case '\n':
					yyerror("unterminated string");
				case '"':
					(void)strncat(buf, yytext, yyleng-1);
					yylval.string = buf;
					BEGIN INITIAL;
					return TOKEN_STRING;
				case '\\':
					(void)strncat(buf, yytext, yyleng-1);
					BEGIN ESCAPEDCHAR;
					break;
				default:
					yyerror("malformed input");
					BEGIN INITIAL;
			}
		}

<ESCAPEDCHAR>.|\n	{
			switch(yytext[0]) {
				case 'r':
					yytext[0] = '\r';
					break;
				case 'n':
					yytext[0] = '\n';
					break;
				case 't':
					yytext[0] = '\t';
					break;
				case '"':
					yytext[0] = '"';
					break;
				case '\n':
					yytext[0] = '\\';
					unput('\n');
			}
			(void)strncat(buf, &yytext[0], 1);
			BEGIN QUOTEDSTR;
		}

^%[ \t]*include		BEGIN INCLUDE;

<INCLUDE>[ \t]+		/* eat the whitespace */
<INCLUDE>[^ \t\n]+	{ /* got the include file name */
			if (depth < MAX_DEPTH) {
				FILE *fp;
				char *cp;
				if (yytext[0] != '/') {
					(void)strcpy(buf, config_file);
					if ((cp = strrchr(buf, '/')) != NULL)
						(void)strcpy(++cp, yytext);
					else	(void)strcpy(buf, yytext);
					cp = buf;
				} else	cp = yytext;
				if ((fp = fopen(cp, "r")) != NULL) {
					include[depth].fp = fp;
					include[depth].name = strdup(cp);
					include[depth].lineno = yylineno;
					include[depth].stack = YY_CURRENT_BUFFER;
					yy_switch_to_buffer(yy_create_buffer(fp, YY_BUF_SIZE));
					yylineno = 1;
					depth++;
				} else	yyerror("can't include \"%s\": %s",
						cp, strerror(errno));
			} else	yyerror("Includes nested too deeply");
			BEGIN INITIAL;
		}

.|\n		{	return yytext[0];	}

<<EOF>>		{
			if (depth > 0) {
				depth--;
				yy_delete_buffer(YY_CURRENT_BUFFER);
				yy_switch_to_buffer(include[depth].stack);
				yylineno = include[depth].lineno;
				fclose(include[depth].fp);
				free(include[depth].name);
			} else {
				yyterminate();
			}
		}

%%

int
yywrap()
{
 	return 1;
}
