/*
 *	Copyright (c) 1999-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/*
 * SNMP specification according to RFC1157.
 */

#define	SNMP_VERSION_1		0
#define	SNMP_VERSION_2		1
#define	SNMPSERVER_PORT		161
#define	SNMPTRAP_PORT		162
#define	MAX_PACKETSZ		4096	/* suggested 484 */

/* ASN.1 types */
#define	ASN_BOOLEAN		0x01
#define	ASN_INTEGER		0x02
#define	ASN_BIT_STR		0x03
#define	ASN_OCTET_STR		0x04
#define	ASN_NULL		0x05
#define	ASN_OBJECT_ID		0x06
#define	ASN_SEQUENCE		0x10
#define	ASN_SET			0x11

#define	ASN_UNIVERSAL		0x00
#define	ASN_APPLICATION		0x40
/* extended ASN types */
#define	 ASN_IPADDRESS		(ASN_APPLICATION | 0)
#define	 ASN_COUNTER		(ASN_APPLICATION | 1)
#define	 ASN_GAUGE		(ASN_APPLICATION | 2)
#define	 ASN_UNSIGNED		ASN_GAUGE	/* RFC 1902 - same as GAUGE */
#define	 ASN_TIMETICKS		(ASN_APPLICATION | 3)
#define	 ASN_OPAQUE		(ASN_APPLICATION | 4)
/* RFC1442 */
#define  ASN_NSAP		(ASN_APPLICATION | 5)	/* obsolete */
#define  ASN_COUNTER64		(ASN_APPLICATION | 6)
#define  ASN_UINTEGER		(ASN_APPLICATION | 7)	/* obsolete */
#define  ASN_FLOAT		(ASN_APPLICATION | 8)
#define  ASN_DOUBLE		(ASN_APPLICATION | 9)

#define	ASN_CONTEXT		0x80
#define	ASN_PRIVATE		0xc0

#define	ASN_PRIMITIVE		0x00
#define	ASN_CONSTRUCTOR		0x20

#define	ASN_LONG_LEN		0x80
#define ASN_EXTENSION_ID	0x1f
#define ASN_BIT8		0x80

/* exception types V2 */
#define	ASN_NOSUCHOBJECT	(ASN_CONTEXT | ASN_PRIMITIVE | 0x0)
#define	ASN_NOSUCHINSTANCE	(ASN_CONTEXT | ASN_PRIMITIVE | 0x1)
#define	ASN_ENDOFMIBVIEW	(ASN_CONTEXT | ASN_PRIMITIVE | 0x2)

#define	IS_CONSTRUCTOR(byte)	((byte) & ASN_CONSTRUCTOR)
#define	IS_EXTENSION_ID(byte)	(((byte) & ASN_EXTENSION_ID) == ASN_EXTENSION_ID)

/* PDU types */
#define GET_REQUEST		(ASN_CONTEXT | ASN_CONSTRUCTOR | 0x0)
#define GETNEXT_REQUEST		(ASN_CONTEXT | ASN_CONSTRUCTOR | 0x1)
#define GET_RESPONSE		(ASN_CONTEXT | ASN_CONSTRUCTOR | 0x2)
#define	SET_REQUEST		(ASN_CONTEXT | ASN_CONSTRUCTOR | 0x3)
#define TRAP_PDU		(ASN_CONTEXT | ASN_CONSTRUCTOR | 0x4)

/* error status */
#define NOERROR			0
#define TOOBIG			1
#define NOSUCHNAME		2
#define BADVALUE		3
#define READONLY		4
#define GENERR			5
/* extended error codes (V2) */
#define	NO_ACCESS		6
#define WRONG_TYPE		7
#define WRONG_LENGTH		8
#define WRONG_ENCODING		9
#define WRONG_VALUE		10
#define NO_CREATION		11
#define INCONSISTENT_VALUE	12
#define RESOURCE_UNAVAILABLE	13
#define COMMIT_FAILED		14
#define UNDO_FAILED		15
#define AUTHORIZATION_ERROR	16
#define NOT_WRITEABLE		17
#define INCONSISTENT_NAME	18

#define	NUM_ERRORS		19

/* max reasonable size of some variables */
#define	MAX_STR_LEN	1024
#define	MAX_OID_LEN	128

/*
 * The standart traps.
 */
#define	TRAP_COLDSTART		0 /* sysUpTime, WhyReload */
#define	TRAP_WARMSTART		1 /* ??? */
#define	TRAP_LINKDOWN		2 /* ifIndex, ifDescr, ifType, locIfReason */
#define	TRAP_LINKUP		3 /* ifIndex, ifDescr, ifType, locIfReason */
#define	TRAP_ENTERPRISESPECIFIC	6

/*
 * Cisco Private Enterprise traps.
 */
#define	TRAP_RELOAD		0 /* sysUpTime, WhyReload */
#define	TRAP_TCPCONNECTIONCLOSE	1

/*
 * BGP Enterprise specific trap according to RFC1657.
 */
/* The BGP Established event is generated when the BGP FSM enters
   the ESTABLISHED state */
#define	TRAP_BGPESTABLISHED	1 /* bgpPeerState, bgpPeerLastError */
/* The BGPBackwardTransition Event is generated when the BGP FSM moves
   from a higher numbered state to a lower numbered state */
#define	TRAP_BGPBACKWARDTRANSITION 2 /* bgpPeerState, bgpPeerLastError */


extern char *snmp_error[];
