/*
 *	Copyright (c) 2000 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif

#include "netmon.h"
#include "snmp.h"
#include "mib.h"

extern CONFIG *cf;

int trap_sock = -1;

/*
 * [Re]initialize SNMP trap server socket.
 */
int
trap_init(enable)
	int enable;
{
	static struct sockaddr_in sin;

	if (enable) {
		if (trap_sock != -1) /* already enabled */
			return 0;

		if ((trap_sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			report(LOG_ERR, "socket: %m");
			return -1;
		}
		memset(&sin, 0, sizeof(sin));
		sin.sin_family = AF_INET;
		sin.sin_port = htons(SNMPTRAP_PORT);
		sin.sin_addr.s_addr = INADDR_ANY;
		if (bind(trap_sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
			report(LOG_ERR, "bind port %d: %m", ntohs(sin.sin_port));
			close(trap_sock);
			trap_sock = -1;
			return -1;
		}
	} else if (trap_sock != -1) {
		close(trap_sock);
		trap_sock = -1;
	}
	return 0;
}

OBJECT *
trap_auth(agent_addr, community, objid, objid_len, type, specific)
	struct in_addr *agent_addr;
	char *community;
	oid *objid;
	int objid_len;
	int type;
	int specific;
{
	OBJECT *object, *service;
	TRAP *trap;
	VARIABLE *var;
	char ipaddr[20];

	/* check for agent address */
	for (object = cf->target; object; object = object->next) {
		if (!memcmp(&object->ip_addr, agent_addr, sizeof(struct in_addr)))
			break;
	}
	if (!object) {
		report(LOG_DEBUG, "Unexpected trap from %s (no target object?)",
		       intoa(ipaddr, *agent_addr));
		return NULL;
	}

	if (type == 6) { /* Enterprise Specific trap -- match services first */
		/*
		 * Find appropriate enterprise identifier and specific code
		 * of the object's service and match it community name.
		 */
		for (service = object->service; service; service = service->next) {
			for (trap = service->trap_list; trap; trap = trap->next) {
				if (trap->type == type &&
				    trap->specific == specific &&
				    trap->objid_len == objid_len &&
				    !memcmp(trap->objid, objid, objid_len * sizeof(oid)) &&
				    !strcmp(trap->community, community))
					return service;
			}
		}
	} else if ((var = get_var(object->var_list, "sysObjID")) != NULL &&
		   var->type == OBJID && var->value.objid) {
		if (var->length / sizeof(oid) != objid_len ||
		    memcmp(var->value.objid, objid, var->length)) {
			report(LOG_WARNING, "Inconsistent trap from %s (wrong sysObjID?)",
			       intoa(ipaddr, *agent_addr));
			return NULL;
		}
	}

	for (trap = object->trap_list; trap; trap = trap->next) {
		if ((!trap->objid ||
		     (trap->type == type &&
		      trap->specific == specific &&
		      trap->objid_len == objid_len &&
		      !memcmp(trap->objid, objid, objid_len * sizeof(oid)))) &&
		    !strcmp(trap->community, community))
			return object;
	}

	report(LOG_WARNING, "Illegal trap from %s (wrong community? [%s])",
	       intoa(ipaddr, *agent_addr), community);
	return NULL;
}

static VARIABLE *
get_var_by_oid(vlist, var)
	register VARIABLE *vlist;
	VARIABLE *var;
{
	for (; vlist; vlist = vlist->next) {
		if (vlist->objid && vlist->objid_len == var->objid_len &&
		    !memcmp(vlist->objid, var->objid, var->objid_len * sizeof(oid)))
			break;
	}
	return vlist;
}

void
trap_parse(trap)
	TRAP *trap;
{
	OBJECT *object = trap->owner;
	VARIABLE *var, *tvar;
	INTERFACE *iface;
	BGP_AS *bgp;
	BGP_PEER *peer;
	int state;
	u_long index;
	struct in_addr ip_addr;
	char tmpbuf[1024];

	/* sanity check */
	if (!object) return;

	dprintf(("trap_parse(\"%s\"): enterprise %s, type %d/%d, uptime %u\n",
		 object->name, oid2str(tmpbuf, trap->objid, trap->objid_len),
		 trap->type, trap->specific, trap->uptime));

	if (trap->type == TRAP_COLDSTART || trap->type == TRAP_WARMSTART) {

		trace(("\t %s Start\n", trap->type == TRAP_COLDSTART ? "Cold" : "Warm"));
		dump_var_list(trap->var_list);
		report(LOG_INFO, "%sStart trap from \"%s\"",
		       trap->type == TRAP_COLDSTART ? "Cold" : "Warm",
		       object->name);
cisco_reload:
		/* reflect var changes */
		for (var = trap->var_list; var; var = var->next) {
			if ((tvar = get_var(object->var_list, var->name)) != 0)
				set_var(tvar, var->type, var->value.integer, var->length);
		}
		/* force to reset sysUpTime at the object */
		if ((var = get_var(object->var_list, "sysUpTime")) != 0)
			set_var(var, UNSIGNED, &trap->uptime, sizeof(trap->uptime));

		/* mark DOWN and restart polling if required */
		if (object->snmpdata)
			object->snmpdata->mode = 0;
		remove_event(start_method_list, object);
		method_finished(object, NULL, NULL, 0);
		return;
	}

	if (trap->type == TRAP_LINKDOWN || trap->type == TRAP_LINKUP) {

		trace(("\t Link %s\n", trap->type == TRAP_LINKUP ? "Up" : "Down"));
		dump_var_list(trap->var_list);

/*		if (object->state != STATE_UP) {
			trace(("\t link trap from died router?\n"));
			return;
		}
*/
		state = (trap->type == TRAP_LINKUP ? STATE_UP : STATE_DOWN);

		/* search for trap source (interface) */
		var = get_var(trap->var_list, "ifIndex");
		if (!var || var->type != INTEGER || !var->value.integer) {
			trace(("\t can't get ifIndex\n"));
			report(LOG_WARNING, "Cant find $ifIndex at Link%s trap from \"%s\"",
			       state == STATE_UP ? "Up" : "Down", object->name);
			return;
		}
		index = oid2long(var->instance, var->instance_len);
		if (!index)
			index = *var->value.integer;
		if (!index) {
			report(LOG_WARNING, "No $ifIndex instance at Link%s trap from \"%s\"",
			       state == STATE_UP ? "Up" : "Down", object->name);
			return;
		}

		/* match it against tabIdx */
		for (iface = object->interface; iface; iface = iface->next) {
			if (iface->tabIdx && iface->tabIdx == index)
				break;
		}
		if (!iface) { /* try to match by ifDescr */
			var = get_var(trap->var_list, "ifDescr");
			if (!var || var->type != STRING || !var->value.string) {
				trace(("\t can't get ifDescr\n"));
				report(LOG_INFO, "Cant get $ifDescr at Link%s trap from \"%s\"",
				       state == STATE_UP ? "Up" : "Down", object->name);
				return;
			}
			for (iface = object->interface; iface; iface = iface->next) {
				if (match_ifdescr(iface->name, var->value.string))
					break;
			}
		}
		if (iface && iface->name)
			sprintf(tmpbuf, "\"%s\"", iface->name);
		else	sprintf(tmpbuf, "Interface %u", index);
		report(LOG_INFO, "Link trap from \"%s\": %s is %s",
		       object->name, tmpbuf, state2str(IF_STATE, state));
		if (!iface) {
			trace(("\t link trap from unknown interface\n"));
			return;
		}
		iface->prev_reply = iface->last_reply;
		gettimeofday(&iface->last_reply, 0);

		if (iface->state == state) { /* discard duplicates */
			trace(("\t link trap duplicated?\n"));
			return;
		}
		iface->prev_state = iface->state;
		iface->state = state;
		iface->last_change = iface->last_reply;

		/* reflect var changes */
		for (var = trap->var_list; var; var = var->next) {
			if ((tvar = get_var(iface->var_list, var->name)) != 0)
				set_var(tvar, var->type, var->value.integer, var->length);
		}

		/* force to clear ifLastChange at the interface */
		index = 0;
		if ((var = get_var(iface->var_list, "ifLastChange")) != 0)
			set_var(var, UNSIGNED, &index, sizeof(index));

		save_state(iface->save_list, TYPE_INTERFACE, iface);
		save_when(iface->save_list, TYPE_INTERFACE, iface);
		return;
	}

	if (trap->type != TRAP_ENTERPRISESPECIFIC) {
		report(LOG_WARNING, "Unknown Generic trap from \"%s\"", object->name);
		return;
	}

	if (trap->objid_len == bgpTrapEnterprise_len &&
	    !memcmp(trap->objid, bgpTrapEnterprise, trap->objid_len * sizeof(oid)) &&
	    (trap->specific == TRAP_BGPESTABLISHED ||
	     trap->specific == TRAP_BGPBACKWARDTRANSITION)) {

		trace(("\t BGP trap %d\n", trap->specific));
		dump_var_list(trap->var_list);

/*		if (object->state != STATE_UP) {
			trace(("\t bgp trap from died router?\n"));
			return;
		}
*/
		/* search for trap source (bgp peer) */
		var = get_var(trap->var_list, "bgpPeerState");
		if (!var || var->type != INTEGER || !var->value.integer) {
			trace(("\t can't get bgpPeerState\n"));
			report(LOG_WARNING, "Cant find $bgpPeerState at BGP trap from \"%s\"",
			       object->name);
		}
		state = *var->value.integer;
		index = oid2long(var->instance, var->instance_len);
		if (!index) {
			report(LOG_WARNING, "No $bgpPeerState instance at BGP trap from \"%s\"",
			       object->name);
			return;
		}

		/* match it against tabIdx */
		peer = 0;
		for (bgp = object->bgp; bgp; bgp = bgp->next) {
			for (peer = bgp->peer; peer; peer = peer->next) {
				if (peer->tabIdx && peer->tabIdx == index)
					break;
			}
			if (peer) break;
		}
		if (peer)
			ip_addr = peer->ip_addr;
		else	ip_addr.s_addr = htonl(index);
		report(LOG_INFO, "BGP trap from \"%s\": %s is %s", object->name,
		       intoa(tmpbuf, ip_addr), state2str(BGP_STATE, state));
		if (!peer) {
			trace(("\t bgp trap from unknown peer\n"));
			return;
		}
		peer->prev_reply = peer->last_reply;
		gettimeofday(&peer->last_reply, 0);

		if (peer->state == state) { /* discard duplicates */
			trace(("\t bgp trap duplicated?\n"));
			return;
		}
		peer->prev_state = peer->state;
		peer->state = state;
		peer->last_change = peer->last_reply;

		/* reflect var changes */
		for (var = trap->var_list; var; var = var->next) {
			if ((tvar = get_var(peer->var_list, var->name)) != 0)
				set_var(tvar, var->type, var->value.integer, var->length);
		}

		if (state == BGP_ESTABLISHED) {
			/* force to clear bgpPeerFsmEstablishedTime at the peer */
			index = 0;
			if ((var = get_var(peer->var_list, "bgpPeerFsmEstablishedTime")) != 0)
				set_var(var, UNSIGNED, &index, sizeof(index));
		}
		save_state(bgp->save_list, TYPE_BGPPEER, peer);
		save_when(bgp->save_list, TYPE_BGPPEER, peer);
		return;
	}

	if (trap->objid_len == ciscoProductID_len - 2 &&
	    !memcmp(trap->objid, ciscoProductID, trap->objid_len * sizeof(oid)) &&
	    trap->specific == TRAP_RELOAD) {

		trace(("\t Cisco Reload\n"));
		dump_var_list(trap->var_list);
		report(LOG_INFO, "CiscoReload trap from \"%s\"", object->name);

		goto cisco_reload;
	}

	/* match variables for user defined traps */
	for (var = trap->var_list; var; var = var->next) {
		if ((tvar = get_var_by_oid(object->var_list, var)) != 0)
			set_var(tvar, var->type, var->value.integer, var->length);
	}
	save_when(object->save_list, object->parent ? TYPE_SERVICE : TYPE_OBJECT, object);
}

void
trap_free(trap)
	TRAP *trap;
{
	if (trap) {
		if (trap->name) free(trap->name);
		if (trap->community) free(trap->community);
		if (trap->objid) free(trap->objid);
		free_var_list(trap->var_list);
		free(trap);
	}
}


/*
 * Built-in generic trap initiator block.
 */

TRAP	trap_generic = {
	"GENERIC",		/* name of traping method */
	"public",		/* community string */
	NULL,			/* enterprise object identifier */
	0,			/* number of subid's in the enterprise */
	0,			/* generic trap type */
	0,			/* specific trap code */
	0,			/* time stamp in 100*seconds */

	NULL,			/* trap owner */
	NULL,			/* list of variables */

	NULL			/* next traping method */
};
