/*
 *	Copyright (c) 1999-2003 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <signal.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>
#include <math.h>
#ifdef	__STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#ifdef	HAVE_DMALLOC_H
#include <dmalloc.h>
#else
#include "malloc.h"
#endif
#include "netmon.h"
#include "regex.h"

extern int errno;

void
#ifdef	__STDC__
report(int pri, const char *fmt, ...)
#else
report(pri, fmt, va_alist)
	int pri;
	const char *fmt;
	va_dcl
#endif
{
	va_list ap;
#ifdef	__STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
#endif
	if (!syslog_initialized) {
		int save_errno = errno;
		char *cp, ebuf[1024];

		cp = ebuf + strlen(strcpy(ebuf, fmt)) - 1;
		if (cp > ebuf && cp[-1] == '%' && cp[0] == 'm')
			cp[-1] = '\0';
		else	cp = NULL;
		vfprintf(stderr, ebuf, ap);
		if (cp) fputs(strerror(save_errno), stderr);
		fputc('\n', stderr);
		fflush(stderr);
	} else vsyslog(pri, fmt, ap);

	va_end(ap);
}

#define	DELIMITERS	" -,;_=\t\r\n"

int
match_ifdescr(name, descr)
	const char *name;
	const char *descr;
{
	if (name && *name && descr && *descr) {
		char *cp, *next, *aname[100], *adescr[100];
		int i, cname = 0, cdescr = 0, lastfound = -1;
		char namebuf[1024], descrbuf[1024];

		next = strcpy(namebuf, name);
		while ((cp = my_strsep(&next, DELIMITERS)) != NULL && cname < 100)
			aname[cname++] = cp;
		if (cname < 1) return 0;

		next = strcpy(descrbuf, descr);
		while ((cp = my_strsep(&next, DELIMITERS)) != NULL && cdescr < 100)
			adescr[cdescr++] = cp;
		if (cdescr < 1) return 0;

		for (i = 0; i < cname; i++) {
			int j;
			for (j = lastfound + 1; j < cdescr; j++) {
				if (!strcasecmp(aname[i], adescr[j])) {
					lastfound = j;
					break;
				}
			}
			if (j >= cdescr) return 0;
		}
		return 1;
	}
	return 0;
}

char *
strippath(path)
	char *path;
{
	register char *cp;
	if ((cp = strrchr(path, '/')) != NULL)
		cp++;
	else	cp = path;
	return cp;
}

char *
stripquotes(str)
	char *str;
{
	register char *cp = str;
	if (cp) {
		int len = strlen(cp);
		if (len > 0) {
			if (cp[0] == '"' && cp[len-1] == '"') {
				cp[len-1] = '\0';
				cp++;
			}
		}
	}
	return cp;
}

void
free_var(var)
	VARIABLE *var;
{
	if (!var) return;
	if (var->name) free(var->name);
	if (var->objid) free(var->objid);
	if (var->instance) free(var->instance);
	if (var->value.integer) free(var->value.integer);
	if (var->statistic) free(var->statistic);
	if (var->old) free_var(var->old);
	free(var);
}

void
free_var_list(var)
	register VARIABLE *var;
{
	VARIABLE *next;
	while (var) {
		next = var->next;
		free_var(var);
		var = next;
	}
}

VARIABLE *
dup_var(orig)
	const VARIABLE *orig;
{
	VARIABLE *var;

	if ((var = (VARIABLE *)malloc(sizeof(VARIABLE))) == NULL)
		return NULL;
	memcpy(var, orig, sizeof(VARIABLE));
	if (var->name && (var->name = strdup(var->name)) == NULL)
		return NULL;
	if (var->objid && var->objid_len) {
		if ((var->objid = (oid *)malloc(var->objid_len * sizeof(oid))) != NULL)
			memcpy(var->objid, orig->objid, var->objid_len * sizeof(oid));
		else	var->objid_len = 0;
	}
	if (var->instance && var->instance_len) {
		if ((var->instance = (oid *)malloc(var->instance_len * sizeof(oid))) != NULL)
			memcpy(var->instance, orig->instance, var->instance_len * sizeof(oid));
		else	var->instance_len = 0;
	}
	if (var->value.integer) {
		if ((var->value.integer = (void *)malloc(var->length)) != NULL)
			memcpy(var->value.integer, orig->value.integer, var->length);
		else	var->length = 0;
	}
	var->statistic = 0;
	var->old = 0;
	var->next = 0;

	return var;
}

/* make copy of variable reference and return it */
REFERENCE *
dup_var_ref(orig)
	REFERENCE *orig;
{
	REFERENCE *first, *curr, *last;

	for (first = NULL; orig; orig = orig->next) {
		if ((curr = (REFERENCE *)malloc(sizeof(REFERENCE))) == NULL)
			return NULL;
		memcpy(curr, orig, sizeof(REFERENCE));
		if (curr->name) {
			if ((curr->name = strdup(curr->name)) == NULL)
				return NULL;
		}
		curr->next = NULL;
		if (first)
			last->next = curr;
		else	first = curr;
		last = curr;
	}
	return first;
}

SNMP_REQ *
dup_snmp_request(old)
	SNMP_REQ *old;
{
	SNMP_REQ *new;

	if (!old) return NULL;
	if ((new = (SNMP_REQ *)malloc(sizeof(SNMP_REQ))) == NULL)
		return NULL;
	memcpy(new, old, sizeof(SNMP_REQ));
	if (old->var_ref) {
		if ((new->var_ref = dup_var_ref(old->var_ref)) == NULL)
			return NULL;
	}
	return new;
}

void
free_var_ref(ref)
	REFERENCE *ref;
{
	REFERENCE *next;
	while (ref) {
		next = ref->next;
		if (ref->name) free(ref->name);
		free(ref);
		ref = next;
	}
}

void
free_chatscript(cs)
	CHATSCRIPT *cs;
{
	CHATSCRIPT *next;
	while (cs) {
		next = cs->next;
		if (cs->send) free(cs->send);
		if (cs->expect) free(cs->expect);
		free_var_ref(cs->var_ref);
		free(cs);
		cs = next;
	}
}

void
free_method(method)
	METHOD *method;
{
	if (!method) return;
	if (method->name) free(method->name);
	if (method->argument) free(method->argument);
	if (method->address) free(method->address);
	if (method->sd) session_free(method->sd);
	if (method->chatscript) free_chatscript(method->chatscript);
	if (method->snmpreq) {
		free_var_ref(method->snmpreq->var_ref);
		free(method->snmpreq);
	}
	free(method);
}

void
free_method_list(method)
	METHOD *method;
{
	METHOD *next;
	while (method) {
		next = method->next;
		free_method(method);
		method = next;
	}
}

void
free_trap_list(trap)
	TRAP *trap;
{
	TRAP *next;
	while (trap) {
		next = trap->next;
		trap_free(trap);
		trap = next;
	}
}

void
free_save(save)
	SAVE *save;
{
	if (!save) return;
	if (save->name) free(save->name);
	if (save->file) free(save->file);
	if (save->saving) free(save->saving);
	if (save->state) free(save->state);
	if (save->when) free(save->when);
	if (save->when_fmt) free(save->when_fmt);
	free(save);
}

void
free_save_list(save)
	SAVE *save;
{
	SAVE *next;
	while (save) {
		next = save->next;
		remove_event(NULL, save);
		free_save(save);
		save = next;
	}
}

SAVE *
dup_save(orig)
	const SAVE *orig;
{
	SAVE *save;

	if ((save = (SAVE *)malloc(sizeof(SAVE))) == NULL)
		return NULL;
	memcpy(save, orig, sizeof(SAVE));
	if (save->name && (save->name = strdup(save->name)) == NULL)
		return NULL;
	if (save->file && (save->file = strdup(save->file)) == NULL)
		return NULL;
	if (save->saving && (save->saving = strdup(save->saving)) == NULL)
		return NULL;
	if (save->state && (save->state = strdup(save->state)) == NULL)
		return NULL;
	if (save->when && (save->when = strdup(save->when)) == NULL)
		return NULL;
	if (save->when_fmt && (save->when_fmt = strdup(save->when_fmt)) == NULL)
		return NULL;
	save->next = 0;
	return save;
}

void
free_group_list(chain, name)
	GROUP_LIST **chain;
	const char *name;	/* 0 - free all chain */
{
	GROUP_LIST *curr_list, *next_list, *prev_list = 0;
	GROUP_ENTRY *curr_entry, *next_entry;

	/* sanity check */
	if (!chain) return;

	for (curr_list = *chain; curr_list; curr_list = next_list) {
		next_list = curr_list->next;
		if (!name || !curr_list->name ||
		    !strcasecmp(name, curr_list->name)) {
			if (curr_list->name)
				free(curr_list->name);
			curr_entry = curr_list->entry;
			while (curr_entry) {
				next_entry = curr_entry->next;
				if (curr_entry->pattern)
					free(curr_entry->pattern);
				free(curr_entry);
				curr_entry = next_entry;
			}
			if (prev_list)
				prev_list->next = next_list;
			else	*chain = next_list;
		} else	prev_list = curr_list;
	}
}

void
free_group_ref(chain)
	GROUP_REF **chain;
{
	GROUP_REF *ref, *next;
	if (!chain) return;
	ref = *chain;
	while (ref) {
		next = ref->next;
		if (ref->name) free(ref->name);
		free(ref);
		ref = next;
	}
	*chain = 0;
}

int
check_group_ref(group_ref, str, altstr)
	GROUP_REF *group_ref;
	char *str, *altstr;
{
	GROUP_REF *ref;
	GROUP_LIST *list;
	GROUP_ENTRY *entry;
	yre_t prog;
	char *yre_error;
	extern CONFIG *cf;

	if (!cf) return 0;
	for (ref = group_ref; ref; ref = ref->next) {
		if (!ref->name || !*ref->name)
			continue;
		for (list = cf->group_list; list; list = list->next) {
			if (list->name && !strcasecmp(list->name, ref->name))
				break;
		}
		if (!list) {
			report(LOG_WARNING, "Group \"%s\": No such name", ref->name);
			continue;
		}
		for (entry = list->entry; entry; entry = entry->next) {
			yre_error = yre_comp(&prog, entry->pattern);
			if (yre_error) {
				report(LOG_ERR, "Group \"%s\": \"%s\": %s",
				       list->name, entry->pattern, yre_error);
				continue;
			}
			if (str && *str && yre_exec(&prog, str))
				return entry->permit;
			if (altstr && *altstr && yre_exec(&prog, altstr))
				return entry->permit;
		}
	}
	return 0;
}

/* emulate group reference */
int
check_group(ref_name, str, altstr)
	char *ref_name;
	char *str, *altstr;
{
	GROUP_REF group_ref;
	memset(&group_ref, 0, sizeof(GROUP_REF));
	group_ref.name = ref_name;
	return check_group_ref(&group_ref, str, altstr);
}

#ifdef	unused
int
check_group_var(ref_name, var)
	char *ref_name;
	VARIABLE *var;
{
	char buf[1024];
	GROUP_REF group_ref;
	memset(&group_ref, 0, sizeof(GROUP_REF));
	group_ref.name = ref_name;
	return check_group_ref(&group_ref, stripquotes(var2str(buf, var)), 0);
}
#endif

int
make_argv(buf, avp, max)
	char *buf, ***avp;
	int max;
{
	int ac, op;
	char *cp, *ap, **av;

	ac = op = 0;
	cp = buf;
	av = (char **)avp;
	while (cp && *cp && ac < max) {
		switch (*cp) {
		case ' ':
			if (op == ' ') {
				*cp = '\0';
				*av++ = ap;
				ac++;
				op = 0;
			}
			break;
		case '"':
			if (!op) {
				if (cp == buf || cp[-1] == ' ' || cp[-1] == '\0') {
					ap = ++cp;
					op = '"';
					continue;
				}
			} else if (op == '"' && (cp[1] == ' ' || cp[1] == '\0')) {
				*cp = '\0';
				*av++ = ap;
				ac++;
				op = 0;
			}
			break;
#ifdef	notdef
		case '<':
		case '(':
			if (!op && (cp == buf || cp[-1] == ' ' || cp[-1] == '\0')) {
				op = *cp;
				ap = ++cp;
				continue;
			}
			break;
		case '>':
		case ')':
			if (((op == '<' && *cp == '>') ||
			     (op == '(' && *cp == ')')) &&
			    (cp[1] == ' ' || cp[1] == '\0')) {
				*cp = '\0';
				*av++ = ap;
				ac++;
				op = 0;
				break;
			}
#endif
		default:
			if (!op && (cp == buf || cp[-1] == ' ' || cp[-1] == '\0')) {
				ap = cp;
				op = ' ';
			}
		}
		cp++;
	}
	if (op) {
		*av++ = ap;
		ac++;
	}
	return ac;
}

static char *
str_sep(stringp, delim)
	register char **stringp;
	register const char *delim;
{
	register char *s;
	register const char *spanp;
	register int c, sc;
	char *tok;

	if ((s = *stringp) == NULL)
		return NULL;
	for (tok = s;;) {
		c = *s++;
		spanp = delim;
		do {
			if ((sc = *spanp++) == c) {
				if (c == 0)
					s = NULL;
				else	s[-1] = 0;
				*stringp = s;
				return tok;
			}
		} while (sc != 0);
	}
}

char *
my_strsep(stringp, delim)
	register char **stringp, *delim;
{
	register char *cp;

	while ((cp = str_sep(stringp, delim)) != NULL && *cp == '\0');

	return cp;
}

/*
 * Checksum routine for Internet Protocol family headers.
 */
u_short
in_cksum(addr, len)
	u_short *addr;
	int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *)(&answer) = *(u_char *)w ;
		sum += answer;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return answer;
}

/*
 * A faster and thread-safe replacement for inet_ntoa().
 * XXX IPv6 incompatible.
 */
char *
intoa(tbuf, saddr)
	char *tbuf;
	struct in_addr saddr;
{
	in_addr_t addr;
	register char *cp;
	register u_int byte;
	register int n;
	char buf[sizeof(".xxx.xxx.xxx.xxx")];

	memcpy(&addr, &saddr, sizeof(addr));
	addr = ntohl(addr);
	cp = &buf[sizeof(buf)];
	*--cp = '\0';

	n = 4;
	do {
		byte = addr & 0xff;
		*--cp = byte % 10 + '0';
		byte /= 10;
		if (byte > 0) {
			*--cp = byte % 10 + '0';
			byte /= 10;
			if (byte > 0)
				*--cp = byte + '0';
		}
		*--cp = '.';
		addr >>= 8;
	} while (--n > 0);

	return strcpy(tbuf, cp + 1);
}

char *
type2str(type)
	int type;
{
	static char *types[7] = {
		"integer", "unsigned", "unsigned64",
		"ipaddr", "double", "string", "objid" };

	if (type >= 0 && type < 7)
		return types[type];
	return "unknown";
}

static int
str2type(str)
	char *str;
{
	register char *cp;
	int digits = 0, octets = 0, sign = 0;

	for (cp = str; cp && *cp; cp++) {
		if (!digits && *cp == '-') {
			sign++;
			cp++;
		}
		for (str = cp; *cp >= '0' && *cp <= '9'; cp++);
		if (cp > str && (*cp == '.' || *cp == '\0')) {
			digits++;
			if (!sign) octets += (atoi(str) < 256);
			if (*cp == '\0' || cp[1] == '\0') break;
		} else break;
	}
	if (cp && *cp == '\0') {
		if (!sign) {
			if (digits > 4)
				return (OBJID);
			if (digits == 4 && octets == 4)
				return (IPADDRESS);
		}
		if (digits == 2)
			return (FLOATING);
		if (digits == 1)
			return (sign ? INTEGER : UNSIGNED);
	}
	return (STRING);
}

/*
 * Convert SNMP variable to printable string.
 */
char *
oid2str(buf, var, len)
	char *buf;
	const oid *var;
	int len;
{
	char *cp = buf;

	*buf = '\0';
	while (len-- > 0) {
		if (cp != buf) *cp++ = '.';
		cp += sprintf(cp, "%d", (int)*var++);
	}
	return buf;
}

void
str2oid(buf, size, str)
	oid *buf;
	int *size; /* IN: max buf length in oids, OUT: actual length */
	char *str;
{
	register char *cp;
	char *num = 0;
	int len = 0;
	for (cp = str; cp; cp++) {
		if (*cp == '\0' || *cp == '.') {
			if (num) {
				if (len >= *size) break;
				buf[len++] = (oid)atoi(num);
				num = 0;
			}
			if (*cp == '\0') break;
		} else if (*cp >= '0' && *cp <= '9') {
			if (!num) num = cp;
		} else {
			len = 0;
			break;
		}
	}
	*size = len;
}

/*
 * Convert value of variable to printable string.
 */
char *
var2str(buf, var)
	char *buf;
	VARIABLE *var;
{
	int i;
	char *cp;
	char *tp;
	struct in_addr in;
	char tmpbuf[1024];

	if (!var || !var->value.integer) {
		(void)strcpy(buf, "Unused");
	} else {
		switch (var->type) {
		case INTEGER:
			sprintf(buf, "%ld", (long)(*var->value.integer));
			break;
		case UNSIGNED:
			sprintf(buf, "%lu", *var->value.integer);
			break;
		case UNSIGNED64:
			sprintf(buf, "%llu", *var->value.integer64);
			break;
		case FLOATING:
			sprintf(buf, "%.12g", *var->value.floating);
			break;
		case IPADDRESS:
			memcpy(&in, var->value.string,
			       var->length > sizeof(in) ? sizeof(in) : var->length);
			(void)strcpy(buf, intoa(tmpbuf, in));
			break;
		case STRING:
			cp = var->value.string;
			for (i = 0; i < var->length; i++) {
				if ((*cp >= ' ' && *cp < 127) ||
				    *cp == '\r' || *cp == '\n' || *cp == '\t') {
					cp++;
					continue;
				}
				break;
			}
			cp = var->value.string;
			tp = buf;
			*tp++ = '"';
			if (i < var->length) { /* in HEX */
				for (i = 0; i < var->length && i < 256; i++) {
					if (i) *tp++ = '_';
					tp += sprintf(tp, "%02X", (int)(*cp & 0xff));
					cp++;
				}
			} else {
				for (i = 0; i < var->length && i < 256; i++) {
					if (*cp == '\r' || *cp == '\n' || *cp == '\t') {
						*tp++ = ' ';
					} else	*tp++ = *cp;
					cp++;
				}
			}
			*tp++ = '"';
			*tp = '\0';
			break;
		case OBJID:
			(void)strcpy(buf, oid2str(tmpbuf, var->value.objid, var->length / sizeof(oid)));
			break;
		default:
			(void)strcpy(buf, "Unknown value type");
		}
	}
	return buf;
}

void
str2var(var, str)
	VARIABLE *var;
	char *str;
{
	int type, length;
	void *data;
	long integer;
	u_int64_t integer64;
	double floating;
	in_addr_t addr;
	oid objid[100];
	char *cp, *next, tmpbuf[1024];

	if (!var) return;

	type = str2type(str);
	switch (type) {
	case INTEGER:
		integer = strtol(str, (char **)0, 10);
		data = &integer;
		length = sizeof(integer);
		break;
	case UNSIGNED:
		errno = 0;
		integer = strtoul(str, (char **)0, 10);
		if (errno != ERANGE) {
			data = &integer;
			length = sizeof(integer);
			break;
		}
		/* LONG_MAX? try to biggest number... */
		type = UNSIGNED64;
	case UNSIGNED64:
		integer64 = strtoull(str, (char **)0, 10);
		data = &integer64;
		length = sizeof(integer64);
		break;
	case FLOATING:
		floating = strtod(str, (char **)0);
		data = &floating;
		length = sizeof(floating);
		break;
	case IPADDRESS:
		if ((addr = inet_addr(str)) == -1) return;
		data = &addr;
		length = sizeof(addr);
		break;
	case OBJID:
		next = strcpy(tmpbuf, str);
		length = 0;
		while ((cp = my_strsep(&next, ".")) != NULL && length < 100)
			objid[length++] = atoi(cp);
		data = objid;
		length *= sizeof(oid);
		break;
	default:
		data = str;
		length = strlen(str);
	}
	set_var(var, type, data, length);
}

void
set_var(var, type, data, length)
	VARIABLE *var;
	int type;
	void *data;
	int length;
{
	int tzero = 0;

	if (!var) return;

	if (type == STRING) tzero++;

	/* save previous value to old */
	if (var->old) free_var(var->old);
	if ((var->old = dup_var(var)) == NULL) {
		report(LOG_ERR, "set_var: Out of memory");
		return;
	}
	if (var->value.integer) free(var->value.integer);
	if (data && length + tzero) {
		if ((var->value.integer = (u_int32_t *)malloc(length + tzero)) == NULL) {
			report(LOG_ERR, "set_var: Out of memory");
			return;
		}
		memcpy(var->value.integer, data, length);
		if (tzero) var->value.string[length] = '\0';
	} else	var->value.integer = NULL;
	var->type = type;
	var->length = length;
#ifdef	USE_VAR_MTIME
	gettimeofday(&var->mtime, NULL);
#endif

	/* make statistic */
	if (var->value.integer &&
	    (type == INTEGER || type == UNSIGNED || type == UNSIGNED64 ||
	     type == FLOATING)) {
		register STATISTIC *st;
		u_int64_t value;

		if (!var->statistic) {
			if ((var->statistic = (STATISTIC *)malloc(sizeof(STATISTIC))) == NULL) {
				report(LOG_ERR, "set_var: Out of memory");
				return;
			}
			memset(var->statistic, 0, sizeof(STATISTIC));
		}
		st = var->statistic;

		if (type == FLOATING)
			value = rint(*var->value.floating);
		else if (type == UNSIGNED64)
			value = *var->value.integer64;
		else	value = *var->value.integer;
		st->count++;

		/* abs values */
		st->sum += value;
		if (value > st->max) st->max = value;
		if (!st->min) st->min = value;
		if (value < st->min) st->min = value;

		/* delta value */
		if (var->old->value.integer) {
			u_int64_t prev;

			if (type == FLOATING)
				prev = rint(*var->old->value.floating);
			else if (type == UNSIGNED64)
				prev = *var->old->value.integer64;
			else	prev = *var->old->value.integer;
			if (value >= prev)
				var->delta = value - prev;
			st->delta += var->delta;
		}
	}
}

int
is_var_changed(vlist, name)
	VARIABLE *vlist;
	const char *name;
{
	VARIABLE *var, *old;
	if ((var = get_var(vlist, name)) == 0)
		return 0;
	old = var->old;
	if (!old) return 1;
	return (var->type != old->type || var->length != old->length ||
		memcmp(var->value.integer, old->value.integer, var->length));
}

void
set_instance(var, instance, length)
	VARIABLE *var;
	const oid *instance;
	int length;	/* suboids */
{
	if (var->instance) {
		if (length && instance && var->instance_len == length) {
			if (var->instance != instance)
				memcpy(var->instance, instance, length * sizeof(oid));
			return;
		}
		free(var->instance);
		var->instance = NULL;
	}
	var->instance_len = length;
	if (length && instance) {
		if ((var->instance = (oid *)malloc(length * sizeof(oid))) != NULL)
			memcpy(var->instance, instance, length * sizeof(oid));
		else	var->instance_len = 0;
	}
}

VARIABLE *
get_var(vlist, name)
	VARIABLE *vlist;
	const char *name;
{
	if (vlist && name) {
		VARIABLE *var = vlist;
		while (var) {
			if (var->name && !strcasecmp(var->name, name))
				return var;
			var = var->next;
		}
	}
	return 0;
}

VARIABLE *
add_var(var_list, name)
	VARIABLE **var_list;
	const char *name;
{
	VARIABLE *curr, *last = NULL;

	/* prevent dups */
	for (curr = *var_list; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, name))
			return curr;
		last = curr;
	}
	if ((curr = (VARIABLE *)malloc(sizeof(VARIABLE))) == NULL)
		return NULL;

	memset(curr, 0, sizeof(VARIABLE));
	if ((curr->name = strdup(name)) == NULL) {
		free(curr);
		return NULL;
	}
	if (last)
		last->next = curr;
	else	*var_list = curr;
	return curr;
}

void
set_objid(var, objid, objid_len)
	VARIABLE *var;
	const oid *objid;
	int objid_len;
{
	int i, instance_len = 0;
	const oid *instance = 0;

	if (!var) return;
	if (var->objid) {
		free(var->objid);
		var->objid = 0;
	}
	var->objid_len = 0;
	if (!objid || objid_len < 4 || !*objid) {
		set_instance(var, 0, 0);
		return;
	}
	var->objid = (oid *)malloc(objid_len * sizeof(oid));
	if (!var->objid)
		return;
	var->objid_len = objid_len;
	for (i = 0; i < objid_len; i++) {
		if (objid[i] == 0) {
			if (!instance)
				instance = &objid[i];
			instance_len++;
		}
		var->objid[i] = objid[i];
	}
	set_instance(var, instance, instance_len);
}

VARIABLE *
make_var(name, objid, objid_len, mode)
	const char *name;
	const oid *objid;
	int objid_len;
	int mode;
{
	VARIABLE *var;

	if ((var = (VARIABLE *)malloc(sizeof(VARIABLE))) == 0)
		return 0;
	memset(var, 0, sizeof(VARIABLE));
	if (name && (var->name = strdup(name)) == 0) {
		free(var);
		return 0;
	}
	if (objid) set_objid(var, objid, objid_len);
	var->mode = mode;
	return var;
}

/* add copy of variable to the list */
VARIABLE *
add_var_copy(list, from)
	VARIABLE **list;
	const VARIABLE *from;
{
	VARIABLE *var, *curr, *last;

	/* sanity check */
	if (!list || !from)
		return 0;

	if ((var = dup_var(from)) == 0)
		return 0;
	last = 0;
	for (curr = *list; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, var->name)) { /* replace */
			var->next = curr->next;
			free_var(curr);
			if (last)
				last->next = var;
			else	*list = var;
			return var;
		}
		last = curr;
	}
	/* append */
	if (last)
		last->next = var;
	else    *list = var;
	return var;
}

/* add copy of save method to the list */
SAVE *
add_save_copy(list, from)
	SAVE **list;
	const SAVE *from;
{
	SAVE *save, *curr, *last;

	/* sanity check */
	if (!list || !from)
		return 0;

	if ((save = dup_save(from)) == 0)
		return 0;
	last = 0;
	for (curr = *list; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, save->name)) { /* replace */
			save->next = curr->next;
			free_save(curr);
			if (last)
				last->next = save;
			else	*list = save;
			return save;
		}
		last = curr;
	}
	/* append */
	if (last)
		last->next = save;
	else    *list = save;
	return save;
}

GROUP_REF *
add_group_ref_copy(list, from)
	GROUP_REF **list;
	const GROUP_REF *from;
{
	GROUP_REF *curr, *last = 0;

	/* sanity check */
	if (!list || !from || !from->name)
		return 0;

	for (curr = *list; curr; curr = curr->next) {
		if (!strcasecmp(curr->name, from->name))
			return curr; /* prevent dups */
		last = curr;
	}
	/* append */
	curr = (GROUP_REF *)malloc(sizeof(GROUP_REF));
	if (!curr) return 0;
	memset(curr, 0, sizeof(GROUP_REF));
	curr->name = strdup(from->name);
	if (!curr->name) {
		free(curr);
		return 0;
	}
	if (last)
		last->next = curr;
	else    *list = curr;
	return curr;
}

void
print_var_list(prepend, var_list)
	char *prepend;
	VARIABLE *var_list;
{
	if (var_list) {
		VARIABLE *var;
		char tmpbuf[1024];

		printf("%sSoft variables...\n", prepend);
		for (var = var_list; var; var = var->next) {
			printf("%s\t$%s", prepend, var->name);
			if (var->instance && *var->instance)
				printf("(%s)", oid2str(tmpbuf, var->instance, var->instance_len));
			if (var->objid)
				printf(" [%s]", oid2str(tmpbuf, var->objid, var->objid_len));
			printf(" = %s\n", var2str(tmpbuf, var));
		}
	}
}

void
print_method_list(prepend, method)
	char *prepend;
	METHOD *method;
{
	CHATSCRIPT *cs;
	REFERENCE *ref;
	struct protoent *proto;
	struct servent *serv;
	char ipaddr[20];

	for (; method; method = method->next) {
		printf("%sMethod = \"%s\"\n", prepend, method->name);
		if (method->argument)
			printf("%s\tArguments = \"%s\"\n", prepend, method->argument);
		if (method->address)
			printf("%s\tAddress = \"%s\" [%s]\n", prepend,
			       method->address, intoa(ipaddr, method->ip_addr));
		if ((proto = getprotobynumber(method->protocol)) == NULL) {
			printf("%s\tUnknown protocol %d\n", prepend, method->protocol);
			continue;
		}
		printf("%s\t%s ", prepend, proto->p_name);
		switch (method->protocol) {
		case IPPROTO_ICMP:
			if (method->rport)
				printf("packets Size = %d, Send = %d, Expect = %d",
				       method->rport, method->ping_send, method->ping_expect);
			else	printf("packet for Record Routes");
			break;
		case IPPROTO_TCP:
		case IPPROTO_UDP:
			printf("Port = %d", method->rport);
			if ((serv = getservbyport(htons(method->rport), proto->p_name)) != NULL)
				printf(" (%s)", serv->s_name);
			if (method->protocol == IPPROTO_UDP &&
			    (!strcasecmp(method->name, "ROUTER") ||
			     !strcasecmp(method->name, "SNMP"))) {
				printf(", Version = %d", method->snmp_version+1);
				if (method->router_batch)
					printf(", Batch = %d", method->router_batch);
			} else if (method->lport_min) {
				printf(", LocalPort = %d", method->lport_min);
				if (method->lport_max > method->lport_min)
				printf("..%d", method->lport_max);
			}
			break;
		default:
			printf("Unsupported");
		}
		printf("\n");

		printf("%s\tTimeout = %d sec.\n", prepend, method->timeout);
		printf("%s\tRetries = %d\n", prepend, method->retries);
		if (method->chatscript) {
			printf("%s\tChatScript...\n", prepend);
			for (cs = method->chatscript; cs; cs = cs->next) {
				if (cs->send && *cs->send)
					printf("%s\t\tSend \"%s\"\n",
					       prepend, cs->send);
				else	printf("%s\t\tNoSend\n", prepend);
				if (cs->expect && *cs->expect) {
					printf("%s\t\tExpect \"%s\"\n",
					       prepend, cs->expect);
				} else	printf("%s\t\tNoExpect\n", prepend);
				if (cs->var_ref) {
					printf("%s\t\t\tExtract for ", prepend);
					for (ref = cs->var_ref; ref; ref = ref->next) {
						if (ref != cs->var_ref)
							printf(", ");
						if (ref->var && ref->var->name)
							printf("$%s", ref->var->name);
						else	printf("$(null)");
					}
					printf("\n");
				}
			}
		}
		if (method->snmpreq) {
			printf("%s\tSNMP-Request...\n", prepend);
			if (method->snmpreq->var_ref) {
				printf("%s\t\tExtract for ", prepend);
				for (ref = method->snmpreq->var_ref; ref; ref = ref->next) {
					if (ref != method->snmpreq->var_ref)
						printf(", ");
					if (ref->var && ref->var->name)
						printf("$%s", ref->var->name);
					else	printf("$(null)");
				}
				printf("\n");
			}
		}
	}
}

void
print_trap_list(prepend, trap)
	char *prepend;
	TRAP *trap;
{
	char tmpbuf[1024];

	for (; trap; trap = trap->next) {
		printf("%sTrap = \"%s\"\n", prepend, trap->name);
		printf("%s\tCommunity = \"%s\"\n", prepend, trap->community);
		if (trap->objid)
			printf("%s\tEnterprise = %s\n", prepend, oid2str(tmpbuf, trap->objid, trap->objid_len));
		if (trap->type)
			printf("%s\tType = EnterpriseSpecific(%d)\n", prepend, trap->specific);
		else	printf("%s\tType = Reload, Link, BGP\n", prepend);
	}
}

void
print_save_list(prepend, save)
	char *prepend;
	SAVE *save;
{
	for (; save; save = save->next) {
		printf("%sSave = \"%s\"\n", prepend, save->name);
		printf("%s\tMethod = ", prepend);
		switch (save->method) {
			case SAVE_FILE:	printf("FILE\n"); break;
			case SAVE_EXEC:	printf("EXEC\n"); break;
			case SAVE_PIPE:	printf("PIPE\n"); break;
			default:	printf("Unknown\n");
		}
		printf("%s\tFile = \"%s\"\n", prepend, save->file);
		if (save->saving)
			printf("%s\tSaving = \"%s\"\n", prepend, save->saving);
		if (save->state)
			printf("%s\tState = \"%s\"\n", prepend, save->state);
		if (save->when) {
			printf("%s\tWhen = \"%s\"", prepend, save->when);
			if (save->delay > 0)
				printf(" %d sec.", save->delay);
			if (save->when_fmt)
				printf(" \"%s\"", save->when_fmt);
			printf("\n");
		}
	}
}

void
print_group_list(prepend, list)
	char *prepend;
	GROUP_LIST *list;
{
	GROUP_ENTRY *entry;
	for (; list; list = list->next) {
		printf("%sGroup = \"%s\"\n", prepend,
		       list->name ? list->name : "unspecified");
		for (entry = list->entry; entry; entry = entry->next) {
			printf("%s\t%s = \"%s\"\n", prepend,
			       entry->permit ? "Permit" : "Deny", entry->pattern);
		}
	}
}

void
print_group_ref(prepend, ref)
	char *prepend;
	GROUP_REF *ref;
{
	for (; ref; ref = ref->next) {
		printf("%sGroup = \"%s\"\n", prepend, ref->name);
	}
}

void
dump_config(cf)
	CONFIG *cf;
{
	OBJECT *target, *service;
	INTERFACE *iface;
	BGP_AS *bgp;
	BGP_PEER *peer;
	ENV_MON *env;
	ENV_GAUGE *gauge;
	char ipaddr[20];

	printf("RootDir = \"%s\"\n", cf->rootdir);
	printf("TimeFmt = \"%s\"\n", cf->timefmt ? cf->timefmt : "%s");
	if (cf->polling > 0)
		printf("Polling = %d sec.\n", cf->polling);
	else	printf("Polling disabled\n");
	printf("\tTimeout = %d sec.\n", cf->timeout);
	printf("\tRetries = %d\n", cf->retries);
	if (cf->saving > 0)
		printf("Saving = %d sec.\n", cf->saving);
	else	printf("Saving disabled\n");

	print_group_list("", cf->group_list);

	printf("NetState %s\n", cf->ns_port ? "enabled" : "disabled");
	if (cf->ns_port) {
		printf("\tPort = %d\n", cf->ns_port);
#ifndef	HAVE_PTHREAD
		printf("\tTimeout = %d sec.\n", cf->ns_timo);
#endif
		print_group_ref("\t", cf->ns_acl);
	}

	printf("Traps ");
	if (cf->enable_traps > 0) {
		printf("enabled");
		if (cf->source_traps > 0) printf(" (sourcecheck)");
	} else	printf("disabled");
	printf("\n");

	for (target = cf->target; target; target = target->next) {
		printf("Object = \"%s\"\n", target->name);
		if (target->descr)
			printf("\tDescription = \"%s\"\n", target->descr);
		printf("\tAddress = \"%s\" [%s]\n", target->address,
		       intoa(ipaddr, target->ip_addr));
		if (target->polling > 0)
			printf("\tPolling = %d sec.\n", target->polling);
		else	printf("\tPolling disabled\n");
		if (target->saving > 0) {
			printf("\tSaving = %d sec.", target->saving);
			if (target->sync)
				printf(" (sync by %d poll)\n", target->sync);
			else	printf(" (async)\n");
		} else	printf("\tSaving disabled\n");
		if (target->ns_acl)
			print_group_ref("\tNetState ", target->ns_acl);
		else	printf("\tNetState Group = free access\n");
		print_var_list("\t", target->var_list);
		print_method_list("\t", target->method_list);
		print_trap_list("\t", target->trap_list);
		print_save_list("\t", target->save_list);

		for (iface = target->interface; iface; iface = iface->next) {
			if (iface->name)
				printf("\tInterface = \"%s\"\n", iface->name);
			if (iface->index > 0)
				printf("\tInterface = %d\n", iface->index);
			if (iface->descr)
				printf("\t\tDescription = \"%s\"\n", iface->descr);
			if (iface->bandwidth)
				printf("\t\tBandwidth = %d bps\n", (int)iface->bandwidth);
			if (iface->ns_acl)
				print_group_ref("\t\tNetState ", iface->ns_acl);
			else	printf("\t\tNetState Group = free access\n");
			print_var_list("\t\t", iface->var_list);
			print_save_list("\t\t", iface->save_list);
		}
		for (iface = target->ifgroup; iface; iface = iface->next) {
			printf("\tInterface Group = \"%s\"\n", iface->name);
			if (iface->descr)
				printf("\t\tDescription = \"%s\"\n", iface->descr);
			if (iface->bandwidth)
				printf("\t\tBandwidth = %d bps\n", (int)iface->bandwidth);
			if (iface->ns_acl)
				print_group_ref("\t\tNetState ", iface->ns_acl);
			else	printf("\t\tNetState Group = free access\n");
			print_var_list("\t\t", iface->var_list);
			print_save_list("\t\t", iface->save_list);
		}

		for (bgp = target->bgp; bgp; bgp = bgp->next) {
			printf("\tBGP = %d", bgp->asn);
			if (bgp->peer) {
				printf(" [");
				for (peer = bgp->peer; peer; peer = peer->next) {
					if (peer != bgp->peer) printf(", ");
					printf("%s", intoa(ipaddr, peer->ip_addr));
				}
				printf("]");
			}
			printf("\n");
			if (bgp->descr)
				printf("\t\tDescription = \"%s\"\n", bgp->descr);
			if (bgp->ns_acl)
				print_group_ref("\t\tNetState ", bgp->ns_acl);
			else	printf("\t\tNetState Group = free access\n");
			print_save_list("\t\t", bgp->save_list);
		}

		for (env = target->env; env; env = env->next) {
			printf("\tEnvMon = %s", get_name_by_type(env->type));
			if (env->gauge) {
				printf(" [");
				for (gauge = env->gauge; gauge; gauge = gauge->next) {
					if (gauge != env->gauge) printf(", ");
					printf("\"%s\"", gauge->name);
				}
				printf("]");
			}
			printf("\n");
			if (env->descr)
				printf("\t\tDescription = \"%s\"\n", env->descr);
			if (env->ns_acl)
				print_group_ref("\t\tNetState ", env->ns_acl);
			else	printf("\t\tNetState Group = free access\n");
			print_save_list("\t\t", env->save_list);
		}

		for (service = target->service; service; service = service->next) {
			printf("\tService = \"%s\"\n", service->name);
			if (service->descr)
				printf("\t\tDescription = \"%s\"\n", service->descr);
			if (service->ns_acl)
				print_group_ref("\t\tNetState ", service->ns_acl);
			else	printf("\t\tNetState Group = free access\n");
			print_var_list("\t\t", service->var_list);
			print_method_list("\t\t", service->method_list);
			print_trap_list("\t\t", service->trap_list);
			print_save_list("\t\t", service->save_list);
		}
	}
}

char *
itoa(number)
	int number;
{
	static char buf[20];
	sprintf(buf, "%d", number);
	return buf;
}

#ifdef	TRACE
void
dump_var(var)
	VARIABLE *var;
{
	if (var) {
		char tmpbuf[1024];
		trace(("\t %s", var->name ? var->name : "Null"));
		if (var->instance)
			trace((".%s", oid2str(tmpbuf, var->instance, var->instance_len)));
		trace((" = %s (%s)", var2str(tmpbuf, var), type2str(var->type)));
		if (var->mode)
			trace((" mode=%06o", var->mode));
		trace(("\n"));
	}
}

void
dump_var_list(var)
	VARIABLE *var;
{
	for (; var; var = var->next) {
		dump_var(var);
	}
}
#endif	/* TRACE */

#ifdef	DEBUG_MEMORY

extern CONFIG *cf;

void
set_cksum()
{
	ROUTER *rp;
	INTERFACE *ifp;
	HOST *hp;

	if (!cf) return;
	for (rp = cf->router; rp; rp = rp->next) {
		rp->cksum = in_cksum((u_short *)rp->name, strlen(rp->name));
		for (ifp = rp->interface; ifp; ifp = ifp->next) {
			ifp->cksum = in_cksum((u_short *)ifp->name, strlen(ifp->name));
		}
	}
	for (hp = cf->host; hp; hp = hp->next) {
		hp->cksum = in_cksum((u_short *)hp->name, strlen(hp->name));
	}
}

int
get_cksum()
{
	ROUTER *rp;
	INTERFACE *ifp;
	HOST *hp;

	if (!cf) return 1;
	for (rp = cf->router; rp; rp = rp->next) {
		if (rp->cksum != in_cksum((u_short *)rp->name, strlen(rp->name)))
			return 0;
		for (ifp = rp->interface; ifp; ifp = ifp->next) {
			if (ifp->cksum != in_cksum((u_short *)ifp->name, strlen(ifp->name)))
				return 0;
		}
	}
	for (hp = cf->host; hp; hp = hp->next) {
		if (hp->cksum != in_cksum((u_short *)hp->name, strlen(hp->name)))
			return 0;
	}
	return 1;
}

dprint_addr()
{
	ROUTER *rp;
	INTERFACE *ifp;
	HOST *hp;

	if (!cf) return;
	for (rp = cf->router; rp; rp = rp->next) {
		printf("\trp=%p rp->name=%p\n", rp, rp->name);
		for (ifp = rp->interface; ifp; ifp = ifp->next) {
			printf("\t\tifp=%p ifp->name=%p\n", ifp, ifp->name);
		}
	}
	for (hp = cf->host; hp; hp = hp->next) {
		printf("\thp=%p hp->name=%p\n", hp, hp->name);
	}
}

#endif
